// SPDX-License-Identifier: GPL-2.0-only
/* Copyright (c) 2023, Intel Corporation. */

#include "idpf.h"

/**
 * idpf_ptp_enable_shtime - Enable shadow time and execute a command
 * @adapter: Driver specific private structure
 */
static void idpf_ptp_enable_shtime(struct idpf_adapter *adapter)
{
	u32 syn_cmd, shtime_enable, exec_cmd;

	/* Get offsets */
	syn_cmd = le32_to_cpu(adapter->ptp.caps->legacy_cross_time_reg.cmd_sync);
	shtime_enable = adapter->ptp.regs.shtime_enable_mask;
	exec_cmd = adapter->ptp.regs.exec_cmd_mask;

	/* Set the shtime en and the sync field */
	writel(shtime_enable, idpf_get_reg_addr(adapter, syn_cmd));
	writel(exec_cmd | shtime_enable, idpf_get_reg_addr(adapter, syn_cmd));
}

#if defined(HAVE_PTP_CROSSTIMESTAMP) && !IS_ENABLED(CONFIG_ARM64)
#if IS_ENABLED(CONFIG_PCIE_PTM)
/**
 * idpf_ptp_get_sync_device_time - Get the cross time stamp info
 * @device: Current device time
 * @system: System counter value read synchronously with device time
 * @ctx: Context provided by timekeeping code
 *
 * Read device and system (ART) clock simultaneously and return the corrected
 * clock values in ns.
 */
static int
idpf_ptp_get_sync_device_time(ktime_t *device,
			      struct system_counterval_t *system,
			      void *ctx)
{
	struct idpf_adapter *adapter = ctx;
	u32 ts_lo, ts_hi, offset_lo, offset_hi;
	struct idpf_ptp *ptp = &adapter->ptp;
	u64 ns_time = 0;

	idpf_ptp_enable_shtime(adapter);

	/* Get offsets */
	offset_lo = le32_to_cpu(ptp->caps->legacy_cross_time_reg.shadow_time_l);
	offset_hi = le32_to_cpu(ptp->caps->legacy_cross_time_reg.shadow_time_h);

	/* Read MTS */
	ts_lo = readl(idpf_get_reg_addr(adapter, offset_lo));
	ts_hi = readl(idpf_get_reg_addr(adapter, offset_hi));

	ns_time = (u64)ts_hi << 32;
	ns_time |= (u64)ts_lo;

	*device = ns_to_ktime(ns_time);

	/* Get the ART offsets */
	offset_lo = le32_to_cpu(ptp->caps->ptm_cross_time_reg.art_l);
	offset_hi = le32_to_cpu(ptp->caps->ptm_cross_time_reg.art_h);

	ts_lo = readl(idpf_get_reg_addr(adapter, offset_lo));
	ts_hi = readl(idpf_get_reg_addr(adapter, offset_hi));

	ns_time = (u64)ts_hi << 32;
	ns_time |= (u64)ts_lo;

	*system = convert_art_ns_to_tsc(ns_time);

	return 0;
}

/**
 * idpf_ptp_get_crosststamp - Capture a device cross timestamp
 * @info: the driver's PTP info structure
 * @cts: The memory to fill the cross timestamp info
 *
 * Capture a cross timestamp between the ART and the device PTP hardware
 * clock. Fill the cross timestamp information and report it back to the
 * caller.
 */
static int
idpf_ptp_get_crosststamp(struct ptp_clock_info *info,
			 struct system_device_crosststamp *cts)
{
	struct idpf_adapter *adapter = idpf_ptp_info_to_adapter(info);

	return get_device_system_crosststamp(idpf_ptp_get_sync_device_time,
					     adapter, NULL, cts);
}

#endif /* CONFIG_PCIE_PTM */
#endif /* HAVE_PTP_CROSSTIMESTAMP && !CONFIG_ARM64 */
/**
 * idpf_ptp_gettimex64 - Get the time of the clock
 * @info: the driver's PTP info structure
 * @ts: timespec64 structure to hold the current time value
 * @sts: Optional parameter for holding a pair of system timestamps from
 *       the system clock. Will be ignored if NULL is given.
 *
 * Read the device clock and return the correct value in ns, after converting it
 * into a timespec struct.
 */
static int
idpf_ptp_gettimex64(struct ptp_clock_info *info, struct timespec64 *ts,
		    struct ptp_system_timestamp *sts)
{
	struct idpf_adapter *adapter = idpf_ptp_info_to_adapter(info);
	struct idpf_ptp *ptp = &adapter->ptp;
	u32 hi, lo, offset_lo, offset_hi;
	u64 ns;

	/* Get offsets */
	offset_lo = le32_to_cpu(ptp->caps->legacy_cross_time_reg.shadow_time_l);
	offset_hi = le32_to_cpu(ptp->caps->legacy_cross_time_reg.shadow_time_h);

	/* Read the system timestamp pre PHC read */
	ptp_read_system_prets(sts);

	idpf_ptp_enable_shtime(adapter);
	lo = readl(idpf_get_reg_addr(adapter, offset_lo));

	/* Read the system timestamp post PHC read */
	ptp_read_system_postts(sts);

	hi = readl(idpf_get_reg_addr(adapter, offset_hi));

	ns = ((u64)hi << 32) | lo;
	dev_dbg(idpf_adapter_to_dev(adapter), "Master time: %llx\n", ns);

	*ts = ns_to_timespec64(ns);

	return 0;
}

#ifndef HAVE_PTP_CLOCK_INFO_GETTIMEX64
/**
 * idpf_ptp_gettime64 - Get the time of the clock
 * @info: the driver's PTP info structure
 * @ts: timespec64 structure to hold the current time value
 *
 * Read the device clock and return the correct value on ns, after converting it
 * into a timespec struct.
 */
static int
idpf_ptp_gettime64(struct ptp_clock_info *info, struct timespec64 *ts)
{
	return idpf_ptp_gettimex64(info, ts, NULL);
}

#ifndef HAVE_PTP_CLOCK_INFO_GETTIME64
/**
 * idpf_ptp_gettime32 - Get the time of the clock
 * @info: the driver's PTP info structure
 * @ts: timespec structure to hold the current time value
 *
 * Read the device clock and return the correct value on ns, after converting it
 * into a timespec struct.
 */
static int idpf_ptp_gettime32(struct ptp_clock_info *info, struct timespec *ts)
{
	struct timespec64 ts64;

	if (idpf_ptp_gettime64(info, &ts64))
		return -EFAULT;

	*ts = timespec64_to_timespec(ts64);
	return 0;
}

#endif /* !HAVE_PTP_CLOCK_INFO_GETTIME64 */
#endif /* !HAVE_PTP_CLOCK_INFO_GETTIMEX64 */
/**
 * idpf_ptp_adjtime - Adjust the time of the clock by the indicated delta
 * @info: the driver's PTP info structure
 * @delta: Offset in nanoseconds to adjust the time by
 */
static int idpf_ptp_adjtime(struct ptp_clock_info *info, s64 delta)
{
	return -EOPNOTSUPP;
}

/**
 * idpf_ptp_adjfine - Adjust clock increment rate
 * @info: the driver's PTP info structure
 * @scaled_ppm: Parts per million with 16-bit fractional field
 *
 * Adjust the frequency of the clock by the indicated scaled ppm from the
 * base frequency.
 */
static int idpf_ptp_adjfine(struct ptp_clock_info *info, long scaled_ppm)
{
	return -EOPNOTSUPP;
}

#ifndef HAVE_PTP_CLOCK_INFO_ADJFINE
/**
 * idpf_ptp_adjfreq - Adjust the frequency of the clock
 * @info: the driver's PTP info structure
 * @ppb: Parts per billion adjustment from the base
 *
 * Adjust the frequency of the clock by the indicated parts per billion from the
 * base frequency.
 */
static int idpf_ptp_adjfreq(struct ptp_clock_info *info, s32 ppb)
{
	return -EOPNOTSUPP;
}

#endif /* HAVE_PTP_CLOCK_INFO_ADJFINE */
/**
 * idpf_ptp_settime64 - Set the time of the clock
 * @info: the driver's PTP info structure
 * @ts: timespec64 structure that holds the new time value
 *
 * Set the device clock to the user input value. The conversion from timespec
 * to ns happens in the write function.
 */
static int
idpf_ptp_settime64(struct ptp_clock_info *info, const struct timespec64 *ts)
{
	return -EOPNOTSUPP;
}

#ifndef HAVE_PTP_CLOCK_INFO_GETTIME64
/**
 * idpf_ptp_settime32 - Set the time of the clock
 * @info: the driver's PTP info structure
 * @ts: timespec structure that holds the new time value
 *
 * Set the device clock to the user input value. The conversion from timespec
 * to ns happens in the write function.
 */
static int
idpf_ptp_settime32(struct ptp_clock_info *info, const struct timespec *ts)
{
	return -EOPNOTSUPP;
}

#endif /* !HAVE_PTP_CLOCK_INFO_GETTIME64 */
/**
 * idpf_ptp_verify_pin - Verify if pin supports requested pin function
 * @info: the driver's PTP info structure
 * @pin: Pin index
 * @func: Assigned function
 * @chan: Assigned channel
 */
static int
idpf_ptp_verify_pin(struct ptp_clock_info *info, unsigned int pin,
		    enum ptp_pin_function func, unsigned int chan)
{
	return -EOPNOTSUPP;
}

/**
 * idpf_ptp_gpio_enable - Enable/disable ancillary features of PHC
 * @info: the driver's PTP info structure
 * @rq: The requested feature to change
 * @on: Enable/disable flag
 */
static int
idpf_ptp_gpio_enable(struct ptp_clock_info *info, struct ptp_clock_request *rq,
		     int on)
{
	return -EOPNOTSUPP;
}

/**
 * idpf_ptp_set_caps - Set PTP capabilities
 * @adapter: Driver specific private structure*
 *
 * This function sets the PTP functions
 */
static void idpf_ptp_set_caps(struct idpf_adapter *adapter)
{
	struct device *dev = idpf_adapter_to_dev(adapter);
	struct ptp_clock_info *info = &adapter->ptp.info;

	snprintf(info->name, sizeof(info->name) - 1, "%s-%s-clk",
		 dev_driver_string(dev), dev_name(dev));

	info->owner = THIS_MODULE;
	info->adjtime = idpf_ptp_adjtime;
#ifdef HAVE_PTP_CLOCK_INFO_ADJFINE
	info->adjfine = idpf_ptp_adjfine;
#else
	info->adjfreq = idpf_ptp_adjfreq;
#endif /* HAVE_PTP_CLOCK_INFO_ADJFINE */
#ifdef HAVE_PTP_CLOCK_INFO_GETTIME64
	info->settime64 = idpf_ptp_settime64;
#else
	info->settime = idpf_ptp_settime32;
#endif /* HAVE_PTP_CLOCK_INFO_GETTIME64 */
	info->verify = idpf_ptp_verify_pin;
	info->enable = idpf_ptp_gpio_enable;
	if (idpf_is_ptp_cap_ena(adapter, VIRTCHNL2_PTP_CAP_LEGACY_CROSS_TIME)) {
#if defined(HAVE_PTP_CLOCK_INFO_GETTIMEX64)
		info->gettimex64 = idpf_ptp_gettimex64;
#elif defined(HAVE_PTP_CLOCK_INFO_GETTIME64)
		info->gettime64 = idpf_ptp_gettime64;
#else
		info->gettime = idpf_ptp_gettime32;
#endif
		dev_dbg(idpf_adapter_to_dev(adapter), "Legacy cross tstamp supported\n");
	} else {
		dev_dbg(idpf_adapter_to_dev(adapter), "Legacy cross time cap not detected\n");
	}
#if defined(HAVE_PTP_CROSSTIMESTAMP) && !IS_ENABLED(CONFIG_ARM64)
#if IS_ENABLED(CONFIG_PCIE_PTM)

	if (idpf_is_ptp_cap_ena(adapter, VIRTCHNL2_PTP_CAP_PTM) &&
	    pcie_ptm_enabled(adapter->pdev) &&
	    boot_cpu_has(X86_FEATURE_ART) &&
	    boot_cpu_has(X86_FEATURE_TSC_KNOWN_FREQ)) {
		info->getcrosststamp = idpf_ptp_get_crosststamp;
		dev_dbg(idpf_adapter_to_dev(adapter), "PTM cross tstamp supported\n");
	} else {
		dev_dbg(idpf_adapter_to_dev(adapter), "PTM not enabled\n");
	}
#endif /* CONFIG_PCIE_PTM */
#endif /* HAVE_PTP_CROSSTIMESTAMP && !CONFIG_ARM64 */
}

/**
 * idpf_ptp_create_clock - Create PTP clock device for userspace
 * @adapter: Driver specific private structure
 *
 * This function creates a new PTP clock device.
 */
static int idpf_ptp_create_clock(struct idpf_adapter *adapter)
{
	struct ptp_clock *clock;

	/* No need to create a clock device if we already have one */
	if (adapter->ptp.clock)
		return 0;

	idpf_ptp_set_caps(adapter);

	/* Attempt to register the clock before enabling the hardware. */
	clock = ptp_clock_register(&adapter->ptp.info,
				   idpf_adapter_to_dev(adapter));
	if (IS_ERR_OR_NULL(clock)) {
		dev_err(idpf_adapter_to_dev(adapter), "PTP clock creation failed\n");
		return -EPERM;
	}

	adapter->ptp.clock = clock;

	return 0;
}

/**
 * idpf_ptp_init - Initialize PTP hardware clock support
 * @adapter: Driver specific private structure
 *
 * Set up the device for interacting with the PTP hardware clock for all
 * functions. Function will allocate and register a ptp_clock with the
 * PTP_1588_CLOCK infrastructure.
 */
int idpf_ptp_init(struct idpf_adapter *adapter)
{
	int err;

	if (!idpf_is_cap_ena(adapter, IDPF_OTHER_CAPS, VIRTCHNL2_CAP_PTP)) {
		dev_dbg(idpf_adapter_to_dev(adapter), "PTP capability not detected\n");
		return 0;
	}

	err = idpf_get_ptp_caps(adapter);
	if (err) {
		dev_err(idpf_adapter_to_dev(adapter), "failed to get PTP caps\n");
		return err;
	}

	if (adapter->dev_ops.reg_ops.ptp_reg_init)
		adapter->dev_ops.reg_ops.ptp_reg_init(adapter);

	err = idpf_ptp_create_clock(adapter);
	if (!err)
		dev_info(idpf_adapter_to_dev(adapter), "PTP init successful\n");
	else
		dev_err(idpf_adapter_to_dev(adapter), "PTP init failed, err=%d\n", err);

	return err;
}

/**
 * idpf_ptp_release - Clear PTP hardware clock support
 * @adapter: Driver specific private structure
 */
void idpf_ptp_release(struct idpf_adapter *adapter)
{
	kfree(adapter->ptp.caps);
	adapter->ptp.caps = NULL;

	if (adapter->ptp.clock) {
		ptp_clock_unregister(adapter->ptp.clock);
		adapter->ptp.clock = NULL;
		memset(&adapter->ptp.info, 0, sizeof(adapter->ptp.info));
	}
}
