// SPDX-License-Identifier: GPL-2.0-only
/* Copyright (C) 2022 Intel Corporation */

#include "idpf.h"
#define BAR4 4
#define LAN_BASE_OFFSET 0x26000000
#define HIF_BASE_OFFSET 0x40000000
#define FXP_BASE_OFFSET 0x29000000

u8 __iomem *bar4_hwaddr, *lan_base, *hif_base, *fxp_base;

#define writebar4(a, value)		writel((value), (bar4_hwaddr + (a)))
#define write_lanbase(a, value)		writel((value), (lan_base + (a)))
#define write_hifbase(a, value)		writel((value), (hif_base + (a)))
#define write_fxpbase(a, value)		writeq((value), (fxp_base + (a)))

int idpf_map_cpf_region(struct idpf_adapter *adapter)
{
	struct pci_dev *pdev = adapter->pdev;

	bar4_hwaddr = ioremap(pci_resource_start(pdev, BAR4),
			      pci_resource_len(pdev, BAR4));
	if (!bar4_hwaddr)
		return -EIO;

	lan_base = bar4_hwaddr + LAN_BASE_OFFSET;
	hif_base = bar4_hwaddr + HIF_BASE_OFFSET;
	fxp_base = bar4_hwaddr + FXP_BASE_OFFSET;

	return 0;
}

static int idpf_write_rx_lan_queue_context(u16 q_id, u64 rx_buf_addr,
					   u64 rx_wb_addr, bool isbufq)
{
	unsigned long long addr;
	dma_addr_t value;

	/* If Buffer Queue, then set bit 14 */
	if (isbufq)
		write_lanbase(0x470028, (q_id | 1 << 14 | 0 << 16 | 1 << 19));
	else
		write_lanbase(0x470028, (q_id | 0 << 16 | 1 << 19));

	value = readl(lan_base + 0x470000);
	/* convert to 128 bits */
	addr = (rx_buf_addr & 0xffffffffffffffff) >> 7;
	value = (value & 0x7f) | ((addr & 0x1ffffff) << 7); /* low 25 bits */
	/* Setting context with 32Byte descriptor format */
	value = value | 0x18;
	write_lanbase(0x470000, value);

	value = (addr >> 25) & 0xffffffff;  /* high 32 bits */
	write_lanbase(0x470000 + 4, value);

	/* write RxMax for length of 1526 bytes */
	write_lanbase(0x470000 + 12, 0x5F6);

	value = rx_wb_addr & 0xffffffff;
	write_lanbase(0x470000 + 24, value);

	value = ((rx_wb_addr >> 32) << 1) & 0xffffffff;
	write_lanbase(0x470000 + 28, value);

	/* Write CQLEN as 512 */
	write_lanbase(0x470008, 0x00420200);

	if (isbufq)
		write_lanbase(0x470028, (q_id | 1 << 14 | 1 << 16 | 1 << 19));
	else
		write_lanbase(0x470028, (q_id | 1 << 16 | 1 << 19));

	/* Write tail for RxQ# 0 */
	if (q_id == 0)
		write_lanbase(0x400000 + (q_id * 4), 0x1FF);

	/* Enable interrupt */
	write_lanbase(0x70004, 0x00010000);

	return 0;
}

static int idpf_write_tx_lan_queue_context(u16 q_id, dma_addr_t buf_addr,
					   dma_addr_t compln_addr)
{
	unsigned long q_len, q_len_lo, q_len_hi;
	unsigned long long addr;
	unsigned long addr_lo;
	unsigned long addr_hi;
	unsigned long val;
	dma_addr_t value;

	write_lanbase(0x341588, (q_id | 0 << 16 | 1 << 19));

	/* convert to 128 bits */
	addr = (buf_addr & 0xffffffffffffffff) >> 7;
	value = addr & 0xffffffff;  /* low 32 bits */
	write_lanbase(0x341560, value);

	value = readl(lan_base + 0x341560 + 4);
	value = (value & 0xfe000000) | ((addr >> 32) & 0x1ffffff);
	write_lanbase(0x341560 + 4, value);

	/* write QLEN */
	q_len = 0x200;
	q_len_lo = q_len & 0xFF;
	write_lanbase(0x00341574, q_len_lo << 23);

	q_len_hi = (q_len >> 8) & 0x1F;
	write_lanbase(0x00341578, q_len_hi);

	write_lanbase(0x341588, (q_id | 1 << 16 | 1 << 19));

	/* MEV_LANPE_LAN_TOP_MAP_TCLAN_REGS_GLTCLAN_CMPQTX_CNTX_DATA_INST */
	#define MEV_LANPE_CMPQTX_CNTX_DATA_INST(_i) (0x00220000 + ((_i) * 4))

	write_lanbase(0x00220020, (q_id | 0 << 16 | 1 << 19));

	val = readl(lan_base + 0x00220000 + 0xC);
	write_lanbase(0x00220000 + 0xC, val | 0x1);

	addr_lo = (compln_addr >> 7) & 0x00000000FFFFFFFFULL;

	write_lanbase(MEV_LANPE_CMPQTX_CNTX_DATA_INST(0), addr_lo);

	addr_hi = (((compln_addr >> 7) >> 32) & 0x00000000FFFFFFFFULL);

	write_lanbase(MEV_LANPE_CMPQTX_CNTX_DATA_INST(1), addr_hi | (1 << 31));

	write_lanbase(MEV_LANPE_CMPQTX_CNTX_DATA_INST(2), 0x10);

	write_lanbase(0x00220020, (q_id | 1 << 16 | 1 << 19));

	/* Enable interrupt - Vector 2 */
	write_lanbase(0x70008, 0x00010000);

	return 0;
}

void idpf_write_queue_context(struct idpf_vport *vport)
{
	/* Write RX Buffer Queue context */
	idpf_write_rx_lan_queue_context(0, /* vport->rx_qs[0].q_id, */
					vport->rx_bufqs[0].dma,
					vport->rx_qs[0].dma, true);
	/* Write Rx Queue context */
	idpf_write_rx_lan_queue_context(0,
					vport->rx_qs[0].dma,
					vport->rx_qs[0].dma, false);

	/* Write TX Lan Queue context */
	idpf_write_tx_lan_queue_context(vport->tx_qs[0].q_id,
					vport->tx_qs[0].dma,
					vport->tx_complqs[0].dma);
}

/* Write default configuration values */
void idpf_write_default_config(void)
{
	/* HIF GL_LAN_INT_ALLOC <- 0x037fe000 */
	writebar4(0x48260238, 0x037fe000);
	/* HIF GL_LCE_INT_ALLOC <- 0x07ffe000 */
	writebar4(0x48260234, 0x07ffe000);
	/* HIF GL_NVME_INT_ALLOC <- 0x07ffe000 */
	writebar4(0x4826023c, 0x07ffe000);

	/*
	 * [mev info] MEV_CONFIG: Initialize RxQ #0:
	 * SplitQ mode, Rx BufferQ #[0, 1], QLen = 128
	 */
	/* LAN QRX_CONTEXT_DATA[0] <- 0x00000001 */
	write_lanbase(0x00470000, 0x00000001);
	/* LAN QRX_CONTEXT_DATA[1] <- 0x00000000 */
	write_lanbase(0x00470004, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[2] <- 0x00420080 */
	write_lanbase(0x00470008, 0x00420080);
	/* LAN QRX_CONTEXT_DATA[3] <- 0x00001000 */
	write_lanbase(0x0047000c, 0x00001000);
	/* LAN QRX_CONTEXT_DATA[4] <- 0x00000008 */
	write_lanbase(0x00470010, 0x00000008);
	/* LAN QRX_CONTEXT_DATA[5] <- 0x04002000 */
	write_lanbase(0x00470014, 0x04002000);
	/* LAN QRX_CONTEXT_DATA[6] <- 0x00000000 */
	write_lanbase(0x00470018, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[7] <- 0x00000000 */
	write_lanbase(0x0047001c, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[8] <- 0x00000000 */
	write_lanbase(0x00470020, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[9] <- 0x00000000 */
	write_lanbase(0x00470024, 0x00000000);
	/* LAN QRX_CONTEXT_CTL <- 0x000b0000 */
	write_lanbase(0x00470028, 0x000b0000);
	/* LAN QRX_CTRL[0] <- 0x00000001 */
	write_lanbase(0x00440000, 0x00000001);

	/* [mev info] MEV_CONFIG: Map RxQ #0 to MSI-X 0, EVV 0 */
	/* LAN GLINT_VECT2FUNC[0] <- 0x00010000 */
	write_lanbase(0x00070000, 0x00010000);
	/* HIF GLINT_VECT2FUNC[0] <- 0x00010000 */
	writebar4(0x48258000, 0x00010000);
	/* HIF GLINT_VECT2EVV_IND_INDX <- 0x00000000 */
	writebar4(0x4800b030, 0x00000000);
	/* HIF GLINT_VECT2EVV_IND_VAL <- 0x00c00000 */
	writebar4(0x4800b034, 0x00400000);
	/* LAN QINT_RQCTL[0] <- 0x40000000 */
	write_lanbase(0x00040000, 0x40000000);
	/* LAN GLINT_ITR[0].GLINT_ITR_RINST[0] <- 0x00000002 */
	write_lanbase(0x00018000, 0x00000002);
	/* LAN GLINT_ITR[1].GLINT_ITR_RINST[0] <- 0x0000000a */
	write_lanbase(0x00020000, 0x0000000a);

	/*
	 * [mev info] MEV_CONFIG: Initialize Rx BufferQ #0:
	 * short (128B) buffers, stride = 16, QLen = 128
	 */
	/* LAN QRX_CONTEXT_DATA[0] <- 0x00000002 */
	write_lanbase(0x00470000, 0x00000002);
	/* LAN QRX_CONTEXT_DATA[1] <- 0x00000000 */
	write_lanbase(0x00470004, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[2] <- 0x00402080 */
	write_lanbase(0x00470008, 0x00402080);
	/* LAN QRX_CONTEXT_DATA[3] <- 0x00001000 */
	write_lanbase(0x0047000c, 0x00001000);
	/* LAN QRX_CONTEXT_DATA[4] <- 0x00300008 */
	write_lanbase(0x00470010, 0x00300008);
	/* LAN QRX_CONTEXT_DATA[5] <- 0x00000000 */
	write_lanbase(0x00470014, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[6] <- 0x00000000 */
	write_lanbase(0x00470018, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[7] <- 0x00000000 */
	write_lanbase(0x0047001c, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[8] <- 0x00000000 */
	write_lanbase(0x00470020, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[9] <- 0x00000000 */
	write_lanbase(0x00470024, 0x00000000);
	/* LAN QRX_CONTEXT_CTL <- 0x000b4000 */
	write_lanbase(0x00470028, 0x000b4000);

	/* [mev info] MEV_CONFIG: Map Rx BufferQ #0 to MSI-X 1, EVV 1 */
	/* LAN GLINT_VECT2FUNC[1] <- 0x00010000 */
	write_lanbase(0x00070004, 0x00010000);
	/* HIF GLINT_VECT2FUNC[1] <- 0x00010000 */
	writebar4(0x48258004, 0x00010000);
	/* HIF GLINT_VECT2EVV_IND_INDX <- 0x00000001 */
	writebar4(0x4800b030, 0x00000001);
	/* HIF GLINT_VECT2EVV_IND_VAL <- 0x00c00001 */
	writebar4(0x4800b034, 0x00400001);
	/* LAN QINT_RBQCTL[0] <- 0x40000001 */
	write_lanbase(0x00050000, 0x40000001);
	/* LAN GLINT_ITR[0].GLINT_ITR_RINST[1] <- 0x00000002 */
	write_lanbase(0x00018004, 0x00000002);
	/* LAN GLINT_ITR[1].GLINT_ITR_RINST[1] <- 0x0000000a */
	write_lanbase(0x00020004, 0x0000000a);

	/*
	 * [mev info] MEV_CONFIG: Initialize Rx BufferQ #1:
	 * long (2048B) buffers, stride = 16, QLen = 128
	 */
	/* LAN QRX_CONTEXT_DATA[0] <- 0x00010082 */
	write_lanbase(0x00470000, 0x00010082);
	/* LAN QRX_CONTEXT_DATA[1] <- 0x00000000 */
	write_lanbase(0x00470004, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[2] <- 0x00420080 */
	write_lanbase(0x00470008, 0x00420080);
	/* LAN QRX_CONTEXT_DATA[3] <- 0x00001000 */
	write_lanbase(0x0047000c, 0x00001000);
	/* LAN QRX_CONTEXT_DATA[4] <- 0x00300008 */
	write_lanbase(0x00470010, 0x00300008);
	/* LAN QRX_CONTEXT_DATA[5] <- 0x00000000 */
	write_lanbase(0x00470014, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[6] <- 0x00000000 */
	write_lanbase(0x00470018, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[7] <- 0x00000000 */
	write_lanbase(0x0047001c, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[8] <- 0x00000000 */
	write_lanbase(0x00470020, 0x00000000);
	/* LAN QRX_CONTEXT_DATA[9] <- 0x00000000 */
	write_lanbase(0x00470024, 0x00000000);
	/* LAN QRX_CONTEXT_CTL <- 0x000b4001 */
	write_lanbase(0x00470028, 0x000b4001);

	/* [mev info] MEV_CONFIG: Map Rx BufferQ #1 to MSI-X 2, EVV 2 */
	/* LAN GLINT_VECT2FUNC[2] <- 0x00010000 */
	write_lanbase(0x00070008, 0x00010000);
	/* HIF GLINT_VECT2FUNC[2] <- 0x00010000 */
	writebar4(0x48258008, 0x00010000);
	/* HIF GLINT_VECT2EVV_IND_INDX <- 0x00000002 */
	writebar4(0x4800b030, 0x00000002);
	/* HIF GLINT_VECT2EVV_IND_VAL <- 0x00c00002 */
	writebar4(0x4800b034, 0x00400002);
	/* LAN QINT_RBQCTL[1] <- 0x40000002 */
	write_lanbase(0x00050004, 0x40000002);
	/* LAN GLINT_ITR[0].GLINT_ITR_RINST[2] <- 0x00000002 */
	write_lanbase(0x00018008, 0x00000002);
	/* LAN GLINT_ITR[1].GLINT_ITR_RINST[2] <- 0x0000000a */
	write_lanbase(0x00020008, 0x0000000a);

	/*
	 * [mev info] MEV_CONFIG: Initialize TxQ #0:
	 * SplitQ mode, Queue scheduling, QLen = 128
	 */
	/* [mev info] MEV_CONFIG:  Using CompQ #0, qnum = 0 */
	/* LAN GLCOMM_QTX_CNTX_DATA[0] <- 0x00000000 */
	write_lanbase(0x00341560, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[1] <- 0x00000000 */
	write_lanbase(0x00341564, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[2] <- 0x00008000 */
	write_lanbase(0x00341568, 0x00008000);
	/* LAN GLCOMM_QTX_CNTX_DATA[3] <- 0x00000005 */
	write_lanbase(0x0034156c, 0x00000005);
	/* LAN GLCOMM_QTX_CNTX_DATA[4] <- 0x00000000 */
	write_lanbase(0x00341570, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[5] <- 0x80000000 */
	write_lanbase(0x00341574, 0x80000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[6] <- 0x00e00200 */
	write_lanbase(0x00341578, 0x00e00200);
	/* LAN GLCOMM_QTX_CNTX_DATA[7] <- 0x00000000 */
	write_lanbase(0x0034157c, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[8] <- 0x00000000 */
	write_lanbase(0x00341580, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_DATA[9] <- 0x00000000 */
	write_lanbase(0x00341584, 0x00000000);
	/* LAN GLCOMM_QTX_CNTX_CTL <- 0x000b0000 */
	write_lanbase(0x00341588, 0x000b0000);

	/* [mev info] MEV_CONFIG: Map TxQ #0 to MSI-X 8, EVV 8 */
	/* LAN GLINT_VECT2FUNC[8] <- 0x00010000 */
	write_lanbase(0x00070020, 0x00010000);
	/* HIF GLINT_VECT2FUNC[8] <- 0x00010000 */
	writebar4(0x48258020, 0x00010000);
	/* HIF GLINT_VECT2EVV_IND_INDX <- 0x00000008 */
	writebar4(0x4800b030, 0x00000008);
	/* HIF GLINT_VECT2EVV_IND_VAL <- 0x00c00008 */
	writebar4(0x4800b034, 0x00400008);
	/* LAN QINT_TQCTL[0] <- 0x40000008 */
	write_lanbase(0x00030000, 0x40000008);
	/* LAN GLINT_ITR[0].GLINT_ITR_RINST[8] <- 0x00000002 */
	write_lanbase(0x00018020, 0x00000002);
	/* LAN GLINT_ITR[1].GLINT_ITR_RINST[8] <- 0x0000000a */
	write_lanbase(0x00020020, 0x0000000a);

	/*
	 * [mev info] MEV_CONFIG: Initialize Tx CompQ #0:
	 * cqe_len = 1 (8B descriptors)
	 */
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[0] <- 0x00000000 */
	write_lanbase(0x00220000, 0x00000000);
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[1] <- 0x80000000 */
	write_lanbase(0x00220004, 0x80000000);
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[2] <- 0x00000004 */
	write_lanbase(0x00220008, 0x00000004);
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[3] <- 0x00000001 */
	write_lanbase(0x0022000c, 0x00000001);
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[4] <- 0x00008000 */
	write_lanbase(0x00220010, 0x00008000);
	/* LAN GLTCLAN_CMPQTX_CNTX_DATA[5] <- 0x00000000 */
	write_lanbase(0x00220014, 0x00000000);
	/* LAN GLTCLAN_CMPQTX_CNTX_CTL <- 0x00090000 */
	write_lanbase(0x00220020, 0x00090000);

	/* LAN QRX_CONTEXT_DATA[5] <- 0x04004001 */
	write_lanbase(0x00470014, 0x04002000);
	/* LAN QINT_TQCTL[0] <- 0x40000008 */
	write_lanbase(0x00030000, 0x40000002);
	/* LAN QINT_RQCTL[0] <- 0x40000001 */
	write_lanbase(0x00040000, 0x40000001);
}

/* Write FXP configuration values */
void idpf_write_fxp_config(void)
{
	write_fxpbase(0x03c40c0, 0x00000000000000a3);
	write_fxpbase(0x03c40c8, 0x00000000000000b3);
	write_fxpbase(0x03c40d0, 0x00000000000000a0);
	write_fxpbase(0x03c40a0, 0x0000000000000104);
	write_fxpbase(0x03c40a8, 0x0000000000000081);
	write_fxpbase(0x03c40b0, 0x0000000000000081);
	write_fxpbase(0x03c40b8, 0x00000000000000a4);
	write_fxpbase(0x03c4148, 0x0000000000000092);
	write_fxpbase(0x03c4150, 0x0000000000000083);
	write_fxpbase(0x03c40f0, 0x0000000000000029);
	write_fxpbase(0x03c40f8, 0x0ffffffffffff8cf);
	write_fxpbase(0x03c4100, 0x0ffffffffff3320f);
	/* write_fxpbase(0x03c4100, 0x0ffffffffff3998f); */
	write_fxpbase(0x03c4108, 0x0fffffffffffffef);
	write_fxpbase(0x03c4110, 0x0ffffffffff321e0);
	write_fxpbase(0x03c4118, 0x0ffffffffe6389e0);
	write_fxpbase(0x03c4120, 0x0fffffffffffb5e0);
	write_fxpbase(0x03e80e8, 0x0fffffffffffffcf);
	write_fxpbase(0x03e80f0, 0x0ffffffffffff8cf);
	write_fxpbase(0x03e80f8, 0x0ffffffffffffccf);
	write_fxpbase(0x03e8100, 0x0fffffffffffc0ef);
	write_fxpbase(0x03e8108, 0x0fffffffffffffef);
	write_fxpbase(0x03e0000, 0x004a39544b622a7e);
	write_fxpbase(0x03e8110, 0x000000000000004f);
	write_fxpbase(0x03e8000, 0x0000000001fc63ff);
	write_fxpbase(0x03e8008, 0x0000000005f06848);
	write_fxpbase(0x03e8010, 0x0000000006f438a7);
	write_fxpbase(0x03e8018, 0x000000000b5bd346);
	write_fxpbase(0x03e8060, 0x0000000058db9efb);
	write_fxpbase(0x03e8068, 0x00000000339db14e);
	write_fxpbase(0x03e8070, 0x0000000015d8a6c7);
	write_fxpbase(0x03e8078, 0x0000000037cfee35);
	write_fxpbase(0x031d9d8, 0x0000000000000007);
	write_fxpbase(0x031d9d0, 0x00000000000002a3);
	write_fxpbase(0x031d968, 0x0000000000000000);
	write_fxpbase(0x031c618, 0x00ffffffffffffff);
	write_fxpbase(0x031ce18, 0x00ffffffffffffff);
	write_fxpbase(0x031d618, 0x0000000000000001);
	write_fxpbase(0x031c090, 0x00ffffffffffffff);
	write_fxpbase(0x031c890, 0x00f7feffffffffff);
	write_fxpbase(0x031d090, 0x0000000000000007);
	write_fxpbase(0x031c380, 0x00f7ffffffffffff);
	write_fxpbase(0x031cb80, 0x00fffeffffffffff);
	write_fxpbase(0x031d380, 0x0000000000000005);
	write_fxpbase(0x031c3e0, 0x00fffeffffffffff);
	write_fxpbase(0x031cbe0, 0x00f7ffffffffffff);
	write_fxpbase(0x031d3e0, 0x0000000000000002);
	write_fxpbase(0x0318000, 0x00000000000bfff4);
	write_fxpbase(0x031a000, 0x0000000000000014);
	write_fxpbase(0x0318008, 0x000000000003fffc);
	write_fxpbase(0x031a008, 0x0000000000000008);
	write_fxpbase(0x0318010, 0x000000000023ffdc);
	write_fxpbase(0x031a010, 0x0000000000000016);
	write_fxpbase(0x0318018, 0x0000000000d3ff2c);
	write_fxpbase(0x031a018, 0x0000000000000072);
	write_fxpbase(0x0318020, 0x0000000002d3fd2c);
	write_fxpbase(0x031a020, 0x0000000000000076);
	write_fxpbase(0x0318028, 0x000000000153feac);
	write_fxpbase(0x031a028, 0x0000000000000086);
	write_fxpbase(0x0318030, 0x000000000353fcac);
	write_fxpbase(0x031a030, 0x000000000000008a);
	write_fxpbase(0x0318038, 0x000000000053ffac);
	write_fxpbase(0x031a038, 0x0000000000000046);
	write_fxpbase(0x0318040, 0x000000002013dfec);
	write_fxpbase(0x031a040, 0x000000000000009a);
	write_fxpbase(0x0318048, 0x000000000013ffec);
	write_fxpbase(0x031a048, 0x000000000000000c);
	write_fxpbase(0x0318050, 0x000000001013efec);
	write_fxpbase(0x031a050, 0x0000000000000012);
	write_fxpbase(0x0318058, 0x00000000000dfff2);
	write_fxpbase(0x031a058, 0x0000000000000015);
	write_fxpbase(0x0318060, 0x000000000005fffa);
	write_fxpbase(0x031a060, 0x0000000000000009);
	write_fxpbase(0x0318068, 0x000000000025ffda);
	write_fxpbase(0x031a068, 0x0000000000000017);
	write_fxpbase(0x0318070, 0x0000000000d5ff2a);
	write_fxpbase(0x031a070, 0x0000000000000073);
	write_fxpbase(0x0318078, 0x0000000002d5fd2a);
	write_fxpbase(0x031a078, 0x0000000000000077);
	write_fxpbase(0x0318080, 0x000000000155feaa);
	write_fxpbase(0x031a080, 0x0000000000000087);
	write_fxpbase(0x0318088, 0x000000000355fcaa);
	write_fxpbase(0x031a088, 0x000000000000008b);
	write_fxpbase(0x0318090, 0x000000000055ffaa);
	write_fxpbase(0x031a090, 0x0000000000000047);
	write_fxpbase(0x0318098, 0x000000002015dfea);
	write_fxpbase(0x031a098, 0x000000000000009b);
	write_fxpbase(0x03180a0, 0x000000000015ffea);
	write_fxpbase(0x031a0a0, 0x000000000000000d);
	write_fxpbase(0x03180a8, 0x000000001015efea);
	write_fxpbase(0x031a0a8, 0x0000000000000013);
	write_fxpbase(0x03180b0, 0x000000000001fffe);
	write_fxpbase(0x031a0b0, 0x0000000000000001);
	write_fxpbase(0x031d9a0, 0x0000000000000011);
	write_fxpbase(0x031d9a8, 0x00001211100f0e0d);
	write_fxpbase(0x031d9b0, 0x0000666521181716);
	write_fxpbase(0x031d808, 0x0001000000000000);
	write_fxpbase(0x031d848, 0x0000000000000000);
	write_fxpbase(0x031d888, 0x0000000000000000);
	write_fxpbase(0x031d8c8, 0x0000000000000000);
	write_fxpbase(0x031d908, 0x0000000000000000);
	write_fxpbase(0x031d838, 0x000100000000000c);
	write_fxpbase(0x031d878, 0x0000000000000002);
	write_fxpbase(0x031d8b8, 0x0000000000000000);
	write_fxpbase(0x031d8f8, 0x8000000000000000);
	write_fxpbase(0x031d938, 0x0000000000000000);
	write_fxpbase(0x031d828, 0x000100000000000c);
	write_fxpbase(0x031d868, 0x0000000000000002);
	write_fxpbase(0x031d8a8, 0x0000000000000000);
	write_fxpbase(0x031d8e8, 0xc000000000000000);
	write_fxpbase(0x031d928, 0x0000000000000000);
	write_fxpbase(0x031d810, 0x000100000000000c);
	write_fxpbase(0x031d850, 0x0000000000000001);
	write_fxpbase(0x031d890, 0x0000000000000000);
	write_fxpbase(0x031d8d0, 0xa000000000000000);
	write_fxpbase(0x031d910, 0x0000000000000000);
	write_fxpbase(0x031d800, 0x000100000000000c);
	write_fxpbase(0x031d840, 0x0000000000000000);
	write_fxpbase(0x031d880, 0x0000000000000000);
	write_fxpbase(0x031d8c0, 0xe000000000000000);
	write_fxpbase(0x031d900, 0x0000000000000000);
	write_fxpbase(0x031d818, 0x002a7bc5f5b193f9);
	write_fxpbase(0x031d858, 0x000000000002832e);
	write_fxpbase(0x031d898, 0xcf991b1f243aa375);
	write_fxpbase(0x031d8d8, 0xd5f61ce97d08049a);
	write_fxpbase(0x031d918, 0x000000000000d770);
	write_fxpbase(0x031d820, 0x0034ebe44644590d);
	write_fxpbase(0x031d860, 0x000000000003d2ab);
	write_fxpbase(0x031d8a0, 0x941c447b960d9498);
	write_fxpbase(0x031d8e0, 0x2c810e4165ab5f5f);
	write_fxpbase(0x031d920, 0x000000000000cc90);
	write_fxpbase(0x031d830, 0x0030a82b0b2ea561);
	write_fxpbase(0x031d870, 0x000000000003dd2c);
	write_fxpbase(0x031d8b0, 0xbd6548f1a049c4de);
	write_fxpbase(0x031d8f0, 0x1502e6deca1eba50);
	write_fxpbase(0x031d930, 0x000000000000ca63);
	write_fxpbase(0x0308018, 0x0000000000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000000000000003);
	write_fxpbase(0x031d9c8, 0x2000000000000003);
	write_fxpbase(0x030c018, 0x0000000c00060001);
	write_fxpbase(0x0310018, 0x0000000000000000);
	write_fxpbase(0x0308030, 0x0000000001000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000000000000006);
	write_fxpbase(0x031d9c8, 0x2000000000000006);
	write_fxpbase(0x030c030, 0x0000000c0007000a);
	write_fxpbase(0x0310030, 0x0000000000000000);
	write_fxpbase(0x03000f0, 0x00000000057fc008);
	write_fxpbase(0x0314030, 0x000000000000000c);
	write_fxpbase(0x0308058, 0xffffffffffffffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00000000000000b);
	write_fxpbase(0x031d9c8, 0x200000000000000b);
	write_fxpbase(0x030c058, 0x0000000400000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000080000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001000000000b);
	write_fxpbase(0x031d9c8, 0x200001000000000b);
	write_fxpbase(0x0310058, 0x0000000000000000);
	write_fxpbase(0x0300058, 0x00000000003fc001);
	write_fxpbase(0x0314058, 0x000000000000000c);
	write_fxpbase(0x0308120, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000080000);
	write_fxpbase(0x031d9c8, 0xa000000100000004);
	write_fxpbase(0x031d9c8, 0x2000000100000004);
	write_fxpbase(0x030c120, 0x0000000c00040005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000000000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010100000004);
	write_fxpbase(0x031d9c8, 0x2000010100000004);
	write_fxpbase(0x0310120, 0x0000000000000000);
	write_fxpbase(0x0308158, 0x00000000ffffffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000080000);
	write_fxpbase(0x031d9c8, 0xe00000010000000b);
	write_fxpbase(0x031d9c8, 0x200000010000000b);
	write_fxpbase(0x030c158, 0x0000000c00090005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000000000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001010000000b);
	write_fxpbase(0x031d9c8, 0x200001010000000b);
	write_fxpbase(0x0310158, 0x0000000000000000);
	write_fxpbase(0x0300318, 0x0000000005ffc007);
	write_fxpbase(0x0314158, 0x000000000000000c);
	write_fxpbase(0x0308208, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000000200000001);
	write_fxpbase(0x031d9c8, 0x2000000200000001);
	write_fxpbase(0x030c208, 0x0000000c00010005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000000a00000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xa000010200000001);
	write_fxpbase(0x031d9c8, 0x2000010200000001);
	write_fxpbase(0x0310208, 0x000000000000000c);
	write_fxpbase(0x0300408, 0x0000000000008001);
	write_fxpbase(0x03004c8, 0x0000000000014002);
	write_fxpbase(0x0314208, 0x000000000000000c);
	write_fxpbase(0x0308248, 0x000000008100ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000000200000009);
	write_fxpbase(0x031d9c8, 0x2000000200000009);
	write_fxpbase(0x030c248, 0x0000001000180013);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000000100000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xa000010200000009);
	write_fxpbase(0x031d9c8, 0x2000010200000009);
	write_fxpbase(0x0310248, 0x000000000000000c);
	write_fxpbase(0x0300448, 0x0000000000008001);
	write_fxpbase(0x0300508, 0x0000000000014002);
	write_fxpbase(0x0314248, 0x000000000000000c);
	write_fxpbase(0x0308260, 0x0000000088a8ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00000020000000c);
	write_fxpbase(0x031d9c8, 0x200000020000000c);
	write_fxpbase(0x030c260, 0x0000001000180002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000000100000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xa00001020000000c);
	write_fxpbase(0x031d9c8, 0x200001020000000c);
	write_fxpbase(0x0310260, 0x000000000000000c);
	write_fxpbase(0x0300460, 0x0000000000008001);
	write_fxpbase(0x0300520, 0x0000000000014002);
	write_fxpbase(0x0314260, 0x000000000000000c);
	write_fxpbase(0x0308290, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000000200000012);
	write_fxpbase(0x031d9c8, 0x2000000200000012);
	write_fxpbase(0x030c290, 0x00000003000c000a);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xe000010200000012);
	write_fxpbase(0x031d9c8, 0x2000010200000012);
	write_fxpbase(0x0310290, 0x000000000000000c);
	write_fxpbase(0x0300490, 0x0000000000008001);
	write_fxpbase(0x0300550, 0x0000000000014002);
	write_fxpbase(0x0314290, 0x000000000000000c);
	write_fxpbase(0x03082a8, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000000200000015);
	write_fxpbase(0x031d9c8, 0x2000000200000015);
	write_fxpbase(0x030c2a8, 0x0000000100090004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xa000010200000015);
	write_fxpbase(0x031d9c8, 0x2000010200000015);
	write_fxpbase(0x03102a8, 0x000000000000000c);
	write_fxpbase(0x03004a8, 0x0000000000008001);
	write_fxpbase(0x0300568, 0x0000000000014002);
	write_fxpbase(0x03142a8, 0x000000000000000c);
	write_fxpbase(0x03082b0, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000000200000016);
	write_fxpbase(0x031d9c8, 0x2000000200000016);
	write_fxpbase(0x030c2b0, 0x000000110014000c);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000003700000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xe000010200000016);
	write_fxpbase(0x031d9c8, 0x2000010200000016);
	write_fxpbase(0x03102b0, 0x000000000000000c);
	write_fxpbase(0x03004b0, 0x0000000000008001);
	write_fxpbase(0x0300570, 0x0000000000014002);
	write_fxpbase(0x03142b0, 0x000000000000000c);
	write_fxpbase(0x03082b8, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000000200000017);
	write_fxpbase(0x031d9c8, 0x2000000200000017);
	write_fxpbase(0x030c2b8, 0x0000000200040000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000001);
	write_fxpbase(0x031d990, 0x0000000003800000);
	write_fxpbase(0x031d998, 0x0000000000000001);
	write_fxpbase(0x031d9c8, 0xe000010200000017);
	write_fxpbase(0x031d9c8, 0x2000010200000017);
	write_fxpbase(0x03102b8, 0x000000000000000c);
	write_fxpbase(0x03004b8, 0x0000000000008001);
	write_fxpbase(0x0300578, 0x0000000000014002);
	write_fxpbase(0x03142b8, 0x000000000000000c);
	write_fxpbase(0x0308410, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xe000000400000002);
	write_fxpbase(0x031d9c8, 0x2000000400000002);
	write_fxpbase(0x030c410, 0x0000000400000003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010400000002);
	write_fxpbase(0x031d9c8, 0x2000010400000002);
	write_fxpbase(0x0310410, 0x0000000000000004);
	write_fxpbase(0x0300810, 0x0000000009428045);
	write_fxpbase(0x0314410, 0x0000000000000004);
	write_fxpbase(0x0308438, 0x000000008100ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xa000000400000007);
	write_fxpbase(0x031d9c8, 0x2000000400000007);
	write_fxpbase(0x030c438, 0x0000000800100004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010400000007);
	write_fxpbase(0x031d9c8, 0x2000010400000007);
	write_fxpbase(0x0310438, 0x0000000000000004);
	write_fxpbase(0x0300838, 0x000000000a428045);
	write_fxpbase(0x03008f8, 0x00000000097fc001);
	write_fxpbase(0x0314438, 0x0000000000000004);
	write_fxpbase(0x0308440, 0x0000000088a8ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xa000000400000008);
	write_fxpbase(0x031d9c8, 0x2000000400000008);
	write_fxpbase(0x030c440, 0x0000000800100005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010400000008);
	write_fxpbase(0x031d9c8, 0x2000010400000008);
	write_fxpbase(0x0310440, 0x0000000000000004);
	write_fxpbase(0x0300840, 0x0000000009428045);
	write_fxpbase(0x0314440, 0x0000000000000004);
	write_fxpbase(0x0308448, 0x0000000088e7ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xe000000400000009);
	write_fxpbase(0x031d9c8, 0x2000000400000009);
	write_fxpbase(0x030c448, 0x0000000200010000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010400000009);
	write_fxpbase(0x031d9c8, 0x2000010400000009);
	write_fxpbase(0x0310448, 0x0000000000000004);
	write_fxpbase(0x0300848, 0x0000000009428045);
	write_fxpbase(0x0314448, 0x0000000000000004);
	write_fxpbase(0x0308458, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xa00000040000000b);
	write_fxpbase(0x031d9c8, 0x200000040000000b);
	write_fxpbase(0x030c458, 0x0000000300040002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001040000000b);
	write_fxpbase(0x031d9c8, 0x200001040000000b);
	write_fxpbase(0x0310458, 0x0000000000000004);
	write_fxpbase(0x0300858, 0x0000000009428045);
	write_fxpbase(0x0314458, 0x0000000000000004);
	write_fxpbase(0x0308460, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xa00000040000000c);
	write_fxpbase(0x031d9c8, 0x200000040000000c);
	write_fxpbase(0x030c460, 0x0000000200030003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001040000000c);
	write_fxpbase(0x031d9c8, 0x200001040000000c);
	write_fxpbase(0x0310460, 0x0000000000000004);
	write_fxpbase(0x0300860, 0x0000000009428045);
	write_fxpbase(0x0314460, 0x0000000000000004);
	write_fxpbase(0x03084a0, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xe000000400000014);
	write_fxpbase(0x031d9c8, 0x2000000400000014);
	write_fxpbase(0x030c4a0, 0x00000008000c0004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003700000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010400000014);
	write_fxpbase(0x031d9c8, 0x2000010400000014);
	write_fxpbase(0x03104a0, 0x0000000000000004);
	write_fxpbase(0x03008a0, 0x0000000009428045);
	write_fxpbase(0x03144a0, 0x0000000000000004);
	write_fxpbase(0x03084a8, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000100000);
	write_fxpbase(0x031d9c8, 0xa000000400000015);
	write_fxpbase(0x031d9c8, 0x2000000400000015);
	write_fxpbase(0x030c4a8, 0x0000000300040002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003800000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010400000015);
	write_fxpbase(0x031d9c8, 0x2000010400000015);
	write_fxpbase(0x03104a8, 0x0000000000000004);
	write_fxpbase(0x0300968, 0x0000000009428045);
	write_fxpbase(0x03144a8, 0x0000000000000004);
	write_fxpbase(0x0308530, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xa000000500000006);
	write_fxpbase(0x031d9c8, 0x2000000500000006);
	write_fxpbase(0x030c530, 0x0000000400030003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010500000006);
	write_fxpbase(0x031d9c8, 0x2000010500000006);
	write_fxpbase(0x0310530, 0x0000000000000004);
	write_fxpbase(0x0300af0, 0x0000000009ffc007);
	write_fxpbase(0x0314530, 0x0000000000000004);
	write_fxpbase(0x0308538, 0x0000000088e7ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xa000000500000007);
	write_fxpbase(0x031d9c8, 0x2000000500000007);
	write_fxpbase(0x030c538, 0x0000000100040002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010500000007);
	write_fxpbase(0x031d9c8, 0x2000010500000007);
	write_fxpbase(0x0310538, 0x0000000000000004);
	write_fxpbase(0x0300a38, 0x0000000009ffc00b);
	write_fxpbase(0x0314538, 0x0000000000000004);
	write_fxpbase(0x0308560, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xe00000050000000c);
	write_fxpbase(0x031d9c8, 0x200000050000000c);
	write_fxpbase(0x030c560, 0x0000000100000002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001050000000c);
	write_fxpbase(0x031d9c8, 0x200001050000000c);
	write_fxpbase(0x0310560, 0x0000000000000004);
	write_fxpbase(0x0300a60, 0x0000000009ffc007);
	write_fxpbase(0x0314560, 0x0000000000000004);
	write_fxpbase(0x0308568, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xe00000050000000d);
	write_fxpbase(0x031d9c8, 0x200000050000000d);
	write_fxpbase(0x030c568, 0x0000000200010000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001050000000d);
	write_fxpbase(0x031d9c8, 0x200001050000000d);
	write_fxpbase(0x0310568, 0x0000000000000004);
	write_fxpbase(0x0300a68, 0x0000000009ffc00d);
	write_fxpbase(0x0314568, 0x0000000000000004);
	write_fxpbase(0x0308580, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xa000000500000010);
	write_fxpbase(0x031d9c8, 0x2000000500000010);
	write_fxpbase(0x030c580, 0x00000003000c0001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003700000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010500000010);
	write_fxpbase(0x031d9c8, 0x2000010500000010);
	write_fxpbase(0x0310580, 0x0000000000000004);
	write_fxpbase(0x0300b40, 0x0000000009ffc004);
	write_fxpbase(0x0314580, 0x0000000000000004);
	write_fxpbase(0x03085b8, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000200000);
	write_fxpbase(0x031d9c8, 0xa000000500000017);
	write_fxpbase(0x031d9c8, 0x2000000500000017);
	write_fxpbase(0x030c5b8, 0x0000000200030001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003800000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010500000017);
	write_fxpbase(0x031d9c8, 0x2000010500000017);
	write_fxpbase(0x03105b8, 0x0000000000000004);
	write_fxpbase(0x0300b78, 0x0000000009ffc008);
	write_fxpbase(0x03145b8, 0x0000000000000004);
	write_fxpbase(0x0308608, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe000000600000001);
	write_fxpbase(0x031d9c8, 0x2000000600000001);
	write_fxpbase(0x030c608, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010600000001);
	write_fxpbase(0x031d9c8, 0x2000010600000001);
	write_fxpbase(0x0310608, 0x0000000000000002);
	write_fxpbase(0x0300c08, 0x0000000000018003);
	write_fxpbase(0x0300cc8, 0x0000000000024004);
	write_fxpbase(0x0314608, 0x0000000000000002);
	write_fxpbase(0x0308618, 0x000000008808ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe000000600000003);
	write_fxpbase(0x031d9c8, 0x2000000600000003);
	write_fxpbase(0x030c618, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010600000003);
	write_fxpbase(0x031d9c8, 0x2000010600000003);
	write_fxpbase(0x0310618, 0x0000000000000002);
	write_fxpbase(0x0300c18, 0x0000000000018003);
	write_fxpbase(0x0300cd8, 0x0000000000024004);
	write_fxpbase(0x0314618, 0x0000000000000002);
	write_fxpbase(0x0308638, 0x0000000088e5ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa000000600000007);
	write_fxpbase(0x031d9c8, 0x2000000600000007);
	write_fxpbase(0x030c638, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000007);
	write_fxpbase(0x031d9c8, 0x2000010600000007);
	write_fxpbase(0x0310638, 0x0000000000000002);
	write_fxpbase(0x0300c38, 0x0000000000018003);
	write_fxpbase(0x0300cf8, 0x0000000000024004);
	write_fxpbase(0x0314638, 0x0000000000000002);
	write_fxpbase(0x0308640, 0x000000000806ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe000000600000008);
	write_fxpbase(0x031d9c8, 0x2000000600000008);
	write_fxpbase(0x030c640, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010600000008);
	write_fxpbase(0x031d9c8, 0x2000010600000008);
	write_fxpbase(0x0310640, 0x0000000000000002);
	write_fxpbase(0x0300c40, 0x0000000000018003);
	write_fxpbase(0x0300d00, 0x0000000000024004);
	write_fxpbase(0x0314640, 0x0000000000000002);
	write_fxpbase(0x0308660, 0x000000008035ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa00000060000000c);
	write_fxpbase(0x031d9c8, 0x200000060000000c);
	write_fxpbase(0x030c660, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001060000000c);
	write_fxpbase(0x031d9c8, 0x200001060000000c);
	write_fxpbase(0x0310660, 0x0000000000000002);
	write_fxpbase(0x0300c60, 0x0000000000018003);
	write_fxpbase(0x0300d20, 0x0000000000024004);
	write_fxpbase(0x0314660, 0x0000000000000002);
	write_fxpbase(0x0308668, 0x0000000088ccffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa00000060000000d);
	write_fxpbase(0x031d9c8, 0x200000060000000d);
	write_fxpbase(0x030c668, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001060000000d);
	write_fxpbase(0x031d9c8, 0x200001060000000d);
	write_fxpbase(0x0310668, 0x0000000000000002);
	write_fxpbase(0x0300c68, 0x0000000000018003);
	write_fxpbase(0x0300d28, 0x0000000000024004);
	write_fxpbase(0x0314668, 0x0000000000000002);
	write_fxpbase(0x0308670, 0x000000008906ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe00000060000000e);
	write_fxpbase(0x031d9c8, 0x200000060000000e);
	write_fxpbase(0x030c670, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001060000000e);
	write_fxpbase(0x031d9c8, 0x200001060000000e);
	write_fxpbase(0x0310670, 0x0000000000000002);
	write_fxpbase(0x0300c70, 0x0000000000018003);
	write_fxpbase(0x0300d30, 0x0000000000024004);
	write_fxpbase(0x0314670, 0x0000000000000002);
	write_fxpbase(0x0308678, 0x000000008914ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe00000060000000f);
	write_fxpbase(0x031d9c8, 0x200000060000000f);
	write_fxpbase(0x030c678, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001060000000f);
	write_fxpbase(0x031d9c8, 0x200001060000000f);
	write_fxpbase(0x0310678, 0x0000000000000002);
	write_fxpbase(0x0300c78, 0x0000000000018003);
	write_fxpbase(0x0300d38, 0x0000000000024004);
	write_fxpbase(0x0314678, 0x0000000000000002);
	write_fxpbase(0x0308680, 0x000000000842ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xe000000600000010);
	write_fxpbase(0x031d9c8, 0x2000000600000010);
	write_fxpbase(0x030c680, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000010);
	write_fxpbase(0x031d9c8, 0x2000010600000010);
	write_fxpbase(0x0310680, 0x0000000000000002);
	write_fxpbase(0x0300c80, 0x0000000000018003);
	write_fxpbase(0x0300d40, 0x0000000000024004);
	write_fxpbase(0x0314680, 0x0000000000000002);
	write_fxpbase(0x0308688, 0x0000000088f7ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa000000600000011);
	write_fxpbase(0x031d9c8, 0x2000000600000011);
	write_fxpbase(0x030c688, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000011);
	write_fxpbase(0x031d9c8, 0x2000010600000011);
	write_fxpbase(0x0310688, 0x0000000000000002);
	write_fxpbase(0x0300c88, 0x0000000000018003);
	write_fxpbase(0x0300d48, 0x0000000000024004);
	write_fxpbase(0x0314688, 0x0000000000000002);
	write_fxpbase(0x0308690, 0x000000008863ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa000000600000012);
	write_fxpbase(0x031d9c8, 0x2000000600000012);
	write_fxpbase(0x030c690, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000012);
	write_fxpbase(0x031d9c8, 0x2000010600000012);
	write_fxpbase(0x0310690, 0x0000000000000002);
	write_fxpbase(0x0300c90, 0x0000000000018003);
	write_fxpbase(0x0300d50, 0x0000000000024004);
	write_fxpbase(0x0314690, 0x0000000000000002);
	write_fxpbase(0x03086a8, 0x000000008864ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa000000600000015);
	write_fxpbase(0x031d9c8, 0x2000000600000015);
	write_fxpbase(0x030c6a8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000015);
	write_fxpbase(0x031d9c8, 0x2000010600000015);
	write_fxpbase(0x03106a8, 0x0000000000000002);
	write_fxpbase(0x0300ca8, 0x0000000000018003);
	write_fxpbase(0x0300d68, 0x0000000000024004);
	write_fxpbase(0x03146a8, 0x0000000000000002);
	write_fxpbase(0x03086b8, 0x0000000022f3ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a00000);
	write_fxpbase(0x031d9c8, 0xa000000600000017);
	write_fxpbase(0x031d9c8, 0x2000000600000017);
	write_fxpbase(0x030c6b8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010600000017);
	write_fxpbase(0x031d9c8, 0x2000010600000017);
	write_fxpbase(0x03106b8, 0x0000000000000002);
	write_fxpbase(0x0300cb8, 0x0000000000018003);
	write_fxpbase(0x0300d78, 0x0000000000024004);
	write_fxpbase(0x03146b8, 0x0000000000000002);
	write_fxpbase(0x0308720, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003680000);
	write_fxpbase(0x031d9c8, 0xe000000700000004);
	write_fxpbase(0x031d9c8, 0x2000000700000004);
	write_fxpbase(0x030c720, 0x0000001200070009);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010700000004);
	write_fxpbase(0x031d9c8, 0x2000010700000004);
	write_fxpbase(0x0310720, 0x0000000000000006);
	write_fxpbase(0x0300e20, 0x0000000000018003);
	write_fxpbase(0x0300ee0, 0x0000000000024004);
	write_fxpbase(0x0314720, 0x0000000000000002);
	write_fxpbase(0x0308738, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003600000);
	write_fxpbase(0x031d9c8, 0xe000000700000007);
	write_fxpbase(0x031d9c8, 0x2000000700000007);
	write_fxpbase(0x030c738, 0x0000000400080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000300000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010700000007);
	write_fxpbase(0x031d9c8, 0x2000010700000007);
	write_fxpbase(0x0310738, 0x0000000000000002);
	write_fxpbase(0x0300e38, 0x0000000000018003);
	write_fxpbase(0x0300ef8, 0x0000000000024004);
	write_fxpbase(0x0314738, 0x0000000000000002);
	write_fxpbase(0x0308778, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003700000);
	write_fxpbase(0x031d9c8, 0xe00000070000000f);
	write_fxpbase(0x031d9c8, 0x200000070000000f);
	write_fxpbase(0x030c778, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000080680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001070000000f);
	write_fxpbase(0x031d9c8, 0x200001070000000f);
	write_fxpbase(0x0310778, 0x0000000000000ffe);
	write_fxpbase(0x0300e78, 0x0000000007c18003);
	write_fxpbase(0x0300f38, 0x0000000000024004);
	write_fxpbase(0x0314778, 0x0000000000000002);
	write_fxpbase(0x0308788, 0x00003fff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003700000);
	write_fxpbase(0x031d9c8, 0xa000000700000011);
	write_fxpbase(0x031d9c8, 0x2000000700000011);
	write_fxpbase(0x030c788, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010700000011);
	write_fxpbase(0x031d9c8, 0x2000010700000011);
	write_fxpbase(0x0310788, 0x00000000000007fe);
	write_fxpbase(0x0300e88, 0x0000000000018003);
	write_fxpbase(0x0300f48, 0x0000000000024004);
	write_fxpbase(0x0314788, 0x0000000000000002);
	write_fxpbase(0x0308798, 0x20003fff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003700000);
	write_fxpbase(0x031d9c8, 0xa000000700000013);
	write_fxpbase(0x031d9c8, 0x2000000700000013);
	write_fxpbase(0x030c798, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000080680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010700000013);
	write_fxpbase(0x031d9c8, 0x2000010700000013);
	write_fxpbase(0x0310798, 0x00000000000009fe);
	write_fxpbase(0x0300e98, 0x0000000007418003);
	write_fxpbase(0x0300f58, 0x0000000000024004);
	write_fxpbase(0x0314798, 0x0000000000000002);
	write_fxpbase(0x03087a8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003800000);
	write_fxpbase(0x031d9c8, 0xe000000700000015);
	write_fxpbase(0x031d9c8, 0x2000000700000015);
	write_fxpbase(0x030c7a8, 0x0000000200080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000780000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010700000015);
	write_fxpbase(0x031d9c8, 0x2000010700000015);
	write_fxpbase(0x03107a8, 0x0000000000000002);
	write_fxpbase(0x0300ea8, 0x0000000000018003);
	write_fxpbase(0x0300f68, 0x0000000000024004);
	write_fxpbase(0x03147a8, 0x0000000000000002);
	write_fxpbase(0x0308900, 0x0000010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000300000);
	write_fxpbase(0x031d9c8, 0xa000000900000000);
	write_fxpbase(0x031d9c8, 0x2000000900000000);
	write_fxpbase(0x030c900, 0x0000000a000e000e);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000400000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010900000000);
	write_fxpbase(0x031d9c8, 0x2000010900000000);
	write_fxpbase(0x0310900, 0x0000000000000008);
	write_fxpbase(0x03012c0, 0x0000000000070006);
	write_fxpbase(0x0314900, 0x0000000000000008);
	write_fxpbase(0x0308978, 0x0100010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000300000);
	write_fxpbase(0x031d9c8, 0xe00000090000000f);
	write_fxpbase(0x031d9c8, 0x200000090000000f);
	write_fxpbase(0x030c978, 0x00000008000e0005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001090000000f);
	write_fxpbase(0x031d9c8, 0x200001090000000f);
	write_fxpbase(0x0310978, 0x0000000000000008);
	write_fxpbase(0x0301338, 0x0000000000070006);
	write_fxpbase(0x0314978, 0x0000000000000008);
	write_fxpbase(0x0308998, 0x0000000001000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000300000);
	write_fxpbase(0x031d9c8, 0xa000000900000013);
	write_fxpbase(0x031d9c8, 0x2000000900000013);
	write_fxpbase(0x030c998, 0x00000004000a0007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010900000013);
	write_fxpbase(0x031d9c8, 0x2000010900000013);
	write_fxpbase(0x0310998, 0x0000000000000004);
	write_fxpbase(0x0301358, 0x0000000000070006);
	write_fxpbase(0x0314998, 0x0000000000000004);
	write_fxpbase(0x0308a18, 0x0000010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000400000);
	write_fxpbase(0x031d9c8, 0xa000000a00000003);
	write_fxpbase(0x031d9c8, 0x2000000a00000003);
	write_fxpbase(0x030ca18, 0x0000000a000e0007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000500000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010a00000003);
	write_fxpbase(0x031d9c8, 0x2000010a00000003);
	write_fxpbase(0x0310a18, 0x0000000000000008);
	write_fxpbase(0x0301418, 0x00000000003fc000);
	write_fxpbase(0x0314a18, 0x0000000000000008);
	write_fxpbase(0x0308a50, 0x0100010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000400000);
	write_fxpbase(0x031d9c8, 0xa000000a0000000a);
	write_fxpbase(0x031d9c8, 0x2000000a0000000a);
	write_fxpbase(0x030ca50, 0x00000008000e0000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010a0000000a);
	write_fxpbase(0x031d9c8, 0x2000010a0000000a);
	write_fxpbase(0x0310a50, 0x0000000000000008);
	write_fxpbase(0x0301510, 0x00000000003fc006);
	write_fxpbase(0x0314a50, 0x0000000000000008);
	write_fxpbase(0x0308ab8, 0x0000000001000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000400000);
	write_fxpbase(0x031d9c8, 0xa000000a00000017);
	write_fxpbase(0x031d9c8, 0x2000000a00000017);
	write_fxpbase(0x030cab8, 0x00000004000a000a);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010a00000017);
	write_fxpbase(0x031d9c8, 0x2000010a00000017);
	write_fxpbase(0x0310ab8, 0x0000000000000004);
	write_fxpbase(0x03014b8, 0x00000000003fc00b);
	write_fxpbase(0x0314ab8, 0x0000000000000004);
	write_fxpbase(0x0308c60, 0x0000010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000500000);
	write_fxpbase(0x031d9c8, 0xe000000c0000000c);
	write_fxpbase(0x031d9c8, 0x2000000c0000000c);
	write_fxpbase(0x030cc60, 0x0000000a000e000a);
	write_fxpbase(0x031d980, 0x000000001ff80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x000000000bf00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010c0000000c);
	write_fxpbase(0x031d9c8, 0x2000010c0000000c);
	write_fxpbase(0x0310c60, 0x0000000000000008);
	write_fxpbase(0x0301860, 0x00000000003fc000);
	write_fxpbase(0x0314c60, 0x0000000000000008);
	write_fxpbase(0x0308c70, 0x0100010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000500000);
	write_fxpbase(0x031d9c8, 0xa000000c0000000e);
	write_fxpbase(0x031d9c8, 0x2000000c0000000e);
	write_fxpbase(0x030cc70, 0x00000008000e000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010c0000000e);
	write_fxpbase(0x031d9c8, 0x2000010c0000000e);
	write_fxpbase(0x0310c70, 0x0000000000000008);
	write_fxpbase(0x0301930, 0x00000000003fc007);
	write_fxpbase(0x0314c70, 0x0000000000000008);
	write_fxpbase(0x0308c98, 0x0000000001000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000500000);
	write_fxpbase(0x031d9c8, 0xe000000c00000013);
	write_fxpbase(0x031d9c8, 0x2000000c00000013);
	write_fxpbase(0x030cc98, 0x00000004000a0004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010c00000013);
	write_fxpbase(0x031d9c8, 0x2000010c00000013);
	write_fxpbase(0x0310c98, 0x0000000000000004);
	write_fxpbase(0x0301898, 0x00000000003fc00b);
	write_fxpbase(0x0314c98, 0x0000000000000004);
	write_fxpbase(0x0308d30, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000900000);
	write_fxpbase(0x031d9c8, 0xe000000d00000006);
	write_fxpbase(0x031d9c8, 0x2000000d00000006);
	write_fxpbase(0x030cd30, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010d00000006);
	write_fxpbase(0x031d9c8, 0x2000010d00000006);
	write_fxpbase(0x0310d30, 0x0000000000000000);
	write_fxpbase(0x0301af0, 0x00000000003fc007);
	write_fxpbase(0x0314d30, 0x0000000000000000);
	write_fxpbase(0x0308d40, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x000000001ff80000);
	write_fxpbase(0x031d988, 0x000000000bf00000);
	write_fxpbase(0x031d9c8, 0xe000000d00000008);
	write_fxpbase(0x031d9c8, 0x2000000d00000008);
	write_fxpbase(0x030cd40, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x000000001c000000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000000000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010d00000008);
	write_fxpbase(0x031d9c8, 0x2000010d00000008);
	write_fxpbase(0x0310d40, 0x0000000000000000);
	write_fxpbase(0x0301b00, 0x00000000003fc00c);
	write_fxpbase(0x0314d40, 0x0000000000010000);
	write_fxpbase(0x0308d70, 0x000000004000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000900000);
	write_fxpbase(0x031d9c8, 0xa000000d0000000e);
	write_fxpbase(0x031d9c8, 0x2000000d0000000e);
	write_fxpbase(0x030cd70, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000080680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010d0000000e);
	write_fxpbase(0x031d9c8, 0x2000010d0000000e);
	write_fxpbase(0x0310d70, 0x00000000000007fc);
	write_fxpbase(0x0301b30, 0x0000000007ffc009);
	write_fxpbase(0x0314d70, 0x0000000000000000);
	write_fxpbase(0x0308d88, 0x00003fff4000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000900000);
	write_fxpbase(0x031d9c8, 0xa000000d00000011);
	write_fxpbase(0x031d9c8, 0x2000000d00000011);
	write_fxpbase(0x030cd88, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010d00000011);
	write_fxpbase(0x031d9c8, 0x2000010d00000011);
	write_fxpbase(0x0310d88, 0x00000000000003fc);
	write_fxpbase(0x0301a88, 0x00000000003fc003);
	write_fxpbase(0x0314d88, 0x0000000000000000);
	write_fxpbase(0x0308d90, 0x20003fff4000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000900000);
	write_fxpbase(0x031d9c8, 0xa000000d00000012);
	write_fxpbase(0x031d9c8, 0x2000000d00000012);
	write_fxpbase(0x030cd90, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000080680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000010d00000012);
	write_fxpbase(0x031d9c8, 0x2000010d00000012);
	write_fxpbase(0x0310d90, 0x0000000000000dfc);
	write_fxpbase(0x0301b50, 0x00000000077fc005);
	write_fxpbase(0x0314d90, 0x0000000000000000);
	write_fxpbase(0x0308da8, 0x000000006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000900000);
	write_fxpbase(0x031d9c8, 0xe000000d00000015);
	write_fxpbase(0x031d9c8, 0x2000000d00000015);
	write_fxpbase(0x030cda8, 0x0000000000060001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000780000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000010d00000015);
	write_fxpbase(0x031d9c8, 0x2000010d00000015);
	write_fxpbase(0x0310da8, 0x0000000000000000);
	write_fxpbase(0x0301aa8, 0x00000000003fc00b);
	write_fxpbase(0x0314da8, 0x0000000000000000);
	write_fxpbase(0x0308e38, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000680000);
	write_fxpbase(0x031d9c8, 0xe000000e00000007);
	write_fxpbase(0x031d9c8, 0x2000000e00000007);
	write_fxpbase(0x030ce38, 0x00000004000c0003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000002);
	write_fxpbase(0x031d990, 0x0000000000700000);
	write_fxpbase(0x031d998, 0x0000000000000002);
	write_fxpbase(0x031d9c8, 0xa000010e00000007);
	write_fxpbase(0x031d9c8, 0x2000010e00000007);
	write_fxpbase(0x0310e38, 0x000000000000490f);
	write_fxpbase(0x0301c38, 0x000000000b434087);
	write_fxpbase(0x0301cf8, 0x000000000004c088);
	write_fxpbase(0x0314e38, 0x0000000000000004);
	write_fxpbase(0x0308e88, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000780000);
	write_fxpbase(0x031d9c8, 0xa000000e00000011);
	write_fxpbase(0x031d9c8, 0x2000000e00000011);
	write_fxpbase(0x030ce88, 0x0000000000060002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000004);
	write_fxpbase(0x031d990, 0x0000000000800000);
	write_fxpbase(0x031d998, 0x0000000000000004);
	write_fxpbase(0x031d9c8, 0xa000010e00000011);
	write_fxpbase(0x031d9c8, 0x2000010e00000011);
	write_fxpbase(0x0310e88, 0x0000000000000000);
	write_fxpbase(0x0301c88, 0x000000000b440007);
	write_fxpbase(0x0301d48, 0x000000000004c008);
	write_fxpbase(0x0314e88, 0x0000000000000000);
	write_fxpbase(0x0309100, 0x000000004000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa000001100000000);
	write_fxpbase(0x031d9c8, 0x2000001100000000);
	write_fxpbase(0x030d100, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001480000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011100000000);
	write_fxpbase(0x031d9c8, 0x2000011100000000);
	write_fxpbase(0x0311100, 0x0000000000000000);
	write_fxpbase(0x0302200, 0x00000000000501b4);
	write_fxpbase(0x0315100, 0x0000000000000018);
	write_fxpbase(0x0309108, 0x000400ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xe000001100000001);
	write_fxpbase(0x031d9c8, 0x2000001100000001);
	write_fxpbase(0x030d108, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001dc0000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011100000001);
	write_fxpbase(0x031d9c8, 0x2000011100000001);
	write_fxpbase(0x0311108, 0x00000000000007fc);
	write_fxpbase(0x03022c8, 0x00000000000501b4);
	write_fxpbase(0x0315108, 0x0000000000000018);
	write_fxpbase(0x0309128, 0x002900ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa000001100000005);
	write_fxpbase(0x031d9c8, 0x2000001100000005);
	write_fxpbase(0x030d128, 0x0000000000060001);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001ec0000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011100000005);
	write_fxpbase(0x031d9c8, 0x2000011100000005);
	write_fxpbase(0x0311128, 0x0000000000000000);
	write_fxpbase(0x03022e8, 0x00000000000501b4);
	write_fxpbase(0x0315128, 0x0000000000000018);
	write_fxpbase(0x0309130, 0x000100ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa000001100000006);
	write_fxpbase(0x031d9c8, 0x2000001100000006);
	write_fxpbase(0x030d130, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000b00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011100000006);
	write_fxpbase(0x031d9c8, 0x2000011100000006);
	write_fxpbase(0x0311130, 0x0000000000000000);
	write_fxpbase(0x03022f0, 0x00000000000501b4);
	write_fxpbase(0x0315130, 0x0000000000000018);
	write_fxpbase(0x0309138, 0x000600ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa000001100000007);
	write_fxpbase(0x031d9c8, 0x2000001100000007);
	write_fxpbase(0x030d138, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000020);
	write_fxpbase(0x031d990, 0x0000000000c00000);
	write_fxpbase(0x031d998, 0x0000000000000020);
	write_fxpbase(0x031d9c8, 0xa000011100000007);
	write_fxpbase(0x031d9c8, 0x2000011100000007);
	write_fxpbase(0x0311138, 0x00000000000058f0);
	write_fxpbase(0x03022f8, 0x00000000000501b4);
	write_fxpbase(0x0315138, 0x0000000000000018);
	write_fxpbase(0x0309158, 0x001100ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xe00000110000000b);
	write_fxpbase(0x031d9c8, 0x200000110000000b);
	write_fxpbase(0x030d158, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000010);
	write_fxpbase(0x031d990, 0x0000000000d00000);
	write_fxpbase(0x031d998, 0x0000000000000010);
	write_fxpbase(0x031d9c8, 0xe00001110000000b);
	write_fxpbase(0x031d9c8, 0x200001110000000b);
	write_fxpbase(0x0311158, 0x0000000000000000);
	write_fxpbase(0x0302318, 0x00000000000501b4);
	write_fxpbase(0x0315158, 0x0000000000000018);
	write_fxpbase(0x0309160, 0x002f00ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa00000110000000c);
	write_fxpbase(0x031d9c8, 0x200000110000000c);
	write_fxpbase(0x030d160, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001110000000c);
	write_fxpbase(0x031d9c8, 0x200001110000000c);
	write_fxpbase(0x0311160, 0x0000000000000000);
	write_fxpbase(0x0302260, 0x00000000000501b4);
	write_fxpbase(0x0315160, 0x0000000000000018);
	write_fxpbase(0x0309170, 0x003200ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000080700000);
	write_fxpbase(0x031d9c8, 0xe00000110000000e);
	write_fxpbase(0x031d9c8, 0x200000110000000e);
	write_fxpbase(0x030d170, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x000000001ff80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x000000000fe80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001110000000e);
	write_fxpbase(0x031d9c8, 0x200001110000000e);
	write_fxpbase(0x0311170, 0x0000000000000000);
	write_fxpbase(0x0302270, 0x00000000000501b4);
	write_fxpbase(0x0302330, 0x00000000067fc00c);
	write_fxpbase(0x0315170, 0x0000000000000018);
	write_fxpbase(0x0309178, 0x003200ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xa00000110000000f);
	write_fxpbase(0x031d9c8, 0x200000110000000f);
	write_fxpbase(0x030d178, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003c00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001110000000f);
	write_fxpbase(0x031d9c8, 0x200001110000000f);
	write_fxpbase(0x0311178, 0x0000000000000000);
	write_fxpbase(0x0302278, 0x00000000000501b4);
	write_fxpbase(0x0302338, 0x00000000067fc008);
	write_fxpbase(0x0315178, 0x0000000000000018);
	write_fxpbase(0x03091b8, 0x008400ff4000f000);
	write_fxpbase(0x031d980, 0x0000000083f80000);
	write_fxpbase(0x031d988, 0x0000000000700000);
	write_fxpbase(0x031d9c8, 0xe000001100000017);
	write_fxpbase(0x031d9c8, 0x2000001100000017);
	write_fxpbase(0x030d1b8, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000f00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011100000017);
	write_fxpbase(0x031d9c8, 0x2000011100000017);
	write_fxpbase(0x03111b8, 0x0000000000000000);
	write_fxpbase(0x0302378, 0x00000000000501b4);
	write_fxpbase(0x03151b8, 0x0000000000000018);
	write_fxpbase(0x0309318, 0x000000006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xa000001300000003);
	write_fxpbase(0x031d9c8, 0x2000001300000003);
	write_fxpbase(0x030d318, 0x00000028001a0027);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000003);
	write_fxpbase(0x031d9c8, 0x2000011300000003);
	write_fxpbase(0x0311318, 0x0000000000000028);
	write_fxpbase(0x03026d8, 0x00000000004500d4);
	write_fxpbase(0x0315318, 0x0000000000000028);
	write_fxpbase(0x0309320, 0x0400ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xa000001300000004);
	write_fxpbase(0x031d9c8, 0x2000001300000004);
	write_fxpbase(0x030d320, 0x0000002800300027);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001dc0000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000004);
	write_fxpbase(0x031d9c8, 0x2000011300000004);
	write_fxpbase(0x0311320, 0x0000000000000024);
	write_fxpbase(0x0302620, 0x00000000004500d4);
	write_fxpbase(0x0315320, 0x0000000000000028);
	write_fxpbase(0x0309330, 0x2900ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe000001300000006);
	write_fxpbase(0x031d9c8, 0x2000001300000006);
	write_fxpbase(0x030d330, 0x00000028002e0012);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001ec0000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000006);
	write_fxpbase(0x031d9c8, 0x2000011300000006);
	write_fxpbase(0x0311330, 0x0000000000000028);
	write_fxpbase(0x03026f0, 0x00000000004500d4);
	write_fxpbase(0x0315330, 0x0000000000000028);
	write_fxpbase(0x0309338, 0x3a00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe000001300000007);
	write_fxpbase(0x031d9c8, 0x2000001300000007);
	write_fxpbase(0x030d338, 0x0000001700190026);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000b00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000007);
	write_fxpbase(0x031d9c8, 0x2000011300000007);
	write_fxpbase(0x0311338, 0x0000000000000028);
	write_fxpbase(0x0302638, 0x00000000004500d4);
	write_fxpbase(0x0315338, 0x0000000000000028);
	write_fxpbase(0x0309350, 0x0600ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe00000130000000a);
	write_fxpbase(0x031d9c8, 0x200000130000000a);
	write_fxpbase(0x030d350, 0x0000000b00240006);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000020);
	write_fxpbase(0x031d990, 0x0000000000c80000);
	write_fxpbase(0x031d998, 0x0000000000000020);
	write_fxpbase(0x031d9c8, 0xe00001130000000a);
	write_fxpbase(0x031d9c8, 0x200001130000000a);
	write_fxpbase(0x0311350, 0x0000000000000028);
	write_fxpbase(0x0302710, 0x00000000004500d4);
	write_fxpbase(0x0315350, 0x0000000000000028);
	write_fxpbase(0x0309358, 0x1100ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe00000130000000b);
	write_fxpbase(0x031d9c8, 0x200000130000000b);
	write_fxpbase(0x030d358, 0x00000028002a0025);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000010);
	write_fxpbase(0x031d990, 0x0000000000d00000);
	write_fxpbase(0x031d998, 0x0000000000000010);
	write_fxpbase(0x031d9c8, 0xe00001130000000b);
	write_fxpbase(0x031d9c8, 0x200001130000000b);
	write_fxpbase(0x0311358, 0x0000000000000028);
	write_fxpbase(0x0302718, 0x00000000004500d4);
	write_fxpbase(0x0315358, 0x0000000000000028);
	write_fxpbase(0x0309368, 0x2f00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xa00000130000000d);
	write_fxpbase(0x031d9c8, 0x200000130000000d);
	write_fxpbase(0x030d368, 0x00000028002a0004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001130000000d);
	write_fxpbase(0x031d9c8, 0x200001130000000d);
	write_fxpbase(0x0311368, 0x0000000000000028);
	write_fxpbase(0x0302728, 0x00000000004500d4);
	write_fxpbase(0x0315368, 0x0000000000000028);
	write_fxpbase(0x0309378, 0x8400ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe00000130000000f);
	write_fxpbase(0x031d9c8, 0x200000130000000f);
	write_fxpbase(0x030d378, 0x00000028002a0023);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000f00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001130000000f);
	write_fxpbase(0x031d9c8, 0x200001130000000f);
	write_fxpbase(0x0311378, 0x0000000000000028);
	write_fxpbase(0x0302678, 0x00000000004500d4);
	write_fxpbase(0x0315378, 0x0000000000000028);
	write_fxpbase(0x0309390, 0x2c00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xa000001300000012);
	write_fxpbase(0x031d9c8, 0x2000001300000012);
	write_fxpbase(0x030d390, 0x00000028002a0000);
	write_fxpbase(0x031d980, 0x0000000013f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000011100000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000012);
	write_fxpbase(0x031d9c8, 0x2000011300000012);
	write_fxpbase(0x0311390, 0x0000000000000028);
	write_fxpbase(0x0302750, 0x00000000004500d4);
	write_fxpbase(0x0315390, 0x0000000000000028);
	write_fxpbase(0x03093a8, 0x3200ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xa000001300000015);
	write_fxpbase(0x031d9c8, 0x2000001300000015);
	write_fxpbase(0x030d3a8, 0x000000280026000a);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003c00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000015);
	write_fxpbase(0x031d9c8, 0x2000011300000015);
	write_fxpbase(0x03113a8, 0x0000000000000028);
	write_fxpbase(0x0302768, 0x00000000064500d4);
	write_fxpbase(0x03153a8, 0x0000000000000028);
	write_fxpbase(0x03093b0, 0x3b00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000800000);
	write_fxpbase(0x031d9c8, 0xe000001300000016);
	write_fxpbase(0x031d9c8, 0x2000001300000016);
	write_fxpbase(0x030d3b0, 0x0000000800070028);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001480000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011300000016);
	write_fxpbase(0x031d9c8, 0x2000011300000016);
	write_fxpbase(0x03113b0, 0x0000000000000028);
	write_fxpbase(0x0302770, 0x00000000004500d4);
	write_fxpbase(0x03153b0, 0x0000000000000028);
	write_fxpbase(0x0309420, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xc000000003c00000);
	write_fxpbase(0x031d9c8, 0xa000001400000004);
	write_fxpbase(0x031d9c8, 0x2000001400000004);
	write_fxpbase(0x030d420, 0x0000000400030003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003d00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011400000004);
	write_fxpbase(0x031d9c8, 0x2000011400000004);
	write_fxpbase(0x0311420, 0x0000000000003098);
	write_fxpbase(0x0302820, 0x00000000003fc001);
	write_fxpbase(0x0315420, 0x0000000000010004);
	write_fxpbase(0x03094b8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000c80000);
	write_fxpbase(0x031d9c8, 0xa000001400000017);
	write_fxpbase(0x031d9c8, 0x2000001400000017);
	write_fxpbase(0x030d4b8, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000c00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011400000017);
	write_fxpbase(0x031d9c8, 0x2000011400000017);
	write_fxpbase(0x03114b8, 0x00000000000058f0);
	write_fxpbase(0x03154b8, 0x0000000000000000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003d00000);
	write_fxpbase(0x031d9c8, 0xe000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003e00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x0000000000000000);
	write_fxpbase(0x0315500, 0x0000000000010000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d40000);
	write_fxpbase(0x031d9c8, 0xe000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x0000000000080007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x00000000000005fc);
	write_fxpbase(0x0315500, 0x0000000000000000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003f90000);
	write_fxpbase(0x031d9c8, 0xe000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x0000000000060004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x0000000000000000);
	write_fxpbase(0x0315500, 0x0000000000000000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d60000);
	write_fxpbase(0x031d9c8, 0xe000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000020);
	write_fxpbase(0x031d990, 0x0000000000c00000);
	write_fxpbase(0x031d998, 0x0000000000000020);
	write_fxpbase(0x031d9c8, 0xe000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x00000000000058f0);
	write_fxpbase(0x0315500, 0x0000000000000000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d10000);
	write_fxpbase(0x031d9c8, 0xa000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x0000000100020001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000010);
	write_fxpbase(0x031d990, 0x0000000000d00000);
	write_fxpbase(0x031d998, 0x0000000000000010);
	write_fxpbase(0x031d9c8, 0xa000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x0000000000000000);
	write_fxpbase(0x0315500, 0x0000000000000000);
	write_fxpbase(0x0309500, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d10000);
	write_fxpbase(0x031d9c8, 0xa000001500000000);
	write_fxpbase(0x031d9c8, 0x2000001500000000);
	write_fxpbase(0x030d500, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011500000000);
	write_fxpbase(0x031d9c8, 0x2000011500000000);
	write_fxpbase(0x0311500, 0x0000000000000000);
	write_fxpbase(0x0315500, 0x0000000000000000);
	write_fxpbase(0x0309700, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000001480000);
	write_fxpbase(0x031d9c8, 0xa000001700000000);
	write_fxpbase(0x031d9c8, 0x2000001700000000);
	write_fxpbase(0x030d700, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000000);
	write_fxpbase(0x031d9c8, 0x2000011700000000);
	write_fxpbase(0x0311700, 0x0000000000000000);
	write_fxpbase(0x0302ec0, 0x000000000033c00d);
	write_fxpbase(0x0315700, 0x0000000000000002);
	write_fxpbase(0x0309710, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000b00000);
	write_fxpbase(0x031d9c8, 0xa000001700000002);
	write_fxpbase(0x031d9c8, 0x2000001700000002);
	write_fxpbase(0x030d710, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000002);
	write_fxpbase(0x031d9c8, 0x2000011700000002);
	write_fxpbase(0x0311710, 0x0000000000000000);
	write_fxpbase(0x0302e10, 0x000000000033c00d);
	write_fxpbase(0x0315710, 0x0000000000000002);
	write_fxpbase(0x0309718, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000c00000);
	write_fxpbase(0x031d9c8, 0xa000001700000003);
	write_fxpbase(0x031d9c8, 0x2000001700000003);
	write_fxpbase(0x030d718, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000003);
	write_fxpbase(0x031d9c8, 0x2000011700000003);
	write_fxpbase(0x0311718, 0x0000000000000000);
	write_fxpbase(0x0302e18, 0x000000000bc84009);
	write_fxpbase(0x0302ed8, 0x000000000033c00d);
	write_fxpbase(0x0315718, 0x0000000000000012);
	write_fxpbase(0x0309720, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001700000004);
	write_fxpbase(0x031d9c8, 0x2000001700000004);
	write_fxpbase(0x030d720, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000004);
	write_fxpbase(0x031d9c8, 0x2000011700000004);
	write_fxpbase(0x0311720, 0x0000000000000000);
	write_fxpbase(0x0302e20, 0x000000000bc58009);
	write_fxpbase(0x0302ee0, 0x000000000033c00d);
	write_fxpbase(0x0315720, 0x0000000000000008);
	write_fxpbase(0x0309728, 0x03e8ffff00000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xa000001700000005);
	write_fxpbase(0x031d9c8, 0x2000001700000005);
	write_fxpbase(0x030d728, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001800000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011700000005);
	write_fxpbase(0x031d9c8, 0x2000011700000005);
	write_fxpbase(0x0311728, 0x0000000000000008);
	write_fxpbase(0x0302e28, 0x000000000bc58009);
	write_fxpbase(0x0315728, 0x0000000000000008);
	write_fxpbase(0x0309730, 0x12b7ffff00000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001700000006);
	write_fxpbase(0x031d9c8, 0x2000001700000006);
	write_fxpbase(0x030d730, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000001000);
	write_fxpbase(0x031d990, 0x0000000000f80000);
	write_fxpbase(0x031d998, 0x0000000000001000);
	write_fxpbase(0x031d9c8, 0xe000011700000006);
	write_fxpbase(0x031d9c8, 0x2000011700000006);
	write_fxpbase(0x0311730, 0x0000000000000000);
	write_fxpbase(0x0302e30, 0x000000000bc58009);
	write_fxpbase(0x0302ef0, 0x000000000033c00d);
	write_fxpbase(0x0315730, 0x0000000000000008);
	write_fxpbase(0x0309738, 0x12b5ffff00000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001700000007);
	write_fxpbase(0x031d9c8, 0x2000001700000007);
	write_fxpbase(0x030d738, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001780000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011700000007);
	write_fxpbase(0x031d9c8, 0x2000011700000007);
	write_fxpbase(0x0311738, 0x0000000000000000);
	write_fxpbase(0x0302e38, 0x000000000bc58009);
	write_fxpbase(0x0302ef8, 0x000000000033c00d);
	write_fxpbase(0x0315738, 0x0000000000000008);
	write_fxpbase(0x0309740, 0x12b6ffff00000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xa000001700000008);
	write_fxpbase(0x031d9c8, 0x2000001700000008);
	write_fxpbase(0x030d740, 0x00000012000e0010);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001700000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000008);
	write_fxpbase(0x031d9c8, 0x2000011700000008);
	write_fxpbase(0x0311740, 0x000000000000483f);
	write_fxpbase(0x0302e40, 0x000000000bc58009);
	write_fxpbase(0x0302f00, 0x000000000033c00d);
	write_fxpbase(0x0315740, 0x0000000000000008);
	write_fxpbase(0x0309748, 0x2122ffff00000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000d00000);
	write_fxpbase(0x031d9c8, 0xa000001700000009);
	write_fxpbase(0x031d9c8, 0x2000001700000009);
	write_fxpbase(0x030d748, 0x0000000700050007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000009);
	write_fxpbase(0x031d9c8, 0x2000011700000009);
	write_fxpbase(0x0311748, 0x000000000000483f);
	write_fxpbase(0x0302e48, 0x000000000bc58009);
	write_fxpbase(0x0302f08, 0x000000000033c00d);
	write_fxpbase(0x0315748, 0x0000000000000008);
	write_fxpbase(0x0309760, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000f00000);
	write_fxpbase(0x031d9c8, 0xa00000170000000c);
	write_fxpbase(0x031d9c8, 0x200000170000000c);
	write_fxpbase(0x030d760, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001170000000c);
	write_fxpbase(0x031d9c8, 0x200001170000000c);
	write_fxpbase(0x0311760, 0x000000000000000c);
	write_fxpbase(0x0311760, 0x0000000000000000);
	write_fxpbase(0x0302e60, 0x000000000bc88009);
	write_fxpbase(0x0302f20, 0x000000000033c00d);
	write_fxpbase(0x0315760, 0x000000000000000c);
	write_fxpbase(0x0309768, 0x0000000000003000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000e00000);
	write_fxpbase(0x031d9c8, 0xa00000170000000d);
	write_fxpbase(0x031d9c8, 0x200000170000000d);
	write_fxpbase(0x030d768, 0x0000000000020002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001170000000d);
	write_fxpbase(0x031d9c8, 0x200001170000000d);
	write_fxpbase(0x0311768, 0x0000000000000000);
	write_fxpbase(0x0302e68, 0x000000000bc68009);
	write_fxpbase(0x0315768, 0x0000000000000004);
	write_fxpbase(0x0309770, 0x0000000010003000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000e00000);
	write_fxpbase(0x031d9c8, 0xa00000170000000e);
	write_fxpbase(0x031d9c8, 0x200000170000000e);
	write_fxpbase(0x030d770, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001170000000e);
	write_fxpbase(0x031d9c8, 0x200001170000000e);
	write_fxpbase(0x0311770, 0x0000000000000004);
	write_fxpbase(0x0302e70, 0x000000000bc68009);
	write_fxpbase(0x0302f30, 0x000000000033c00d);
	write_fxpbase(0x0315770, 0x0000000000000004);
	write_fxpbase(0x0309778, 0x0000000020003000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000e00000);
	write_fxpbase(0x031d9c8, 0xe00000170000000f);
	write_fxpbase(0x031d9c8, 0x200000170000000f);
	write_fxpbase(0x030d778, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001170000000f);
	write_fxpbase(0x031d9c8, 0x200001170000000f);
	write_fxpbase(0x0311778, 0x0000000000000004);
	write_fxpbase(0x0302e78, 0x000000000bc68009);
	write_fxpbase(0x0302f38, 0x000000000033c00d);
	write_fxpbase(0x0315778, 0x0000000000000004);
	write_fxpbase(0x0309788, 0x0000000030003000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000e00000);
	write_fxpbase(0x031d9c8, 0xe000001700000011);
	write_fxpbase(0x031d9c8, 0x2000001700000011);
	write_fxpbase(0x030d788, 0x0000000000020001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011700000011);
	write_fxpbase(0x031d9c8, 0x2000011700000011);
	write_fxpbase(0x0311788, 0x0000000000000008);
	write_fxpbase(0x0302e88, 0x000000000bc68009);
	write_fxpbase(0x0302f48, 0x000000000033c00d);
	write_fxpbase(0x0315788, 0x0000000000000004);
	write_fxpbase(0x03097a0, 0x6558ffff2000ffff);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000000e00000);
	write_fxpbase(0x031d9c8, 0xa000001700000014);
	write_fxpbase(0x031d9c8, 0x2000001700000014);
	write_fxpbase(0x030d7a0, 0x0000001400160005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011700000014);
	write_fxpbase(0x031d9c8, 0x2000011700000014);
	write_fxpbase(0x03117a0, 0x0000000000000008);
	write_fxpbase(0x0302ea0, 0x000000000bc68009);
	write_fxpbase(0x0302f60, 0x000000000033c00d);
	write_fxpbase(0x03157a0, 0x0000000000000004);
	write_fxpbase(0x03097a8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xa000000003d80000);
	write_fxpbase(0x031d9c8, 0xe000001700000015);
	write_fxpbase(0x031d9c8, 0x2000001700000015);
	write_fxpbase(0x030d7a8, 0x0000000700060004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003c00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011700000015);
	write_fxpbase(0x031d9c8, 0x2000011700000015);
	write_fxpbase(0x03117a8, 0x0000000000000008);
	write_fxpbase(0x0302ea8, 0x000000000bc58009);
	write_fxpbase(0x0302f68, 0x000000000033c00d);
	write_fxpbase(0x03157a8, 0x0000000000000006);
	write_fxpbase(0x03097b0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe00000001ff80000);
	write_fxpbase(0x031d988, 0xa00000000fe80000);
	write_fxpbase(0x031d9c8, 0xa000001700000016);
	write_fxpbase(0x031d9c8, 0x2000001700000016);
	write_fxpbase(0x030d7b0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011700000016);
	write_fxpbase(0x031d9c8, 0x2000011700000016);
	write_fxpbase(0x03117b0, 0x0000000000000000);
	write_fxpbase(0x0302f70, 0x000000000033c00d);
	write_fxpbase(0x03157b0, 0x0000000000000000);
	write_fxpbase(0x0309800, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001480000);
	write_fxpbase(0x031d9c8, 0xa000001800000000);
	write_fxpbase(0x031d9c8, 0x2000001800000000);
	write_fxpbase(0x030d800, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000000);
	write_fxpbase(0x031d9c8, 0x2000011800000000);
	write_fxpbase(0x0311800, 0x0000000000000000);
	write_fxpbase(0x0303000, 0x00000000003fc005);
	write_fxpbase(0x0315800, 0x0000000000000000);
	write_fxpbase(0x0309810, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000b00000);
	write_fxpbase(0x031d9c8, 0xe000001800000002);
	write_fxpbase(0x031d9c8, 0x2000001800000002);
	write_fxpbase(0x030d810, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000002);
	write_fxpbase(0x031d9c8, 0x2000011800000002);
	write_fxpbase(0x0311810, 0x0000000000000000);
	write_fxpbase(0x03030d0, 0x00000000003fc000);
	write_fxpbase(0x0315810, 0x0000000000000000);
	write_fxpbase(0x0309818, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000c00000);
	write_fxpbase(0x031d9c8, 0xe000001800000003);
	write_fxpbase(0x031d9c8, 0x2000001800000003);
	write_fxpbase(0x030d818, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011800000003);
	write_fxpbase(0x031d9c8, 0x2000011800000003);
	write_fxpbase(0x0311818, 0x0000000000000000);
	write_fxpbase(0x0303018, 0x000000000bc84009);
	write_fxpbase(0x0315818, 0x0000000000000012);
	write_fxpbase(0x0309828, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xa000001800000005);
	write_fxpbase(0x031d9c8, 0x2000001800000005);
	write_fxpbase(0x030d828, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000005);
	write_fxpbase(0x031d9c8, 0x2000011800000005);
	write_fxpbase(0x0311828, 0x0000000000000000);
	write_fxpbase(0x03030e8, 0x000000000bc58009);
	write_fxpbase(0x0315828, 0x0000000000000008);
	write_fxpbase(0x0309830, 0x03e8ffff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001800000006);
	write_fxpbase(0x031d9c8, 0x2000001800000006);
	write_fxpbase(0x030d830, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001800000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011800000006);
	write_fxpbase(0x031d9c8, 0x2000011800000006);
	write_fxpbase(0x0311830, 0x0000000000000008);
	write_fxpbase(0x0303030, 0x000000000bc58009);
	write_fxpbase(0x0315830, 0x0000000000000008);
	write_fxpbase(0x0309838, 0x12b7ffff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001800000007);
	write_fxpbase(0x031d9c8, 0x2000001800000007);
	write_fxpbase(0x030d838, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000001000);
	write_fxpbase(0x031d990, 0x0000000000f80000);
	write_fxpbase(0x031d998, 0x0000000000001000);
	write_fxpbase(0x031d9c8, 0xa000011800000007);
	write_fxpbase(0x031d9c8, 0x2000011800000007);
	write_fxpbase(0x0311838, 0x0000000000000000);
	write_fxpbase(0x0303038, 0x000000000bc58009);
	write_fxpbase(0x03030f8, 0x000000000008c10a);
	write_fxpbase(0x0315838, 0x0000000000000008);
	write_fxpbase(0x0309840, 0x12b5ffff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xe000001800000008);
	write_fxpbase(0x031d9c8, 0x2000001800000008);
	write_fxpbase(0x030d840, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001780000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000008);
	write_fxpbase(0x031d9c8, 0x2000011800000008);
	write_fxpbase(0x0311840, 0x0000000000000000);
	write_fxpbase(0x0303100, 0x000000000bc58009);
	write_fxpbase(0x0315840, 0x0000000000000008);
	write_fxpbase(0x0309850, 0x12b6ffff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xe00000180000000a);
	write_fxpbase(0x031d9c8, 0x200000180000000a);
	write_fxpbase(0x030d850, 0x0000001200000010);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001700000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001180000000a);
	write_fxpbase(0x031d9c8, 0x200001180000000a);
	write_fxpbase(0x0311850, 0x000000000000483f);
	write_fxpbase(0x0303110, 0x000000000bc58009);
	write_fxpbase(0x0315850, 0x0000000000000008);
	write_fxpbase(0x0309858, 0x2122ffff00000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d00000);
	write_fxpbase(0x031d9c8, 0xa00000180000000b);
	write_fxpbase(0x031d9c8, 0x200000180000000b);
	write_fxpbase(0x030d858, 0x0000000500040007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001600000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001180000000b);
	write_fxpbase(0x031d9c8, 0x200001180000000b);
	write_fxpbase(0x0311858, 0x000000000000483f);
	write_fxpbase(0x0303058, 0x000000000bc58009);
	write_fxpbase(0x0315858, 0x0000000000000008);
	write_fxpbase(0x0309860, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000f00000);
	write_fxpbase(0x031d9c8, 0xe00000180000000c);
	write_fxpbase(0x031d9c8, 0x200000180000000c);
	write_fxpbase(0x030d860, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001180000000c);
	write_fxpbase(0x031d9c8, 0x200001180000000c);
	write_fxpbase(0x0311860, 0x000000000000000c);
	write_fxpbase(0x0303120, 0x000000000bc88009);
	write_fxpbase(0x0315860, 0x000000000000000c);
	write_fxpbase(0x0309878, 0x0000000000003000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e00000);
	write_fxpbase(0x031d9c8, 0xa00000180000000f);
	write_fxpbase(0x031d9c8, 0x200000180000000f);
	write_fxpbase(0x030d878, 0x0000000000020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001180000000f);
	write_fxpbase(0x031d9c8, 0x200001180000000f);
	write_fxpbase(0x0311878, 0x0000000000000000);
	write_fxpbase(0x0303078, 0x000000000bc68009);
	write_fxpbase(0x0315878, 0x0000000000000004);
	write_fxpbase(0x0309880, 0x0000000010003000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e00000);
	write_fxpbase(0x031d9c8, 0xa000001800000010);
	write_fxpbase(0x031d9c8, 0x2000001800000010);
	write_fxpbase(0x030d880, 0x0000000000020001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000010);
	write_fxpbase(0x031d9c8, 0x2000011800000010);
	write_fxpbase(0x0311880, 0x0000000000000004);
	write_fxpbase(0x0303080, 0x000000000bc68009);
	write_fxpbase(0x0315880, 0x0000000000000004);
	write_fxpbase(0x0309890, 0x0000000020003000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e00000);
	write_fxpbase(0x031d9c8, 0xa000001800000012);
	write_fxpbase(0x031d9c8, 0x2000001800000012);
	write_fxpbase(0x030d890, 0x0000000000020001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011800000012);
	write_fxpbase(0x031d9c8, 0x2000011800000012);
	write_fxpbase(0x0311890, 0x0000000000000004);
	write_fxpbase(0x0303090, 0x000000000bc68009);
	write_fxpbase(0x0315890, 0x0000000000000004);
	write_fxpbase(0x03098a0, 0x0000000030003000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e00000);
	write_fxpbase(0x031d9c8, 0xe000001800000014);
	write_fxpbase(0x031d9c8, 0x2000001800000014);
	write_fxpbase(0x030d8a0, 0x0000000000020002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011800000014);
	write_fxpbase(0x031d9c8, 0x2000011800000014);
	write_fxpbase(0x03118a0, 0x0000000000000008);
	write_fxpbase(0x0303160, 0x000000000bc68009);
	write_fxpbase(0x03158a0, 0x0000000000000004);
	write_fxpbase(0x03098a8, 0x6558ffff2000ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e00000);
	write_fxpbase(0x031d9c8, 0xa000001800000015);
	write_fxpbase(0x031d9c8, 0x2000001800000015);
	write_fxpbase(0x030d8a8, 0x0000001400160016);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011800000015);
	write_fxpbase(0x031d9c8, 0x2000011800000015);
	write_fxpbase(0x03118a8, 0x0000000000000008);
	write_fxpbase(0x0303168, 0x000000000bc68009);
	write_fxpbase(0x03158a8, 0x0000000000000004);
	write_fxpbase(0x03098b0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003d80000);
	write_fxpbase(0x031d9c8, 0xa000001800000016);
	write_fxpbase(0x031d9c8, 0x2000001800000016);
	write_fxpbase(0x030d8b0, 0x0000000500070000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003c00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000016);
	write_fxpbase(0x031d9c8, 0x2000011800000016);
	write_fxpbase(0x03118b0, 0x0000000000000008);
	write_fxpbase(0x0303170, 0x000000000bc58009);
	write_fxpbase(0x03158b0, 0x0000000000000006);
	write_fxpbase(0x03098b8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x000000001ff80000);
	write_fxpbase(0x031d988, 0x000000000fe80000);
	write_fxpbase(0x031d9c8, 0xe000001800000017);
	write_fxpbase(0x031d9c8, 0x2000001800000017);
	write_fxpbase(0x030d8b8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011800000017);
	write_fxpbase(0x031d9c8, 0x2000011800000017);
	write_fxpbase(0x03118b8, 0x0000000000000000);
	write_fxpbase(0x03158b8, 0x0000000000000000);
	write_fxpbase(0x0309a78, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001800000);
	write_fxpbase(0x031d9c8, 0xa000001a0000000f);
	write_fxpbase(0x031d9c8, 0x2000001a0000000f);
	write_fxpbase(0x030da78, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001880000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011a0000000f);
	write_fxpbase(0x031d9c8, 0x2000011a0000000f);
	write_fxpbase(0x0311a78, 0x0000000000004e0f);
	write_fxpbase(0x0303478, 0x00000000000c0012);
	write_fxpbase(0x0303538, 0x000000000033c00d);
	write_fxpbase(0x0315a78, 0x0000000000000000);
	write_fxpbase(0x0309ca8, 0x000000008f00ff00);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001880000);
	write_fxpbase(0x031d9c8, 0xe000001c00000015);
	write_fxpbase(0x031d9c8, 0x2000001c00000015);
	write_fxpbase(0x030dca8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001500000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011c00000015);
	write_fxpbase(0x031d9c8, 0x2000011c00000015);
	write_fxpbase(0x0311ca8, 0x0000000000000000);
	write_fxpbase(0x0315ca8, 0x0000000000000000);
	write_fxpbase(0x0309d00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003c00000);
	write_fxpbase(0x031d9c8, 0xe000001d00000000);
	write_fxpbase(0x031d9c8, 0x2000001d00000000);
	write_fxpbase(0x030dd00, 0x0000000600070004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000008);
	write_fxpbase(0x031d990, 0x0000000003e00000);
	write_fxpbase(0x031d998, 0x0000000000000008);
	write_fxpbase(0x031d9c8, 0xa000011d00000000);
	write_fxpbase(0x031d9c8, 0x2000011d00000000);
	write_fxpbase(0x0311d00, 0x0000000000000000);
	write_fxpbase(0x0303a00, 0x000000000473c00d);
	write_fxpbase(0x0315d00, 0x0000000000010004);
	write_fxpbase(0x0309d10, 0x0000000000000000);
	write_fxpbase(0x031d980, 0xe000000003f80000);
	write_fxpbase(0x031d988, 0xc000000003c00000);
	write_fxpbase(0x031d9c8, 0xa000001d00000002);
	write_fxpbase(0x031d9c8, 0x2000001d00000002);
	write_fxpbase(0x030dd10, 0x0000000700050002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003d00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d00000002);
	write_fxpbase(0x031d9c8, 0x2000011d00000002);
	write_fxpbase(0x0311d10, 0x0000000000003098);
	write_fxpbase(0x0303a10, 0x00000000003fc00b);
	write_fxpbase(0x0315d10, 0x0000000000010004);
	write_fxpbase(0x0309d28, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000d80000);
	write_fxpbase(0x031d9c8, 0xe000001d00000005);
	write_fxpbase(0x031d9c8, 0x2000001d00000005);
	write_fxpbase(0x030dd28, 0x0000000700000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d00000005);
	write_fxpbase(0x031d9c8, 0x2000011d00000005);
	write_fxpbase(0x0311d28, 0x0000000000000008);
	write_fxpbase(0x0315d28, 0x0000000000000006);
	write_fxpbase(0x0309d30, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001500000);
	write_fxpbase(0x031d9c8, 0xe000001d00000006);
	write_fxpbase(0x031d9c8, 0x2000001d00000006);
	write_fxpbase(0x030dd30, 0x0000001400000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001680000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000006);
	write_fxpbase(0x031d9c8, 0x2000011d00000006);
	write_fxpbase(0x0311d30, 0x0000000000000010);
	write_fxpbase(0x0303a30, 0x00000000000b8010);
	write_fxpbase(0x0315d30, 0x0000000000000010);
	write_fxpbase(0x0309d38, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000f80000);
	write_fxpbase(0x031d9c8, 0xe000001d00000007);
	write_fxpbase(0x031d9c8, 0x2000001d00000007);
	write_fxpbase(0x030dd38, 0x0000001600000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000007);
	write_fxpbase(0x031d9c8, 0x2000011d00000007);
	write_fxpbase(0x0311d38, 0x0000000000000008);
	write_fxpbase(0x0303af8, 0x000000000008c10a);
	write_fxpbase(0x0315d38, 0x0000000000000006);
	write_fxpbase(0x0309d40, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001780000);
	write_fxpbase(0x031d9c8, 0xa000001d00000008);
	write_fxpbase(0x031d9c8, 0x2000001d00000008);
	write_fxpbase(0x030dd40, 0x0000001c001e000e);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000008);
	write_fxpbase(0x031d9c8, 0x2000011d00000008);
	write_fxpbase(0x0311d40, 0x0000000000000010);
	write_fxpbase(0x0315d40, 0x0000000000000006);
	write_fxpbase(0x0309d48, 0x00000000000300ff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001700000);
	write_fxpbase(0x031d9c8, 0xe000001d00000009);
	write_fxpbase(0x031d9c8, 0x2000001d00000009);
	write_fxpbase(0x030dd48, 0x0000001c00120016);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000009);
	write_fxpbase(0x031d9c8, 0x2000011d00000009);
	write_fxpbase(0x0311d48, 0x0000000000000010);
	write_fxpbase(0x0315d48, 0x0000000000000006);
	write_fxpbase(0x0309d58, 0x00000000000100ff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001700000);
	write_fxpbase(0x031d9c8, 0xa000001d0000000b);
	write_fxpbase(0x031d9c8, 0x2000001d0000000b);
	write_fxpbase(0x030dd58, 0x0000000f0008000e);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d0000000b);
	write_fxpbase(0x031d9c8, 0x2000011d0000000b);
	write_fxpbase(0x0311d58, 0x000000000000000c);
	write_fxpbase(0x0315d58, 0x0000000000000006);
	write_fxpbase(0x0309d60, 0x00000000000200ff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001700000);
	write_fxpbase(0x031d9c8, 0xa000001d0000000c);
	write_fxpbase(0x031d9c8, 0x2000001d0000000c);
	write_fxpbase(0x030dd60, 0x0000001000160004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d0000000c);
	write_fxpbase(0x031d9c8, 0x2000011d0000000c);
	write_fxpbase(0x0311d60, 0x0000000000000010);
	write_fxpbase(0x0315d60, 0x0000000000000006);
	write_fxpbase(0x0309d68, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001600000);
	write_fxpbase(0x031d9c8, 0xa000001d0000000d);
	write_fxpbase(0x031d9c8, 0x2000001d0000000d);
	write_fxpbase(0x030dd68, 0x0000001c000c0011);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d0000000d);
	write_fxpbase(0x031d9c8, 0x2000011d0000000d);
	write_fxpbase(0x0311d68, 0x0000000000000010);
	write_fxpbase(0x0303a68, 0x0000000000338009);
	write_fxpbase(0x0315d68, 0x0000000000000006);
	write_fxpbase(0x0309d70, 0x0000000000008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xe000001d0000000e);
	write_fxpbase(0x031d9c8, 0x2000001d0000000e);
	write_fxpbase(0x030dd70, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d0000000e);
	write_fxpbase(0x031d9c8, 0x2000011d0000000e);
	write_fxpbase(0x0311d70, 0x0000000000000004);
	write_fxpbase(0x0315d70, 0x0000000000000000);
	write_fxpbase(0x0309d78, 0x0000000080008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xa000001d0000000f);
	write_fxpbase(0x031d9c8, 0x2000001d0000000f);
	write_fxpbase(0x030dd78, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d0000000f);
	write_fxpbase(0x031d9c8, 0x2000011d0000000f);
	write_fxpbase(0x0311d78, 0x0000000000000008);
	write_fxpbase(0x0315d78, 0x0000000000000000);
	write_fxpbase(0x0309d80, 0x0800ffff00008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xe000001d00000010);
	write_fxpbase(0x031d9c8, 0x2000001d00000010);
	write_fxpbase(0x030dd80, 0x0000000300040004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000010);
	write_fxpbase(0x031d9c8, 0x2000011d00000010);
	write_fxpbase(0x0311d80, 0x0000000000000000);
	write_fxpbase(0x0315d80, 0x0000000000000000);
	write_fxpbase(0x0309d88, 0x0800ffff80008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xa000001d00000011);
	write_fxpbase(0x031d9c8, 0x2000001d00000011);
	write_fxpbase(0x030dd88, 0x0000000600020004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011d00000011);
	write_fxpbase(0x031d9c8, 0x2000011d00000011);
	write_fxpbase(0x0311d88, 0x0000000000000004);
	write_fxpbase(0x0315d88, 0x0000000000000000);
	write_fxpbase(0x0309d90, 0x86ddffff00008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xe000001d00000012);
	write_fxpbase(0x031d9c8, 0x2000001d00000012);
	write_fxpbase(0x030dd90, 0x00000004000a0005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001f00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000012);
	write_fxpbase(0x031d9c8, 0x2000011d00000012);
	write_fxpbase(0x0311d90, 0x0000000000000004);
	write_fxpbase(0x0315d90, 0x0000000000000000);
	write_fxpbase(0x0309d98, 0x86ddffff80008000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000e80000);
	write_fxpbase(0x031d9c8, 0xe000001d00000013);
	write_fxpbase(0x031d9c8, 0x2000001d00000013);
	write_fxpbase(0x030dd98, 0x00000008000e0008);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001f00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011d00000013);
	write_fxpbase(0x031d9c8, 0x2000011d00000013);
	write_fxpbase(0x0311d98, 0x0000000000000008);
	write_fxpbase(0x0315d98, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003d00000);
	write_fxpbase(0x031d9c8, 0xe000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x0000000000000000);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d40000);
	write_fxpbase(0x031d9c8, 0xa000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x0000000000080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x00000000000007fc);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003f90000);
	write_fxpbase(0x031d9c8, 0xe000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x0000000000060005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x0000000000000000);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d60000);
	write_fxpbase(0x031d9c8, 0xa000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002100000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x00000000000058f0);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003ff0000);
	write_fxpbase(0x031d988, 0x0000000003d10000);
	write_fxpbase(0x031d9c8, 0xa000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x0000000000000000);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x0309e00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001680000);
	write_fxpbase(0x031d9c8, 0xe000001e00000000);
	write_fxpbase(0x031d9c8, 0x2000001e00000000);
	write_fxpbase(0x030de00, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000002000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000002000);
	write_fxpbase(0x031d9c8, 0xa000011e00000000);
	write_fxpbase(0x031d9c8, 0x2000011e00000000);
	write_fxpbase(0x0311e00, 0x000000000000000c);
	write_fxpbase(0x0303c00, 0x00000000000bc011);
	write_fxpbase(0x0315e00, 0x0000000000000000);
	write_fxpbase(0x030a000, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe000002000000000);
	write_fxpbase(0x031d9c8, 0x2000002000000000);
	write_fxpbase(0x030e000, 0x0000000c000c0000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000003e00000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xa000012000000000);
	write_fxpbase(0x031d9c8, 0x2000012000000000);
	write_fxpbase(0x0312000, 0x000000000000000c);
	write_fxpbase(0x0304000, 0x000000000cc0c00b);
	write_fxpbase(0x03040c0, 0x0000000000014002);
	write_fxpbase(0x0316000, 0x000000000000000c);
	write_fxpbase(0x030a020, 0x000000008100ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe000002000000004);
	write_fxpbase(0x031d9c8, 0x2000002000000004);
	write_fxpbase(0x030e020, 0x00000010000c0001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000001980000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xe000012000000004);
	write_fxpbase(0x031d9c8, 0x2000012000000004);
	write_fxpbase(0x0312020, 0x000000000000000c);
	write_fxpbase(0x0304020, 0x000000000cc0c00b);
	write_fxpbase(0x03040e0, 0x0000000000014002);
	write_fxpbase(0x0316020, 0x000000000000000c);
	write_fxpbase(0x030a040, 0x0000000088a8ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe000002000000008);
	write_fxpbase(0x031d9c8, 0x2000002000000008);
	write_fxpbase(0x030e040, 0x0000001000020004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000001980000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xe000012000000008);
	write_fxpbase(0x031d9c8, 0x2000012000000008);
	write_fxpbase(0x0312040, 0x000000000000000c);
	write_fxpbase(0x0304040, 0x000000000cc0c00b);
	write_fxpbase(0x0304100, 0x0000000000014002);
	write_fxpbase(0x0316040, 0x000000000000000c);
	write_fxpbase(0x030a058, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe00000200000000b);
	write_fxpbase(0x031d9c8, 0x200000200000000b);
	write_fxpbase(0x030e058, 0x0000000000080004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xa00001200000000b);
	write_fxpbase(0x031d9c8, 0x200001200000000b);
	write_fxpbase(0x0312058, 0x000000000000000c);
	write_fxpbase(0x0304058, 0x000000000cc0c00b);
	write_fxpbase(0x0304118, 0x0000000000014002);
	write_fxpbase(0x0316058, 0x000000000000000c);
	write_fxpbase(0x030a090, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xa000002000000012);
	write_fxpbase(0x031d9c8, 0x2000002000000012);
	write_fxpbase(0x030e090, 0x0000000500070003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xa000012000000012);
	write_fxpbase(0x031d9c8, 0x2000012000000012);
	write_fxpbase(0x0312090, 0x000000000000000c);
	write_fxpbase(0x0304090, 0x000000000cc0c00b);
	write_fxpbase(0x0304150, 0x0000000000014002);
	write_fxpbase(0x0316090, 0x000000000000000c);
	write_fxpbase(0x030a0a0, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe000002000000014);
	write_fxpbase(0x031d9c8, 0x2000002000000014);
	write_fxpbase(0x030e0a0, 0x0000000c00050004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000003a00000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xa000012000000014);
	write_fxpbase(0x031d9c8, 0x2000012000000014);
	write_fxpbase(0x03120a0, 0x000000000000000c);
	write_fxpbase(0x03040a0, 0x000000000cc0c00b);
	write_fxpbase(0x0304160, 0x0000000000014002);
	write_fxpbase(0x03160a0, 0x000000000000000c);
	write_fxpbase(0x030a0a8, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001900000);
	write_fxpbase(0x031d9c8, 0xe000002000000015);
	write_fxpbase(0x031d9c8, 0x2000002000000015);
	write_fxpbase(0x030e0a8, 0x0000000a00030002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000040);
	write_fxpbase(0x031d990, 0x0000000003b00000);
	write_fxpbase(0x031d998, 0x0000000000000040);
	write_fxpbase(0x031d9c8, 0xe000012000000015);
	write_fxpbase(0x031d9c8, 0x2000012000000015);
	write_fxpbase(0x03120a8, 0x000000000000000c);
	write_fxpbase(0x03040a8, 0x000000000cc0c00b);
	write_fxpbase(0x0304168, 0x0000000000014002);
	write_fxpbase(0x03160a8, 0x000000000000000c);
	write_fxpbase(0x030a138, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xa000002100000007);
	write_fxpbase(0x031d9c8, 0x2000002100000007);
	write_fxpbase(0x030e138, 0x0000000400000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012100000007);
	write_fxpbase(0x031d9c8, 0x2000012100000007);
	write_fxpbase(0x0312138, 0x0000000000000002);
	write_fxpbase(0x0304238, 0x000000000d42c053);
	write_fxpbase(0x0316138, 0x0000000000000004);
	write_fxpbase(0x030a158, 0x000000008100ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xe00000210000000b);
	write_fxpbase(0x031d9c8, 0x200000210000000b);
	write_fxpbase(0x030e158, 0x0000000800030007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001210000000b);
	write_fxpbase(0x031d9c8, 0x200001210000000b);
	write_fxpbase(0x0312158, 0x0000000000000004);
	write_fxpbase(0x0304258, 0x000000000d42c053);
	write_fxpbase(0x0304318, 0x000000000e7fc000);
	write_fxpbase(0x0316158, 0x0000000000000004);
	write_fxpbase(0x030a170, 0x0000000088a8ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xe00000210000000e);
	write_fxpbase(0x031d9c8, 0x200000210000000e);
	write_fxpbase(0x030e170, 0x0000000800080001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001210000000e);
	write_fxpbase(0x031d9c8, 0x200001210000000e);
	write_fxpbase(0x0312170, 0x0000000000000004);
	write_fxpbase(0x0304270, 0x000000000d42c053);
	write_fxpbase(0x0316170, 0x0000000000000004);
	write_fxpbase(0x030a178, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xa00000210000000f);
	write_fxpbase(0x031d9c8, 0x200000210000000f);
	write_fxpbase(0x030e178, 0x0000000300010000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001210000000f);
	write_fxpbase(0x031d9c8, 0x200001210000000f);
	write_fxpbase(0x0312178, 0x0000000000000004);
	write_fxpbase(0x0304338, 0x000000000d42c053);
	write_fxpbase(0x0316178, 0x0000000000000004);
	write_fxpbase(0x030a188, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xe000002100000011);
	write_fxpbase(0x031d9c8, 0x2000002100000011);
	write_fxpbase(0x030e188, 0x0000000200000003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012100000011);
	write_fxpbase(0x031d9c8, 0x2000012100000011);
	write_fxpbase(0x0312188, 0x0000000000000004);
	write_fxpbase(0x0304348, 0x000000000d42c053);
	write_fxpbase(0x0316188, 0x0000000000000004);
	write_fxpbase(0x030a198, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xe000002100000013);
	write_fxpbase(0x031d9c8, 0x2000002100000013);
	write_fxpbase(0x030e198, 0x0000000300000004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012100000013);
	write_fxpbase(0x031d9c8, 0x2000012100000013);
	write_fxpbase(0x0312198, 0x0000000000000004);
	write_fxpbase(0x0304358, 0x000000000d42c053);
	write_fxpbase(0x0316198, 0x0000000000000004);
	write_fxpbase(0x030a1b0, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001980000);
	write_fxpbase(0x031d9c8, 0xe000002100000016);
	write_fxpbase(0x031d9c8, 0x2000002100000016);
	write_fxpbase(0x030e1b0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003b00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012100000016);
	write_fxpbase(0x031d9c8, 0x2000012100000016);
	write_fxpbase(0x03121b0, 0x0000000000000004);
	write_fxpbase(0x03042b0, 0x000000000d42c053);
	write_fxpbase(0x03161b0, 0x0000000000000004);
	write_fxpbase(0x030a208, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001a00000);
	write_fxpbase(0x031d9c8, 0xa000002200000001);
	write_fxpbase(0x031d9c8, 0x2000002200000001);
	write_fxpbase(0x030e208, 0x0000000400060000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012200000001);
	write_fxpbase(0x031d9c8, 0x2000012200000001);
	write_fxpbase(0x0312208, 0x0000000000000002);
	write_fxpbase(0x0304408, 0x000000000dffc00f);
	write_fxpbase(0x0316208, 0x0000000000000004);
	write_fxpbase(0x030a248, 0x000000008847ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001a00000);
	write_fxpbase(0x031d9c8, 0xa000002200000009);
	write_fxpbase(0x031d9c8, 0x2000002200000009);
	write_fxpbase(0x030e248, 0x0000000000040001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012200000009);
	write_fxpbase(0x031d9c8, 0x2000012200000009);
	write_fxpbase(0x0312248, 0x0000000000000004);
	write_fxpbase(0x0304508, 0x000000000dffc001);
	write_fxpbase(0x0316248, 0x0000000000000004);
	write_fxpbase(0x030a250, 0x000000008848ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001a00000);
	write_fxpbase(0x031d9c8, 0xa00000220000000a);
	write_fxpbase(0x031d9c8, 0x200000220000000a);
	write_fxpbase(0x030e250, 0x0000000000020004);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003900000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001220000000a);
	write_fxpbase(0x031d9c8, 0x200001220000000a);
	write_fxpbase(0x0312250, 0x0000000000000004);
	write_fxpbase(0x0304510, 0x000000000dffc009);
	write_fxpbase(0x0316250, 0x0000000000000004);
	write_fxpbase(0x030a258, 0x000000000800ffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001a00000);
	write_fxpbase(0x031d9c8, 0xa00000220000000b);
	write_fxpbase(0x031d9c8, 0x200000220000000b);
	write_fxpbase(0x030e258, 0x0000000100010000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003a00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001220000000b);
	write_fxpbase(0x031d9c8, 0x200001220000000b);
	write_fxpbase(0x0312258, 0x0000000000000004);
	write_fxpbase(0x0304458, 0x000000000dffc00b);
	write_fxpbase(0x0316258, 0x0000000000000004);
	write_fxpbase(0x030a288, 0x0000000086ddffff);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001a00000);
	write_fxpbase(0x031d9c8, 0xa000002200000011);
	write_fxpbase(0x031d9c8, 0x2000002200000011);
	write_fxpbase(0x030e288, 0x0000000400010000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003b00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012200000011);
	write_fxpbase(0x031d9c8, 0x2000012200000011);
	write_fxpbase(0x0312288, 0x0000000000000004);
	write_fxpbase(0x0304488, 0x000000000dffc00d);
	write_fxpbase(0x0316288, 0x0000000000000004);
	write_fxpbase(0x030a458, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003900000);
	write_fxpbase(0x031d9c8, 0xa00000240000000b);
	write_fxpbase(0x031d9c8, 0x200000240000000b);
	write_fxpbase(0x030e458, 0x0000000400080000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001b00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe00001240000000b);
	write_fxpbase(0x031d9c8, 0x200001240000000b);
	write_fxpbase(0x0312458, 0x0000000000000002);
	write_fxpbase(0x0304858, 0x000000000001c00c);
	write_fxpbase(0x0304918, 0x0000000000024004);
	write_fxpbase(0x0316458, 0x0000000000000002);
	write_fxpbase(0x030a490, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003a00000);
	write_fxpbase(0x031d9c8, 0xa000002400000012);
	write_fxpbase(0x031d9c8, 0x2000002400000012);
	write_fxpbase(0x030e490, 0x00000002000a0001);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012400000012);
	write_fxpbase(0x031d9c8, 0x2000012400000012);
	write_fxpbase(0x0312490, 0x0000000000000bfe);
	write_fxpbase(0x0304890, 0x000000000001c00c);
	write_fxpbase(0x0304950, 0x0000000000024004);
	write_fxpbase(0x0316490, 0x0000000000000002);
	write_fxpbase(0x030a498, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000003b00000);
	write_fxpbase(0x031d9c8, 0xe000002400000013);
	write_fxpbase(0x031d9c8, 0x2000002400000013);
	write_fxpbase(0x030e498, 0x0000000200080003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012400000013);
	write_fxpbase(0x031d9c8, 0x2000012400000013);
	write_fxpbase(0x0312498, 0x0000000000000002);
	write_fxpbase(0x0304898, 0x000000000001c00c);
	write_fxpbase(0x0304958, 0x0000000000024004);
	write_fxpbase(0x0316498, 0x0000000000000002);
	write_fxpbase(0x030a620, 0x0000010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001b00000);
	write_fxpbase(0x031d9c8, 0xa000002600000004);
	write_fxpbase(0x031d9c8, 0x2000002600000004);
	write_fxpbase(0x030e620, 0x0000000200070008);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012600000004);
	write_fxpbase(0x031d9c8, 0x2000012600000004);
	write_fxpbase(0x0312620, 0x0000000000000008);
	write_fxpbase(0x0304c20, 0x000000000007400d);
	write_fxpbase(0x0316620, 0x0000000000000000);
	write_fxpbase(0x030a698, 0x0100010000000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001b00000);
	write_fxpbase(0x031d9c8, 0xa000002600000013);
	write_fxpbase(0x031d9c8, 0x2000002600000013);
	write_fxpbase(0x030e698, 0x0000000800020000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002000000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012600000013);
	write_fxpbase(0x031d9c8, 0x2000012600000013);
	write_fxpbase(0x0312698, 0x0000000000000008);
	write_fxpbase(0x0304d58, 0x000000000007400d);
	write_fxpbase(0x0316698, 0x0000000000000000);
	write_fxpbase(0x030a6b0, 0x0000000001000100);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001b00000);
	write_fxpbase(0x031d9c8, 0xe000002600000016);
	write_fxpbase(0x031d9c8, 0x2000002600000016);
	write_fxpbase(0x030e6b0, 0x0000000400010003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002000000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012600000016);
	write_fxpbase(0x031d9c8, 0x2000012600000016);
	write_fxpbase(0x03126b0, 0x0000000000000004);
	write_fxpbase(0x0304cb0, 0x000000000007400d);
	write_fxpbase(0x03166b0, 0x0000000000000000);
	write_fxpbase(0x030a770, 0x000000004000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002000000);
	write_fxpbase(0x031d9c8, 0xe00000270000000e);
	write_fxpbase(0x031d9c8, 0x200000270000000e);
	write_fxpbase(0x030e770, 0x0000000000080004);
	write_fxpbase(0x0312770, 0x00000000000005fc);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001d80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa00001270000000e);
	write_fxpbase(0x031d9c8, 0x200001270000000e);
	write_fxpbase(0x030a780, 0x000000006000f000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002000000);
	write_fxpbase(0x031d9c8, 0xa000002700000010);
	write_fxpbase(0x031d9c8, 0x2000002700000010);
	write_fxpbase(0x030e780, 0x0000000000060004);
	write_fxpbase(0x0312780, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000001e80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012700000010);
	write_fxpbase(0x031d9c8, 0x2000012700000010);
	write_fxpbase(0x030a928, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001d80000);
	write_fxpbase(0x031d9c8, 0xa000002900000005);
	write_fxpbase(0x031d9c8, 0x2000002900000005);
	write_fxpbase(0x030e928, 0x00000004000c0003);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000080);
	write_fxpbase(0x031d990, 0x0000000001e00000);
	write_fxpbase(0x031d998, 0x0000000000000080);
	write_fxpbase(0x031d9c8, 0xe000012900000005);
	write_fxpbase(0x031d9c8, 0x2000012900000005);
	write_fxpbase(0x0312928, 0x000000000000490f);
	write_fxpbase(0x0305228, 0x000000000f43808e);
	write_fxpbase(0x03052e8, 0x000000000004c088);
	write_fxpbase(0x0316928, 0x0000000000000004);
	write_fxpbase(0x030a988, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000001e80000);
	write_fxpbase(0x031d9c8, 0xa000002900000011);
	write_fxpbase(0x031d9c8, 0x2000002900000011);
	write_fxpbase(0x030e988, 0x0000000000060006);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000100);
	write_fxpbase(0x031d990, 0x0000000001f00000);
	write_fxpbase(0x031d998, 0x0000000000000100);
	write_fxpbase(0x031d9c8, 0xa000012900000011);
	write_fxpbase(0x031d9c8, 0x2000012900000011);
	write_fxpbase(0x0312988, 0x0000000000000000);
	write_fxpbase(0x0305288, 0x000000000f44400e);
	write_fxpbase(0x0305348, 0x000000000004c008);
	write_fxpbase(0x0316988, 0x0000000000000000);
	write_fxpbase(0x030ab08, 0x000000004000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000001);
	write_fxpbase(0x031d9c8, 0x2000002b00000001);
	write_fxpbase(0x030eb08, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000001);
	write_fxpbase(0x031d9c8, 0x2000012b00000001);
	write_fxpbase(0x0312b08, 0x0000000000000000);
	write_fxpbase(0x0305608, 0x00000000000501b4);
	write_fxpbase(0x0316b08, 0x0000000000000014);
	write_fxpbase(0x030ab18, 0x000600ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000003);
	write_fxpbase(0x031d9c8, 0x2000002b00000003);
	write_fxpbase(0x030eb18, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002100000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000003);
	write_fxpbase(0x031d9c8, 0x2000012b00000003);
	write_fxpbase(0x0312b18, 0x00000000000058f0);
	write_fxpbase(0x03056d8, 0x00000000000501b4);
	write_fxpbase(0x0316b18, 0x0000000000000014);
	write_fxpbase(0x030ab20, 0x001100ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000004);
	write_fxpbase(0x031d9c8, 0x2000002b00000004);
	write_fxpbase(0x030eb20, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000004);
	write_fxpbase(0x031d9c8, 0x2000012b00000004);
	write_fxpbase(0x0312b20, 0x0000000000000000);
	write_fxpbase(0x03056e0, 0x00000000000501b4);
	write_fxpbase(0x0316b20, 0x0000000000000014);
	write_fxpbase(0x030ab28, 0x008400ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000005);
	write_fxpbase(0x031d9c8, 0x2000002b00000005);
	write_fxpbase(0x030eb28, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002300000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000005);
	write_fxpbase(0x031d9c8, 0x2000012b00000005);
	write_fxpbase(0x0312b28, 0x0000000000000000);
	write_fxpbase(0x03056e8, 0x00000000000501b4);
	write_fxpbase(0x0316b28, 0x0000000000000014);
	write_fxpbase(0x030ab30, 0x000000004000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e00000);
	write_fxpbase(0x031d9c8, 0xe000002b00000006);
	write_fxpbase(0x031d9c8, 0x2000002b00000006);
	write_fxpbase(0x030eb30, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000006);
	write_fxpbase(0x031d9c8, 0x2000012b00000006);
	write_fxpbase(0x0312b30, 0x0000000000000000);
	write_fxpbase(0x03056f0, 0x00000000000501b4);
	write_fxpbase(0x0316b30, 0x0000000000000014);
	write_fxpbase(0x030ab40, 0x000600ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e00000);
	write_fxpbase(0x031d9c8, 0xe000002b00000008);
	write_fxpbase(0x031d9c8, 0x2000002b00000008);
	write_fxpbase(0x030eb40, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002100000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000008);
	write_fxpbase(0x031d9c8, 0x2000012b00000008);
	write_fxpbase(0x0312b40, 0x00000000000058f0);
	write_fxpbase(0x0305700, 0x00000000000501b4);
	write_fxpbase(0x0316b40, 0x0000000000000014);
	write_fxpbase(0x030ab48, 0x001100ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e00000);
	write_fxpbase(0x031d9c8, 0xe000002b00000009);
	write_fxpbase(0x031d9c8, 0x2000002b00000009);
	write_fxpbase(0x030eb48, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000009);
	write_fxpbase(0x031d9c8, 0x2000012b00000009);
	write_fxpbase(0x0312b48, 0x0000000000000000);
	write_fxpbase(0x0305708, 0x00000000000501b4);
	write_fxpbase(0x0316b48, 0x0000000000000014);
	write_fxpbase(0x030ab50, 0x008400ff4000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001e00000);
	write_fxpbase(0x031d9c8, 0xe000002b0000000a);
	write_fxpbase(0x031d9c8, 0x2000002b0000000a);
	write_fxpbase(0x030eb50, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002300000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b0000000a);
	write_fxpbase(0x031d9c8, 0x2000012b0000000a);
	write_fxpbase(0x0312b50, 0x0000000000000000);
	write_fxpbase(0x0305650, 0x00000000000501b4);
	write_fxpbase(0x0316b50, 0x0000000000000014);
	write_fxpbase(0x030ab68, 0x000000006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f40000);
	write_fxpbase(0x031d9c8, 0xa000002b0000000d);
	write_fxpbase(0x031d9c8, 0x2000002b0000000d);
	write_fxpbase(0x030eb68, 0x0000002800040002);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b0000000d);
	write_fxpbase(0x031d9c8, 0x2000012b0000000d);
	write_fxpbase(0x0312b68, 0x0000000000000028);
	write_fxpbase(0x0305728, 0x00000000000500d4);
	write_fxpbase(0x0316b68, 0x0000000000000028);
	write_fxpbase(0x030ab78, 0x0600ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f40000);
	write_fxpbase(0x031d9c8, 0xa000002b0000000f);
	write_fxpbase(0x031d9c8, 0x2000002b0000000f);
	write_fxpbase(0x030eb78, 0x0000000800130007);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002180000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b0000000f);
	write_fxpbase(0x031d9c8, 0x2000012b0000000f);
	write_fxpbase(0x0312b78, 0x0000000000000028);
	write_fxpbase(0x0305738, 0x00000000000500d4);
	write_fxpbase(0x0316b78, 0x0000000000000028);
	write_fxpbase(0x030ab80, 0x1100ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f40000);
	write_fxpbase(0x031d9c8, 0xe000002b00000010);
	write_fxpbase(0x031d9c8, 0x2000002b00000010);
	write_fxpbase(0x030eb80, 0x000000190015001a);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000010);
	write_fxpbase(0x031d9c8, 0x2000012b00000010);
	write_fxpbase(0x0312b80, 0x0000000000000028);
	write_fxpbase(0x0305740, 0x00000000000500d4);
	write_fxpbase(0x0316b80, 0x0000000000000028);
	write_fxpbase(0x030ab88, 0x8400ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000011);
	write_fxpbase(0x031d9c8, 0x2000002b00000011);
	write_fxpbase(0x030eb88, 0x0000000600040005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002300000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000011);
	write_fxpbase(0x031d9c8, 0x2000012b00000011);
	write_fxpbase(0x0312b88, 0x0000000000000028);
	write_fxpbase(0x0305748, 0x00000000000500d4);
	write_fxpbase(0x0316b88, 0x0000000000000028);
	write_fxpbase(0x030ab90, 0x3b00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f40000);
	write_fxpbase(0x031d9c8, 0xa000002b00000012);
	write_fxpbase(0x031d9c8, 0x2000002b00000012);
	write_fxpbase(0x030eb90, 0x0000001900090019);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000012);
	write_fxpbase(0x031d9c8, 0x2000012b00000012);
	write_fxpbase(0x0312b90, 0x0000000000000028);
	write_fxpbase(0x0305750, 0x00000000000500d4);
	write_fxpbase(0x0316b90, 0x0000000000000028);
	write_fxpbase(0x030ab98, 0x000000006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f00000);
	write_fxpbase(0x031d9c8, 0xe000002b00000013);
	write_fxpbase(0x031d9c8, 0x2000002b00000013);
	write_fxpbase(0x030eb98, 0x0000002800230005);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000013);
	write_fxpbase(0x031d9c8, 0x2000012b00000013);
	write_fxpbase(0x0312b98, 0x0000000000000028);
	write_fxpbase(0x0305698, 0x00000000000500d4);
	write_fxpbase(0x0316b98, 0x0000000000000028);
	write_fxpbase(0x030aba0, 0x0600ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f00000);
	write_fxpbase(0x031d9c8, 0xa000002b00000014);
	write_fxpbase(0x031d9c8, 0x2000002b00000014);
	write_fxpbase(0x030eba0, 0x0000000900150026);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002180000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000014);
	write_fxpbase(0x031d9c8, 0x2000012b00000014);
	write_fxpbase(0x0312ba0, 0x0000000000000028);
	write_fxpbase(0x03056a0, 0x00000000000500d4);
	write_fxpbase(0x0316ba0, 0x0000000000000028);
	write_fxpbase(0x030aba8, 0x1100ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f00000);
	write_fxpbase(0x031d9c8, 0xa000002b00000015);
	write_fxpbase(0x031d9c8, 0x2000002b00000015);
	write_fxpbase(0x030eba8, 0x0000000400100028);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002200000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012b00000015);
	write_fxpbase(0x031d9c8, 0x2000012b00000015);
	write_fxpbase(0x0312ba8, 0x0000000000000028);
	write_fxpbase(0x0305768, 0x00000000000500d4);
	write_fxpbase(0x0316ba8, 0x0000000000000028);
	write_fxpbase(0x030abb0, 0x8400ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f00000);
	write_fxpbase(0x031d9c8, 0xe000002b00000016);
	write_fxpbase(0x031d9c8, 0x2000002b00000016);
	write_fxpbase(0x030ebb0, 0x000000020026001f);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002300000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000016);
	write_fxpbase(0x031d9c8, 0x2000012b00000016);
	write_fxpbase(0x0312bb0, 0x0000000000000028);
	write_fxpbase(0x0305770, 0x00000000000500d4);
	write_fxpbase(0x0316bb0, 0x0000000000000028);
	write_fxpbase(0x030abb8, 0x3b00ff006000f000);
	write_fxpbase(0x031d980, 0x0000000003fc0000);
	write_fxpbase(0x031d988, 0x0000000001f00000);
	write_fxpbase(0x031d9c8, 0xa000002b00000017);
	write_fxpbase(0x031d9c8, 0x2000002b00000017);
	write_fxpbase(0x030ebb8, 0x00000018000b0027);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012b00000017);
	write_fxpbase(0x031d9c8, 0x2000012b00000017);
	write_fxpbase(0x0312bb8, 0x0000000000000028);
	write_fxpbase(0x03056b8, 0x00000000000500d4);
	write_fxpbase(0x0316bb8, 0x0000000000000028);
	write_fxpbase(0x030aca0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002180000);
	write_fxpbase(0x031d9c8, 0xa000002c00000014);
	write_fxpbase(0x031d9c8, 0x2000002c00000014);
	write_fxpbase(0x030eca0, 0x000000000000000b);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000002100000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012c00000014);
	write_fxpbase(0x031d9c8, 0x2000012c00000014);
	write_fxpbase(0x0312ca0, 0x00000000000058f0);
	write_fxpbase(0x0316ca0, 0x0000000000000000);
	write_fxpbase(0x030ae70, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002100000);
	write_fxpbase(0x031d9c8, 0xe000002e0000000e);
	write_fxpbase(0x031d9c8, 0x2000002e0000000e);
	write_fxpbase(0x030ee70, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000400);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000400);
	write_fxpbase(0x031d9c8, 0xa000012e0000000e);
	write_fxpbase(0x031d9c8, 0x2000012e0000000e);
	write_fxpbase(0x0312e70, 0x0000000000000000);
	write_fxpbase(0x0305d30, 0x000000000fd9400f);
	write_fxpbase(0x0316e70, 0x0000000000000004);
	write_fxpbase(0x030aea0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002200000);
	write_fxpbase(0x031d9c8, 0xa000002e00000014);
	write_fxpbase(0x031d9c8, 0x2000002e00000014);
	write_fxpbase(0x030eea0, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000200);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000200);
	write_fxpbase(0x031d9c8, 0xe000012e00000014);
	write_fxpbase(0x031d9c8, 0x2000012e00000014);
	write_fxpbase(0x0312ea0, 0x0000000000000008);
	write_fxpbase(0x0305ca0, 0x000000000fc5c00f);
	write_fxpbase(0x0316ea0, 0x0000000000000004);
	write_fxpbase(0x030aeb8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000002300000);
	write_fxpbase(0x031d9c8, 0xe000002e00000017);
	write_fxpbase(0x031d9c8, 0x2000002e00000017);
	write_fxpbase(0x030eeb8, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000000a80000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xe000012e00000017);
	write_fxpbase(0x031d9c8, 0x2000012e00000017);
	write_fxpbase(0x0312eb8, 0x000000000000000c);
	write_fxpbase(0x0305d78, 0x000000000fc8800f);
	write_fxpbase(0x0316eb8, 0x0000000000000004);
	write_fxpbase(0x030af88, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000a80000);
	write_fxpbase(0x031d9c8, 0xe000002f00000011);
	write_fxpbase(0x031d9c8, 0x2000002f00000011);
	write_fxpbase(0x030ef88, 0x0000000000000000);
	write_fxpbase(0x031d980, 0x0000000003f80000);
	write_fxpbase(0x031d988, 0x0000000000000000);
	write_fxpbase(0x031d990, 0x0000000003e00000);
	write_fxpbase(0x031d998, 0x0000000000000000);
	write_fxpbase(0x031d9c8, 0xa000012f00000011);
	write_fxpbase(0x031d9c8, 0x2000012f00000011);
	write_fxpbase(0x0312f88, 0x0000000000000000);
	write_fxpbase(0x0305f48, 0x0000000000004015);
	write_fxpbase(0x0316f88, 0x0000000000010001);
	write_fxpbase(0x031d800, 0x000100000000000c);
	write_fxpbase(0x020f088, 0x0000000000020310);
	write_fxpbase(0x0208000, 0xd24ba8c4affb6a5c);
	write_fxpbase(0x0208008, 0xec1a147a919f3203);
	write_fxpbase(0x0208010, 0x06b3e4cea0228a46);
	write_fxpbase(0x0208018, 0x305b6fbecbad1268);
	write_fxpbase(0x0208020, 0xe8463152a79dce92);
	write_fxpbase(0x0208028, 0xd9e5d79ba1e9ffaa);
	write_fxpbase(0x0208030, 0xada800c5e8d3b720);
	write_fxpbase(0x0208038, 0xbbf4e1814e945083);
	write_fxpbase(0x0208040, 0x98217c3502110d29);
	write_fxpbase(0x0208080, 0xd7a8e866fc4328a5);
	write_fxpbase(0x0208088, 0x3cdb777d8978e7c4);
	write_fxpbase(0x0208090, 0x5bab94b12db191d1);
	write_fxpbase(0x0208098, 0xdda9ecd968a910ca);
	write_fxpbase(0x02080a0, 0xa1a83040b3161831);
	write_fxpbase(0x02080a8, 0x250c36c0f19fb9e6);
	write_fxpbase(0x02080b0, 0x866adb951b9d4886);
	write_fxpbase(0x02080b8, 0x529718547685411e);
	write_fxpbase(0x02080c0, 0x740b5f1356beefc0);
	write_fxpbase(0x0208100, 0x11c82b4b01404a63);
	write_fxpbase(0x0208108, 0xfa73bd992e8f57a3);
	write_fxpbase(0x0208110, 0x56a8afa81474d8de);
	write_fxpbase(0x0208118, 0x18f98601c86bd0eb);
	write_fxpbase(0x0208120, 0xbd4a00d490d4e8a3);
	write_fxpbase(0x0208128, 0x57f1dfb9e33ce7ab);
	write_fxpbase(0x0208130, 0xeabb06c583426a30);
	write_fxpbase(0x0208138, 0xa480870e5f8ebc84);
	write_fxpbase(0x0208140, 0x395238f64c4c011a);
	write_fxpbase(0x0208180, 0xb82690e9cf6ac119);
	write_fxpbase(0x0208188, 0xf728608f94642d0a);
	write_fxpbase(0x0208190, 0x4a7a27f20723bc05);
	write_fxpbase(0x0208198, 0x2116713cef984b98);
	write_fxpbase(0x02081a0, 0x18a7ad31353d6bbd);
	write_fxpbase(0x02081a8, 0x09bc26435ffcd2ef);
	write_fxpbase(0x02081b0, 0x7a5d40f7e0cc911f);
	write_fxpbase(0x02081b8, 0xce7f1209c2432093);
	write_fxpbase(0x02081c0, 0x2f8e2819636554d5);
	write_fxpbase(0x0208200, 0x95e75b8d916550e7);
	write_fxpbase(0x0208208, 0xaa9a270015c7cc61);
	write_fxpbase(0x0208210, 0x8f43453b9a248438);
	write_fxpbase(0x0208218, 0x36f85add12e367de);
	write_fxpbase(0x0208220, 0xa2673e3d09396c14);
	write_fxpbase(0x0208228, 0xab3da0c65005194e);
	write_fxpbase(0x0208230, 0xb0d64d9664b7367c);
	write_fxpbase(0x0208238, 0x33c4a893245fb757);
	write_fxpbase(0x0208240, 0x7ec2aa354e6325e2);
	write_fxpbase(0x0208280, 0x053fbc6488c37f86);
	write_fxpbase(0x0208288, 0xcc71ec216a55bdc9);
	write_fxpbase(0x0208290, 0xdf1602daf29ed2d7);
	write_fxpbase(0x0208298, 0xa7323116848479d7);
	write_fxpbase(0x02082a0, 0x0f396b00a2d51718);
	write_fxpbase(0x02082a8, 0xd88307753c731751);
	write_fxpbase(0x02082b0, 0x2f4ff6faa7ca9f77);
	write_fxpbase(0x02082b8, 0x7f7c7ee315f07843);
	write_fxpbase(0x02082c0, 0x476c916fb3a06373);
	write_fxpbase(0x0200170, 0x0000000000000000);
	write_fxpbase(0x0200080, 0xa000060000000000);
	write_fxpbase(0x0200170, 0x0000000000000001);
	write_fxpbase(0x0200080, 0xa000060000000001);
	write_fxpbase(0x0200170, 0x0000000000000002);
	write_fxpbase(0x0200080, 0xa000060000000002);
	write_fxpbase(0x0200170, 0x0000000000000003);
	write_fxpbase(0x0200080, 0xa000060000000003);
	write_fxpbase(0x0200170, 0x0000000000000004);
	write_fxpbase(0x0200080, 0xa000060000000004);
	write_fxpbase(0x0200170, 0x0000000000000005);
	write_fxpbase(0x0200080, 0xa000060000000005);
	write_fxpbase(0x0200170, 0x0000000000000006);
	write_fxpbase(0x0200080, 0xa000060000000006);
	write_fxpbase(0x0200170, 0x0000000000000007);
	write_fxpbase(0x0200080, 0xa000060000000007);
	write_fxpbase(0x0200170, 0x0000000000000008);
	write_fxpbase(0x0200080, 0xa000060000000008);
	write_fxpbase(0x0200170, 0x0000000000000009);
	write_fxpbase(0x0200080, 0xa000060000000009);
	write_fxpbase(0x0200170, 0x000000000000000a);
	write_fxpbase(0x0200080, 0xa00006000000000a);
	write_fxpbase(0x0200170, 0x000000000000000b);
	write_fxpbase(0x0200080, 0xa00006000000000b);
	write_fxpbase(0x0200170, 0x000000000000000c);
	write_fxpbase(0x0200080, 0xa00006000000000c);
	write_fxpbase(0x0200170, 0x000000000000000d);
	write_fxpbase(0x0200080, 0xa00006000000000d);
	write_fxpbase(0x0200170, 0x000000000000000e);
	write_fxpbase(0x0200080, 0xa00006000000000e);
	write_fxpbase(0x0200170, 0x000000000000000f);
	write_fxpbase(0x0200080, 0xa00006000000000f);
	write_fxpbase(0x0200170, 0x0000000000000010);
	write_fxpbase(0x0200080, 0xa000060000000010);
	write_fxpbase(0x0200170, 0x0000000000000011);
	write_fxpbase(0x0200080, 0xa000060000000011);
	write_fxpbase(0x0200170, 0x0000000000000012);
	write_fxpbase(0x0200080, 0xa000060000000012);
	write_fxpbase(0x0200170, 0x0000000000000013);
	write_fxpbase(0x0200080, 0xa000060000000013);
	write_fxpbase(0x0200170, 0x0000000000000014);
	write_fxpbase(0x0200080, 0xa000060000000014);
	write_fxpbase(0x0200170, 0x0000000000000015);
	write_fxpbase(0x0200080, 0xa000060000000015);
	write_fxpbase(0x0200170, 0x0000000000000016);
	write_fxpbase(0x0200080, 0xa000060000000016);
	write_fxpbase(0x0200170, 0x0000000000000017);
	write_fxpbase(0x0200080, 0xa000060000000017);
	write_fxpbase(0x0200170, 0x0000000000000018);
	write_fxpbase(0x0200080, 0xa000060000000018);
	write_fxpbase(0x0200170, 0x0000000000000019);
	write_fxpbase(0x0200080, 0xa000060000000019);
	write_fxpbase(0x0200170, 0x000000000000001a);
	write_fxpbase(0x0200080, 0xa00006000000001a);
	write_fxpbase(0x0200170, 0x000000000000001b);
	write_fxpbase(0x0200080, 0xa00006000000001b);
	write_fxpbase(0x0200170, 0x000000000000001c);
	write_fxpbase(0x0200080, 0xa00006000000001c);
	write_fxpbase(0x0200170, 0x000000000000001d);
	write_fxpbase(0x0200080, 0xa00006000000001d);
	write_fxpbase(0x0200170, 0x000000000000001e);
	write_fxpbase(0x0200080, 0xa00006000000001e);
	write_fxpbase(0x0200170, 0x000000000000001f);
	write_fxpbase(0x0200080, 0xa00006000000001f);
	write_fxpbase(0x0200170, 0x0000000000000020);
	write_fxpbase(0x0200080, 0xa000060000000020);
	write_fxpbase(0x0200170, 0x0000000000000021);
	write_fxpbase(0x0200080, 0xa000060000000021);
	write_fxpbase(0x0200170, 0x0000000000000022);
	write_fxpbase(0x0200080, 0xa000060000000022);
	write_fxpbase(0x0200170, 0x0000000000000023);
	write_fxpbase(0x0200080, 0xa000060000000023);
	write_fxpbase(0x0200170, 0x0000000000000024);
	write_fxpbase(0x0200080, 0xa000060000000024);
	write_fxpbase(0x0200170, 0x0000000000000025);
	write_fxpbase(0x0200080, 0xa000060000000025);
	write_fxpbase(0x0200170, 0x0000000000000026);
	write_fxpbase(0x0200080, 0xa000060000000026);
	write_fxpbase(0x0200170, 0x0000000000000027);
	write_fxpbase(0x0200080, 0xa000060000000027);
	write_fxpbase(0x0200170, 0x0000000000000028);
	write_fxpbase(0x0200080, 0xa000060000000028);
	write_fxpbase(0x0200170, 0x0000000000000029);
	write_fxpbase(0x0200080, 0xa000060000000029);
	write_fxpbase(0x0200170, 0x000000000000002a);
	write_fxpbase(0x0200080, 0xa00006000000002a);
	write_fxpbase(0x0200170, 0x000000000000002b);
	write_fxpbase(0x0200080, 0xa00006000000002b);
	write_fxpbase(0x0200170, 0x000000000000002c);
	write_fxpbase(0x0200080, 0xa00006000000002c);
	write_fxpbase(0x0200170, 0x000000000000002d);
	write_fxpbase(0x0200080, 0xa00006000000002d);
	write_fxpbase(0x0200170, 0x000000000000002e);
	write_fxpbase(0x0200080, 0xa00006000000002e);
	write_fxpbase(0x0200170, 0x000000000000002f);
	write_fxpbase(0x0200080, 0xa00006000000002f);
	write_fxpbase(0x0200170, 0x0000000000000030);
	write_fxpbase(0x0200080, 0xa000060000000030);
	write_fxpbase(0x0200170, 0x0000000000000031);
	write_fxpbase(0x0200080, 0xa000060000000031);
	write_fxpbase(0x0200170, 0x0000000000000032);
	write_fxpbase(0x0200080, 0xa000060000000032);
	write_fxpbase(0x0200170, 0x0000000000000033);
	write_fxpbase(0x0200080, 0xa000060000000033);
	write_fxpbase(0x0200170, 0x0000000000000034);
	write_fxpbase(0x0200080, 0xa000060000000034);
	write_fxpbase(0x0200170, 0x0000000000000035);
	write_fxpbase(0x0200080, 0xa000060000000035);
	write_fxpbase(0x0200170, 0x0000000000000036);
	write_fxpbase(0x0200080, 0xa000060000000036);
	write_fxpbase(0x0200170, 0x0000000000000037);
	write_fxpbase(0x0200080, 0xa000060000000037);
	write_fxpbase(0x0200170, 0x0000000000000038);
	write_fxpbase(0x0200080, 0xa000060000000038);
	write_fxpbase(0x0200170, 0x0000000000000039);
	write_fxpbase(0x0200080, 0xa000060000000039);
	write_fxpbase(0x0200170, 0x000000000000003a);
	write_fxpbase(0x0200080, 0xa00006000000003a);
	write_fxpbase(0x0200170, 0x000000000000003b);
	write_fxpbase(0x0200080, 0xa00006000000003b);
	write_fxpbase(0x0200170, 0x000000000000003c);
	write_fxpbase(0x0200080, 0xa00006000000003c);
	write_fxpbase(0x0200170, 0x000000000000003d);
	write_fxpbase(0x0200080, 0xa00006000000003d);
	write_fxpbase(0x0200170, 0x000000000000003e);
	write_fxpbase(0x0200080, 0xa00006000000003e);
	write_fxpbase(0x0200170, 0x000000000000003f);
	write_fxpbase(0x0200080, 0xa00006000000003f);
	write_fxpbase(0x0200170, 0x0000000000000040);
	write_fxpbase(0x0200080, 0xa000060000000040);
	write_fxpbase(0x0200170, 0x0000000000000041);
	write_fxpbase(0x0200080, 0xa000060000000041);
	write_fxpbase(0x0200170, 0x0000000000000042);
	write_fxpbase(0x0200080, 0xa000060000000042);
	write_fxpbase(0x0200170, 0x0000000000000043);
	write_fxpbase(0x0200080, 0xa000060000000043);
	write_fxpbase(0x0200170, 0x0000000000000044);
	write_fxpbase(0x0200080, 0xa000060000000044);
	write_fxpbase(0x0200170, 0x0000000000000045);
	write_fxpbase(0x0200080, 0xa000060000000045);
	write_fxpbase(0x0200170, 0x0000000000000046);
	write_fxpbase(0x0200080, 0xa000060000000046);
	write_fxpbase(0x0200170, 0x0000000000000047);
	write_fxpbase(0x0200080, 0xa000060000000047);
	write_fxpbase(0x0200170, 0x0000000000000048);
	write_fxpbase(0x0200080, 0xa000060000000048);
	write_fxpbase(0x0200170, 0x0000000000000049);
	write_fxpbase(0x0200080, 0xa000060000000049);
	write_fxpbase(0x0200170, 0x000000000000004a);
	write_fxpbase(0x0200080, 0xa00006000000004a);
	write_fxpbase(0x0200170, 0x000000000000004b);
	write_fxpbase(0x0200080, 0xa00006000000004b);
	write_fxpbase(0x0200170, 0x000000000000004c);
	write_fxpbase(0x0200080, 0xa00006000000004c);
	write_fxpbase(0x0200170, 0x000000000000004d);
	write_fxpbase(0x0200080, 0xa00006000000004d);
	write_fxpbase(0x0200170, 0x000000000000004e);
	write_fxpbase(0x0200080, 0xa00006000000004e);
	write_fxpbase(0x0200170, 0x000000000000004f);
	write_fxpbase(0x0200080, 0xa00006000000004f);
	write_fxpbase(0x0200170, 0x0000000000000050);
	write_fxpbase(0x0200080, 0xa000060000000050);
	write_fxpbase(0x0200170, 0x0000000000000051);
	write_fxpbase(0x0200080, 0xa000060000000051);
	write_fxpbase(0x0200170, 0x0000000000000052);
	write_fxpbase(0x0200080, 0xa000060000000052);
	write_fxpbase(0x0200170, 0x0000000000000053);
	write_fxpbase(0x0200080, 0xa000060000000053);
	write_fxpbase(0x0200170, 0x0000000000000054);
	write_fxpbase(0x0200080, 0xa000060000000054);
	write_fxpbase(0x0200170, 0x0000000000000055);
	write_fxpbase(0x0200080, 0xa000060000000055);
	write_fxpbase(0x0200170, 0x0000000000000056);
	write_fxpbase(0x0200080, 0xa000060000000056);
	write_fxpbase(0x0200170, 0x0000000000000057);
	write_fxpbase(0x0200080, 0xa000060000000057);
	write_fxpbase(0x0200170, 0x0000000000000058);
	write_fxpbase(0x0200080, 0xa000060000000058);
	write_fxpbase(0x0200170, 0x0000000000000059);
	write_fxpbase(0x0200080, 0xa000060000000059);
	write_fxpbase(0x0200170, 0x000000000000005a);
	write_fxpbase(0x0200080, 0xa00006000000005a);
	write_fxpbase(0x0200170, 0x000000000000005b);
	write_fxpbase(0x0200080, 0xa00006000000005b);
	write_fxpbase(0x0200170, 0x000000000000005c);
	write_fxpbase(0x0200080, 0xa00006000000005c);
	write_fxpbase(0x0200170, 0x000000000000005d);
	write_fxpbase(0x0200080, 0xa00006000000005d);
	write_fxpbase(0x0200170, 0x000000000000005e);
	write_fxpbase(0x0200080, 0xa00006000000005e);
	write_fxpbase(0x0200170, 0x000000000000005f);
	write_fxpbase(0x0200080, 0xa00006000000005f);
	write_fxpbase(0x0200170, 0x0000000000000060);
	write_fxpbase(0x0200080, 0xa000060000000060);
	write_fxpbase(0x0200170, 0x0000000000000061);
	write_fxpbase(0x0200080, 0xa000060000000061);
	write_fxpbase(0x0200170, 0x0000000000000062);
	write_fxpbase(0x0200080, 0xa000060000000062);
	write_fxpbase(0x0200170, 0x0000000000000063);
	write_fxpbase(0x0200080, 0xa000060000000063);
	write_fxpbase(0x0200170, 0x0000000000000064);
	write_fxpbase(0x0200080, 0xa000060000000064);
	write_fxpbase(0x0200170, 0x0000000000000065);
	write_fxpbase(0x0200080, 0xa000060000000065);
	write_fxpbase(0x0200170, 0x0000000000000066);
	write_fxpbase(0x0200080, 0xa000060000000066);
	write_fxpbase(0x0200170, 0x0000000000000067);
	write_fxpbase(0x0200080, 0xa000060000000067);
	write_fxpbase(0x0200170, 0x0000000000000068);
	write_fxpbase(0x0200080, 0xa000060000000068);
	write_fxpbase(0x0200170, 0x0000000000000069);
	write_fxpbase(0x0200080, 0xa000060000000069);
	write_fxpbase(0x0200170, 0x000000000000006a);
	write_fxpbase(0x0200080, 0xa00006000000006a);
	write_fxpbase(0x0200170, 0x000000000000006b);
	write_fxpbase(0x0200080, 0xa00006000000006b);
	write_fxpbase(0x0200170, 0x000000000000006c);
	write_fxpbase(0x0200080, 0xa00006000000006c);
	write_fxpbase(0x0200170, 0x000000000000006d);
	write_fxpbase(0x0200080, 0xa00006000000006d);
	write_fxpbase(0x0200170, 0x000000000000006e);
	write_fxpbase(0x0200080, 0xa00006000000006e);
	write_fxpbase(0x0200170, 0x000000000000006f);
	write_fxpbase(0x0200080, 0xa00006000000006f);
	write_fxpbase(0x0200170, 0x0000000000000070);
	write_fxpbase(0x0200080, 0xa000060000000070);
	write_fxpbase(0x0200170, 0x0000000000000071);
	write_fxpbase(0x0200080, 0xa000060000000071);
	write_fxpbase(0x0200170, 0x0000000000000072);
	write_fxpbase(0x0200080, 0xa000060000000072);
	write_fxpbase(0x0200170, 0x0000000000000073);
	write_fxpbase(0x0200080, 0xa000060000000073);
	write_fxpbase(0x0200170, 0x0000000000000074);
	write_fxpbase(0x0200080, 0xa000060000000074);
	write_fxpbase(0x0200170, 0x0000000000000075);
	write_fxpbase(0x0200080, 0xa000060000000075);
	write_fxpbase(0x0200170, 0x0000000000000076);
	write_fxpbase(0x0200080, 0xa000060000000076);
	write_fxpbase(0x0200170, 0x0000000000000077);
	write_fxpbase(0x0200080, 0xa000060000000077);
	write_fxpbase(0x0200170, 0x0000000000000078);
	write_fxpbase(0x0200080, 0xa000060000000078);
	write_fxpbase(0x0200170, 0x0000000000000079);
	write_fxpbase(0x0200080, 0xa000060000000079);
	write_fxpbase(0x0200170, 0x000000000000007a);
	write_fxpbase(0x0200080, 0xa00006000000007a);
	write_fxpbase(0x0200170, 0x000000000000007b);
	write_fxpbase(0x0200080, 0xa00006000000007b);
	write_fxpbase(0x0200170, 0x000000000000007c);
	write_fxpbase(0x0200080, 0xa00006000000007c);
	write_fxpbase(0x0200170, 0x000000000000007d);
	write_fxpbase(0x0200080, 0xa00006000000007d);
	write_fxpbase(0x0200170, 0x000000000000007e);
	write_fxpbase(0x0200080, 0xa00006000000007e);
	write_fxpbase(0x0200170, 0x000000000000007f);
	write_fxpbase(0x0200080, 0xa00006000000007f);
	write_fxpbase(0x0200170, 0x0000000000000080);
	write_fxpbase(0x0200080, 0xa000060000000080);
	write_fxpbase(0x0200170, 0x0000000000000081);
	write_fxpbase(0x0200080, 0xa000060000000081);
	write_fxpbase(0x0200170, 0x0000000000000082);
	write_fxpbase(0x0200080, 0xa000060000000082);
	write_fxpbase(0x0200170, 0x0000000000000083);
	write_fxpbase(0x0200080, 0xa000060000000083);
	write_fxpbase(0x0200170, 0x0000000000000084);
	write_fxpbase(0x0200080, 0xa000060000000084);
	write_fxpbase(0x0200170, 0x0000000000000085);
	write_fxpbase(0x0200080, 0xa000060000000085);
	write_fxpbase(0x0200170, 0x0000000000000086);
	write_fxpbase(0x0200080, 0xa000060000000086);
	write_fxpbase(0x0200170, 0x0000000000000087);
	write_fxpbase(0x0200080, 0xa000060000000087);
	write_fxpbase(0x0200170, 0x0000000000000088);
	write_fxpbase(0x0200080, 0xa000060000000088);
	write_fxpbase(0x0200170, 0x0000000000000089);
	write_fxpbase(0x0200080, 0xa000060000000089);
	write_fxpbase(0x0200170, 0x000000000000008a);
	write_fxpbase(0x0200080, 0xa00006000000008a);
	write_fxpbase(0x0200170, 0x000000000000008b);
	write_fxpbase(0x0200080, 0xa00006000000008b);
	write_fxpbase(0x0200170, 0x000000000000008c);
	write_fxpbase(0x0200080, 0xa00006000000008c);
	write_fxpbase(0x0200170, 0x000000000000008d);
	write_fxpbase(0x0200080, 0xa00006000000008d);
	write_fxpbase(0x0200170, 0x000000000000008e);
	write_fxpbase(0x0200080, 0xa00006000000008e);
	write_fxpbase(0x0200170, 0x000000000000008f);
	write_fxpbase(0x0200080, 0xa00006000000008f);
	write_fxpbase(0x0200170, 0x0000000000000090);
	write_fxpbase(0x0200080, 0xa000060000000090);
	write_fxpbase(0x0200170, 0x0000000000000091);
	write_fxpbase(0x0200080, 0xa000060000000091);
	write_fxpbase(0x0200170, 0x0000000000000092);
	write_fxpbase(0x0200080, 0xa000060000000092);
	write_fxpbase(0x0200170, 0x0000000000000093);
	write_fxpbase(0x0200080, 0xa000060000000093);
	write_fxpbase(0x0200170, 0x0000000000000094);
	write_fxpbase(0x0200080, 0xa000060000000094);
	write_fxpbase(0x0200170, 0x0000000000000095);
	write_fxpbase(0x0200080, 0xa000060000000095);
	write_fxpbase(0x0200170, 0x0000000000000096);
	write_fxpbase(0x0200080, 0xa000060000000096);
	write_fxpbase(0x0200170, 0x0000000000000097);
	write_fxpbase(0x0200080, 0xa000060000000097);
	write_fxpbase(0x0200170, 0x0000000000000098);
	write_fxpbase(0x0200080, 0xa000060000000098);
	write_fxpbase(0x0200170, 0x0000000000000099);
	write_fxpbase(0x0200080, 0xa000060000000099);
	write_fxpbase(0x0200170, 0x000000000000009a);
	write_fxpbase(0x0200080, 0xa00006000000009a);
	write_fxpbase(0x0200170, 0x000000000000009b);
	write_fxpbase(0x0200080, 0xa00006000000009b);
	write_fxpbase(0x0200020, 0x0300020003000200);
	write_fxpbase(0x0200098, 0xffffffffffffffff);
	write_fxpbase(0x02000a0, 0xffffffffffffffff);
	write_fxpbase(0x0200088, 0x0000000000000000);
	write_fxpbase(0x0200090, 0x0000000100000000);
	write_fxpbase(0x0200080, 0xa000000000000000);
	write_fxpbase(0x02000a8, 0x000000000000010b);
	write_fxpbase(0x02000b0, 0x00000fffffffffff);
	write_fxpbase(0x02000b8, 0x00000ffffbffffff);
	write_fxpbase(0x0200080, 0xa0000100000007ff);
	write_fxpbase(0x0200100, 0x0000020200000002);
	write_fxpbase(0x0200108, 0x000000ff00000402);
	write_fxpbase(0x0200110, 0x000000ff000000ff);
	write_fxpbase(0x0200118, 0x000000ff000000ff);
	write_fxpbase(0x0200120, 0x000000ff000000ff);
	write_fxpbase(0x0200128, 0x000000ff000000ff);
	write_fxpbase(0x0200130, 0x000000ff000000ff);
	write_fxpbase(0x0200138, 0x00000000000000ff);
	write_fxpbase(0x0200080, 0xa00003000000010b);
	write_fxpbase(0x0200140, 0xffffffffffffffff);
	write_fxpbase(0x0200148, 0xffffffffffffffff);
	write_fxpbase(0x0200150, 0xffffffffffffffff);
	write_fxpbase(0x0200158, 0x0000ffffffffffff);
	write_fxpbase(0x0200158, 0xffffffffffffffff);
	write_fxpbase(0x0200080, 0xa00004000000010b);
	write_fxpbase(0x02000c0, 0x073816b14c730000);
	write_fxpbase(0x02000c8, 0x0000000000600001);
	write_fxpbase(0x0200080, 0xa00002000000010b);
	write_fxpbase(0x02000a8, 0x00000000000003d9);
	write_fxpbase(0x02000b0, 0x00000ffffbffffff);
	write_fxpbase(0x02000b8, 0x00000fffffffffff);
	write_fxpbase(0x0200080, 0xa0000100000007fe);
	write_fxpbase(0x0200100, 0x0000020200000002);
	write_fxpbase(0x0200108, 0x000000ff00000402);
	write_fxpbase(0x0200110, 0x000000ff000000ff);
	write_fxpbase(0x0200118, 0x000000ff000000ff);
	write_fxpbase(0x0200120, 0x000000ff000000ff);
	write_fxpbase(0x0200128, 0x000000ff000000ff);
	write_fxpbase(0x0200130, 0x000000ff000000ff);
	write_fxpbase(0x0200138, 0x00000000000000ff);
	write_fxpbase(0x0200080, 0xa0000300000003d9);
	write_fxpbase(0x0200140, 0xffffffffffffffff);
	write_fxpbase(0x0200148, 0xffffffffffffffff);
	write_fxpbase(0x0200150, 0xffffffffffffffff);
	write_fxpbase(0x0200158, 0xffffffffffffffff);
	write_fxpbase(0x0200158, 0xffffffffffffffff);
	write_fxpbase(0x0200080, 0xa0000400000003d9);
	write_fxpbase(0x02000c0, 0x0198d729cd530000);
	write_fxpbase(0x02000c8, 0x0000000000600001);
	write_fxpbase(0x0200080, 0xa0000200000003d9);
	write_fxpbase(0x020c198, 0x0000400200060000);
	write_fxpbase(0x020c1a0, 0x0000000000000000);
	write_fxpbase(0x020c1b0, 0x0000000000000000);
	write_fxpbase(0x020c1a8, 0x0000000000000000);
	write_fxpbase(0x020c180, 0xa000020000000000);
	write_fxpbase(0x0200168, 0x0000000000000000);
	write_fxpbase(0x0200160, 0x0000000000000031);
	write_fxpbase(0x0200080, 0xa000050000000000);
	write_fxpbase(0x0200000, 0x0000000000000083);
	write_fxpbase(0x020c120, 0x0000000000000800);
	write_fxpbase(0x0211430, 0x5cbddd8cede2585f);
	write_fxpbase(0x0211428, 0x0000000000000003);
	write_fxpbase(0x0218020, 0x0000000000000001);
	write_fxpbase(0x0211000, 0x0000000000000008);
	write_fxpbase(0x0211008, 0x0000000000000008);
	write_fxpbase(0x0211010, 0x0000000000000008);
	write_fxpbase(0x0211018, 0x0000000000000008);
	write_fxpbase(0x0218008, 0x8000847277b1db89);
	write_fxpbase(0x0218000, 0xa000000000000000);
	write_fxpbase(0x0218000, 0x0000000000000000);
	write_fxpbase(0x0218008, 0x8000ed062dfafd6a);
	write_fxpbase(0x0218000, 0xa000000000000001);
	write_fxpbase(0x0218000, 0x0000000000000000);
	write_fxpbase(0x0218008, 0x8000b584f45bb760);
	write_fxpbase(0x0218000, 0xa000000000000002);
	write_fxpbase(0x0218000, 0x0000000000000000);
	write_fxpbase(0x0218008, 0x80007faee98c75ab);
	write_fxpbase(0x0218000, 0xa000000000000003);
	write_fxpbase(0x0218000, 0x0000000000000000);
	write_fxpbase(0x0218018, 0x8000000000000000);
	write_fxpbase(0x0218010, 0xa000000000000000);
	write_fxpbase(0x0218010, 0x0000000000000000);
	write_fxpbase(0x0218018, 0x8000000000008000);
	write_fxpbase(0x0218010, 0xa000000000000001);
	write_fxpbase(0x0218010, 0x0000000000000000);
	write_fxpbase(0x0218018, 0x8000000000010000);
	write_fxpbase(0x0218010, 0xa000000000000002);
	write_fxpbase(0x0218010, 0x0000000000000000);
	write_fxpbase(0x0218018, 0x8000000000018000);
	write_fxpbase(0x0218010, 0xa000000000000003);
	write_fxpbase(0x0218010, 0x0000000000000000);
	write_fxpbase(0x01a0000, 0x0000ffffffffffff);
	write_fxpbase(0x01a0100, 0x001f001f001f001f);
	write_fxpbase(0x01a0280, 0xffffffffffffffff);
	write_fxpbase(0x01a0288, 0x000000000000ffff);
	write_fxpbase(0x01a0290, 0xffffffffffffffff);
	write_fxpbase(0x01a0298, 0x000000000000ffff);
	write_fxpbase(0x01a02b0, 0xa000000000000000);
	write_fxpbase(0x01a0280, 0x0000300100000000);
	write_fxpbase(0x01a0288, 0x0000000000000000);
	write_fxpbase(0x01a02b0, 0xa000010000000000);
	write_fxpbase(0x01a02c0, 0x0139023700000000);
	write_fxpbase(0x01a02d0, 0x000701e500000000);
	write_fxpbase(0x0028690, 0x1084212a30001004);
	write_fxpbase(0x0028698, 0x0000000000000237);
	write_fxpbase(0x00286a0, 0x0000000000000000);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xa000000000000237);
	write_fxpbase(0x0028688, 0x2516000000000237);
	write_fxpbase(0x0028600, 0x000009ff00009aff);
	write_fxpbase(0x0028608, 0x0001b7ff000117ff);
	write_fxpbase(0x0028610, 0x00012cff000047ff);
	write_fxpbase(0x0028618, 0x0001a1ff000126ff);
	write_fxpbase(0x0028620, 0x000034ff0000a5ff);
	write_fxpbase(0x0028628, 0x000070ff0001a4ff);
	write_fxpbase(0x0028630, 0x0001d3ff0001d9ff);
	write_fxpbase(0x0028638, 0x000010ff000198ff);
	write_fxpbase(0x0028640, 0x00019aff0000a0ff);
	write_fxpbase(0x0028648, 0x0000dfff00019cff);
	write_fxpbase(0x0028650, 0x0000c9ff00010aff);
	write_fxpbase(0x0028658, 0x0001e1ff000065ff);
	write_fxpbase(0x0028660, 0x0001ddff0000c6ff);
	write_fxpbase(0x0028668, 0x0000beff000027ff);
	write_fxpbase(0x0028670, 0x000177ff000080ff);
	write_fxpbase(0x0028678, 0x0001ceff000100ff);
	write_fxpbase(0x0028680, 0xa000000000000237);
	write_fxpbase(0x0028680, 0x2da8000000000237);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xada8010000000237);
	write_fxpbase(0x0028680, 0x39f2010000000237);
	write_fxpbase(0x0028690, 0x1084212af0001004);
	write_fxpbase(0x0028698, 0x0000000000000139);
	write_fxpbase(0x00286a0, 0x0000000000000000);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xa516000000000139);
	write_fxpbase(0x0028688, 0x2933000000000139);
	write_fxpbase(0x0028600, 0x0000e4ff0000f5ff);
	write_fxpbase(0x0028608, 0x0000f7ff0001cbff);
	write_fxpbase(0x0028610, 0x0000bfff0000c0ff);
	write_fxpbase(0x0028618, 0x000141ff0001feff);
	write_fxpbase(0x0028620, 0x000155ff000104ff);
	write_fxpbase(0x0028628, 0x000116ff0001acff);
	write_fxpbase(0x0028630, 0x000105ff000139ff);
	write_fxpbase(0x0028638, 0x000037ff000099ff);
	write_fxpbase(0x0028640, 0x00010fff0000fdff);
	write_fxpbase(0x0028648, 0x000178ff00017fff);
	write_fxpbase(0x0028650, 0x0000f1ff00002eff);
	write_fxpbase(0x0028658, 0x000153ff000032ff);
	write_fxpbase(0x0028660, 0x000105ff0000b6ff);
	write_fxpbase(0x0028668, 0x0001c1ff00012aff);
	write_fxpbase(0x0028670, 0x000172ff0001c0ff);
	write_fxpbase(0x0028678, 0x0000a7ff00000dff);
	write_fxpbase(0x0028680, 0xb9f2000000000139);
	write_fxpbase(0x0028680, 0x2665000000000139);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xa665010000000139);
	write_fxpbase(0x0028680, 0x2d6a010000000139);
	write_fxpbase(0x0028690, 0x1084212a20001004);
	write_fxpbase(0x0028698, 0x00000000000001e5);
	write_fxpbase(0x00286a0, 0x0000000000000000);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xa9330000000001e5);
	write_fxpbase(0x0028688, 0x20d70000000001e5);
	write_fxpbase(0x0028600, 0x000048ff00019dff);
	write_fxpbase(0x0028608, 0x000031ff00018fff);
	write_fxpbase(0x0028610, 0x0000cfff00019cff);
	write_fxpbase(0x0028618, 0x000079ff0001faff);
	write_fxpbase(0x0028620, 0x000175ff00004cff);
	write_fxpbase(0x0028628, 0x000037ff000091ff);
	write_fxpbase(0x0028630, 0x000166ff000151ff);
	write_fxpbase(0x0028638, 0x00013fff000133ff);
	write_fxpbase(0x0028640, 0x000064ff000173ff);
	write_fxpbase(0x0028648, 0x000035ff000041ff);
	write_fxpbase(0x0028650, 0x000160ff0000a5ff);
	write_fxpbase(0x0028658, 0x000155ff0001eaff);
	write_fxpbase(0x0028660, 0x0001d6ff0001d8ff);
	write_fxpbase(0x0028668, 0x000042ff000107ff);
	write_fxpbase(0x0028670, 0x000131ff00002bff);
	write_fxpbase(0x0028678, 0x000116ff0000fdff);
	write_fxpbase(0x0028680, 0xad6a0000000001e5);
	write_fxpbase(0x0028680, 0x35020000000001e5);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xb5020100000001e5);
	write_fxpbase(0x0028680, 0x21f90100000001e5);
	write_fxpbase(0x0028690, 0x1084212a20001004);
	write_fxpbase(0x0028698, 0x0000000000000007);
	write_fxpbase(0x00286a0, 0x0000000000000000);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xa0d7000000000007);
	write_fxpbase(0x0028688, 0x3c54000000000007);
	write_fxpbase(0x0028600, 0x000006ff0000e2ff);
	write_fxpbase(0x0028608, 0x000102ff00007eff);
	write_fxpbase(0x0028610, 0x000090ff00013bff);
	write_fxpbase(0x0028618, 0x00019bff0001b4ff);
	write_fxpbase(0x0028620, 0x0000f0ff000198ff);
	write_fxpbase(0x0028628, 0x000124ff00015bff);
	write_fxpbase(0x0028630, 0x00018bff0000cfff);
	write_fxpbase(0x0028638, 0x000182ff000186ff);
	write_fxpbase(0x0028640, 0x000022ff0000a8ff);
	write_fxpbase(0x0028648, 0x000139ff0001c2ff);
	write_fxpbase(0x0028650, 0x00019fff000147ff);
	write_fxpbase(0x0028658, 0x00013bff0000dfff);
	write_fxpbase(0x0028660, 0x000120ff000018ff);
	write_fxpbase(0x0028668, 0x000196ff000193ff);
	write_fxpbase(0x0028670, 0x000101ff0001c5ff);
	write_fxpbase(0x0028678, 0x000070ff0000fdff);
	write_fxpbase(0x0028680, 0xa1f9000000000007);
	write_fxpbase(0x0028680, 0x3d81000000000007);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xbd81010000000007);
	write_fxpbase(0x0028680, 0x3f21010000000007);
	write_fxpbase(0x0028690, 0x1084212aa0001409);
	write_fxpbase(0x0028698, 0x00000000000002a7);
	write_fxpbase(0x00286a0, 0x28a51455204a8833);
	write_fxpbase(0x00286a8, 0x69000004204a0400);
	write_fxpbase(0x0028688, 0xbc540000000002a7);
	write_fxpbase(0x0028688, 0x289d0000000002a7);
	write_fxpbase(0x0028600, 0x000192ff00001fff);
	write_fxpbase(0x0028608, 0x00011fff000043ff);
	write_fxpbase(0x0028610, 0x0001f6ff0000ceff);
	write_fxpbase(0x0028618, 0x00006fff000108ff);
	write_fxpbase(0x0028620, 0x0000abff0001b9ff);
	write_fxpbase(0x0028628, 0x0001a7ff0000f4ff);
	write_fxpbase(0x0028630, 0x0001a2ff00008fff);
	write_fxpbase(0x0028638, 0x000116ff000089ff);
	write_fxpbase(0x0028640, 0x00000eff00012cff);
	write_fxpbase(0x0028648, 0x000178ff0001e1ff);
	write_fxpbase(0x0028650, 0x000114ff0000f3ff);
	write_fxpbase(0x0028658, 0x000150ff00010aff);
	write_fxpbase(0x0028660, 0x0001f2ff0000a0ff);
	write_fxpbase(0x0028668, 0x00005aff0001f9ff);
	write_fxpbase(0x0028670, 0x0001d8ff0000daff);
	write_fxpbase(0x0028678, 0x000069ff0000eaff);
	write_fxpbase(0x0028680, 0xbf210000000002a7);
	write_fxpbase(0x0028680, 0x2b4f0000000002a7);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xab4f0100000002a7);
	write_fxpbase(0x0028680, 0x3b4f0100000002a7);
	write_fxpbase(0x0028690, 0x1084212a10001409);
	write_fxpbase(0x0028698, 0x0000000000000171);
	write_fxpbase(0x00286a0, 0x8a05c040a2990808);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xa89d000000000171);
	write_fxpbase(0x0028688, 0x25df000000000171);
	write_fxpbase(0x0028600, 0x000155ff000195ff);
	write_fxpbase(0x0028608, 0x000015ff00015dff);
	write_fxpbase(0x0028610, 0x00019aff000050ff);
	write_fxpbase(0x0028618, 0x00015fff00016cff);
	write_fxpbase(0x0028620, 0x0000d4ff0000cdff);
	write_fxpbase(0x0028628, 0x000130ff0000dfff);
	write_fxpbase(0x0028630, 0x0001c3ff00010cff);
	write_fxpbase(0x0028638, 0x0000fcff000087ff);
	write_fxpbase(0x0028640, 0x0001dfff00011cff);
	write_fxpbase(0x0028648, 0x00010bff000194ff);
	write_fxpbase(0x0028650, 0x000034ff0001c2ff);
	write_fxpbase(0x0028658, 0x00017aff0001daff);
	write_fxpbase(0x0028660, 0x0001bfff000131ff);
	write_fxpbase(0x0028668, 0x000146ff000014ff);
	write_fxpbase(0x0028670, 0x0001c9ff000149ff);
	write_fxpbase(0x0028678, 0x00003fff000113ff);
	write_fxpbase(0x0028680, 0xbb4f000000000171);
	write_fxpbase(0x0028680, 0x20d0000000000171);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xa0d0010000000171);
	write_fxpbase(0x0028680, 0x204c010000000171);
	write_fxpbase(0x0028690, 0x1084212a10001409);
	write_fxpbase(0x0028698, 0x000000000000002d);
	write_fxpbase(0x00286a0, 0x28a51455204a8833);
	write_fxpbase(0x00286a8, 0x69000004204a0400);
	write_fxpbase(0x0028688, 0xa5df00000000002d);
	write_fxpbase(0x0028688, 0x2d9c00000000002d);
	write_fxpbase(0x0028600, 0x000053ff00011eff);
	write_fxpbase(0x0028608, 0x0000d7ff000188ff);
	write_fxpbase(0x0028610, 0x000089ff0001deff);
	write_fxpbase(0x0028618, 0x000034ff0000b4ff);
	write_fxpbase(0x0028620, 0x0001caff0001d6ff);
	write_fxpbase(0x0028628, 0x0001acff000199ff);
	write_fxpbase(0x0028630, 0x00007dff00012fff);
	write_fxpbase(0x0028638, 0x00007bff00015eff);
	write_fxpbase(0x0028640, 0x000054ff000184ff);
	write_fxpbase(0x0028648, 0x000109ff000093ff);
	write_fxpbase(0x0028650, 0x000082ff00014cff);
	write_fxpbase(0x0028658, 0x000081ff0000b8ff);
	write_fxpbase(0x0028660, 0x0000d0ff000107ff);
	write_fxpbase(0x0028668, 0x000055ff0000deff);
	write_fxpbase(0x0028670, 0x000081ff0001a7ff);
	write_fxpbase(0x0028678, 0x000159ff000047ff);
	write_fxpbase(0x0028680, 0xa04c00000000002d);
	write_fxpbase(0x0028680, 0x28a800000000002d);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xa8a801000000002d);
	write_fxpbase(0x0028680, 0x2d8701000000002d);
	write_fxpbase(0x0028690, 0x1084212a70001409);
	write_fxpbase(0x0028698, 0x0000000000000095);
	write_fxpbase(0x00286a0, 0x8a05c040a2990808);
	write_fxpbase(0x00286a8, 0x0000000000000000);
	write_fxpbase(0x0028688, 0xad9c000000000095);
	write_fxpbase(0x0028688, 0x2f50000000000095);
	write_fxpbase(0x0028600, 0x000088ff00000cff);
	write_fxpbase(0x0028608, 0x000064ff000061ff);
	write_fxpbase(0x0028610, 0x0000f2ff000172ff);
	write_fxpbase(0x0028618, 0x0000d1ff000156ff);
	write_fxpbase(0x0028620, 0x0001b2ff000199ff);
	write_fxpbase(0x0028628, 0x000008ff00001bff);
	write_fxpbase(0x0028630, 0x0000d5ff0000ccff);
	write_fxpbase(0x0028638, 0x0001d5ff0000dbff);
	write_fxpbase(0x0028640, 0x000127ff000197ff);
	write_fxpbase(0x0028648, 0x0000b7ff000154ff);
	write_fxpbase(0x0028650, 0x0001a2ff000118ff);
	write_fxpbase(0x0028658, 0x0000fcff0001f5ff);
	write_fxpbase(0x0028660, 0x00010eff0001e6ff);
	write_fxpbase(0x0028668, 0x000099ff00011dff);
	write_fxpbase(0x0028670, 0x000007ff0001a8ff);
	write_fxpbase(0x0028678, 0x000100ff0000a9ff);
	write_fxpbase(0x0028680, 0xad87000000000095);
	write_fxpbase(0x0028680, 0x2b01000000000095);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xab01010000000095);
	write_fxpbase(0x0028680, 0x2ec8010000000095);
	write_fxpbase(0x0028690, 0x1084212ac0001409);
	write_fxpbase(0x0028698, 0x00000000000002ba);
	write_fxpbase(0x00286a0, 0xa8a51455204a8833);
	write_fxpbase(0x00286a8, 0xa9000004204a0400);
	write_fxpbase(0x0028688, 0xaf500000000002ba);
	write_fxpbase(0x0028688, 0x23f00000000002ba);
	write_fxpbase(0x0028600, 0x000046ff0001eaff);
	write_fxpbase(0x0028608, 0x000180ff000181ff);
	write_fxpbase(0x0028610, 0x0000e1ff00001aff);
	write_fxpbase(0x0028618, 0x000104ff000057ff);
	write_fxpbase(0x0028620, 0x00000dff000196ff);
	write_fxpbase(0x0028628, 0x0001d2ff000042ff);
	write_fxpbase(0x0028630, 0x000064ff000172ff);
	write_fxpbase(0x0028638, 0x000077ff000134ff);
	write_fxpbase(0x0028640, 0x000017ff000130ff);
	write_fxpbase(0x0028648, 0x00000bff00005dff);
	write_fxpbase(0x0028650, 0x000106ff0001efff);
	write_fxpbase(0x0028658, 0x000011ff000127ff);
	write_fxpbase(0x0028660, 0x000019ff0000aaff);
	write_fxpbase(0x0028668, 0x00000cff00004eff);
	write_fxpbase(0x0028670, 0x000155ff0000a6ff);
	write_fxpbase(0x0028678, 0x0001d8ff000106ff);
	write_fxpbase(0x0028680, 0xaec80000000002ba);
	write_fxpbase(0x0028680, 0x3e920000000002ba);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xbe920100000002ba);
	write_fxpbase(0x0028680, 0x2b2b0100000002ba);
	write_fxpbase(0x0028690, 0x1084212a60001409);
	write_fxpbase(0x0028698, 0x0000000000000353);
	write_fxpbase(0x00286a0, 0xea05c08022990b0b);
	write_fxpbase(0x00286a8, 0x00000000ee0001c8);
	write_fxpbase(0x0028688, 0xa3f0000000000353);
	write_fxpbase(0x0028688, 0x3aac000000000353);
	write_fxpbase(0x0028600, 0x0001fbff000065ff);
	write_fxpbase(0x0028608, 0x000185ff00011eff);
	write_fxpbase(0x0028610, 0x0000f6ff000123ff);
	write_fxpbase(0x0028618, 0x0001e4ff000087ff);
	write_fxpbase(0x0028620, 0x0001cbff000196ff);
	write_fxpbase(0x0028628, 0x00005eff0001cfff);
	write_fxpbase(0x0028630, 0x00002dff0000daff);
	write_fxpbase(0x0028638, 0x000138ff00016dff);
	write_fxpbase(0x0028640, 0x0000b5ff000183ff);
	write_fxpbase(0x0028648, 0x000183ff0000f3ff);
	write_fxpbase(0x0028650, 0x0000fdff0001e7ff);
	write_fxpbase(0x0028658, 0x00007dff0001b9ff);
	write_fxpbase(0x0028660, 0x00017cff000021ff);
	write_fxpbase(0x0028668, 0x0000b7ff000048ff);
	write_fxpbase(0x0028670, 0x00011aff0000a6ff);
	write_fxpbase(0x0028678, 0x0001a5ff00015aff);
	write_fxpbase(0x0028680, 0xab2b000000000353);
	write_fxpbase(0x0028680, 0x344a000000000353);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xb44a010000000353);
	write_fxpbase(0x0028680, 0x323a010000000353);
	write_fxpbase(0x0028690, 0x1084212ad0001409);
	write_fxpbase(0x0028698, 0x0000000000000196);
	write_fxpbase(0x00286a0, 0xa8a51455204a8833);
	write_fxpbase(0x00286a8, 0xa9000004204a0400);
	write_fxpbase(0x0028688, 0xbaac000000000196);
	write_fxpbase(0x0028688, 0x3b04000000000196);
	write_fxpbase(0x0028600, 0x000064ff000118ff);
	write_fxpbase(0x0028608, 0x0000a9ff00016aff);
	write_fxpbase(0x0028610, 0x000137ff00011dff);
	write_fxpbase(0x0028618, 0x000126ff00008dff);
	write_fxpbase(0x0028620, 0x0000e3ff000086ff);
	write_fxpbase(0x0028628, 0x00019cff0001a7ff);
	write_fxpbase(0x0028630, 0x0001f7ff000138ff);
	write_fxpbase(0x0028638, 0x000067ff00017eff);
	write_fxpbase(0x0028640, 0x000193ff0000ebff);
	write_fxpbase(0x0028648, 0x00012dff00018dff);
	write_fxpbase(0x0028650, 0x00003fff00007cff);
	write_fxpbase(0x0028658, 0x000060ff0000e2ff);
	write_fxpbase(0x0028660, 0x00012cff000025ff);
	write_fxpbase(0x0028668, 0x00003aff00015eff);
	write_fxpbase(0x0028670, 0x000127ff0001b9ff);
	write_fxpbase(0x0028678, 0x000025ff00007eff);
	write_fxpbase(0x0028680, 0xb23a000000000196);
	write_fxpbase(0x0028680, 0x2d6e000000000196);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xad6e010000000196);
	write_fxpbase(0x0028680, 0x2acd010000000196);
	write_fxpbase(0x0028690, 0x1084212ae0001409);
	write_fxpbase(0x0028698, 0x0000000000000335);
	write_fxpbase(0x00286a0, 0xea05c08022990b0b);
	write_fxpbase(0x00286a8, 0x00000000ee0001c8);
	write_fxpbase(0x0028688, 0xbb04000000000335);
	write_fxpbase(0x0028688, 0x2910000000000335);
	write_fxpbase(0x0028600, 0x000180ff0001bcff);
	write_fxpbase(0x0028608, 0x0000b5ff00008fff);
	write_fxpbase(0x0028610, 0x00014aff0001b2ff);
	write_fxpbase(0x0028618, 0x000198ff000118ff);
	write_fxpbase(0x0028620, 0x0000cfff0000f5ff);
	write_fxpbase(0x0028628, 0x00016eff0001cbff);
	write_fxpbase(0x0028630, 0x000195ff000107ff);
	write_fxpbase(0x0028638, 0x00009eff000198ff);
	write_fxpbase(0x0028640, 0x0001cdff0001a2ff);
	write_fxpbase(0x0028648, 0x00000cff0001feff);
	write_fxpbase(0x0028650, 0x00017fff0001d6ff);
	write_fxpbase(0x0028658, 0x0000e3ff0001cbff);
	write_fxpbase(0x0028660, 0x00003aff0000e5ff);
	write_fxpbase(0x0028668, 0x0000d3ff0001c9ff);
	write_fxpbase(0x0028670, 0x0001c4ff0000faff);
	write_fxpbase(0x0028678, 0x000041ff00012eff);
	write_fxpbase(0x0028680, 0xaacd000000000335);
	write_fxpbase(0x0028680, 0x39cc000000000335);
	write_fxpbase(0x0028600, 0xffffffffffffffff);
	write_fxpbase(0x0028608, 0xffffffffffffffff);
	write_fxpbase(0x0028610, 0xffffffffffffffff);
	write_fxpbase(0x0028618, 0xffffffffffffffff);
	write_fxpbase(0x0028620, 0xffffffffffffffff);
	write_fxpbase(0x0028628, 0xffffffffffffffff);
	write_fxpbase(0x0028630, 0xffffffffffffffff);
	write_fxpbase(0x0028638, 0xffffffffffffffff);
	write_fxpbase(0x0028680, 0xb9cc010000000335);
	write_fxpbase(0x0028680, 0x38c7010000000335);
	write_fxpbase(0x0028000, 0x3ac570f7fe1817ee);
	write_fxpbase(0x0028008, 0x75779897cb24b16f);
	write_fxpbase(0x0028010, 0xa7c1afb9985abcd7);
	write_fxpbase(0x0028018, 0xdafe66d961e0854a);
	write_fxpbase(0x0028020, 0x5468f0dacb0cde04);
	write_fxpbase(0x0028028, 0x6db6663521b8f635);
	write_fxpbase(0x0028030, 0x0b513d7918f31683);
	write_fxpbase(0x0028038, 0xae513b2edac5e85b);
	write_fxpbase(0x0028040, 0x06365456ac81ae3d);
	write_fxpbase(0x0028080, 0xb2b53059f278e47b);
	write_fxpbase(0x0028088, 0xea65e9da2da6ba32);
	write_fxpbase(0x0028090, 0x3293ac25095a81f3);
	write_fxpbase(0x0028098, 0x7b92d83ccf4b3085);
	write_fxpbase(0x00280a0, 0xd6c00b23b41a57a4);
	write_fxpbase(0x00280a8, 0x5fd12cdce1187cdc);
	write_fxpbase(0x00280b0, 0x853dd3dcb87f6884);
	write_fxpbase(0x00280b8, 0x7a1ccf226d62f0f4);
	write_fxpbase(0x00280c0, 0x7bf9ddf9dd0f91eb);
	write_fxpbase(0x0028100, 0xb32fcaa695fcf7a8);
	write_fxpbase(0x0028108, 0xe96a30b382a68cbb);
	write_fxpbase(0x0028110, 0x55107ee584e5b45b);
	write_fxpbase(0x0028118, 0x1af072b395b4a06d);
	write_fxpbase(0x0028120, 0x1777621068709988);
	write_fxpbase(0x0028128, 0xc9ee01d4717ff9e6);
	write_fxpbase(0x0028130, 0x81b06234b62e7422);
	write_fxpbase(0x0028138, 0xf62f1fcfba25512e);
	write_fxpbase(0x0028140, 0x5f8b60d23e4e80a3);
	write_fxpbase(0x0028180, 0x34b24d16400bfbcf);
	write_fxpbase(0x0028188, 0x5b43d4d21b0c35b7);
	write_fxpbase(0x0028190, 0x31bd25d510b29428);
	write_fxpbase(0x0028198, 0xcdf73d0b9067cf5b);
	write_fxpbase(0x00281a0, 0xe01fd1b0465ad509);
	write_fxpbase(0x00281a8, 0xa8344cd7e5ec96e8);
	write_fxpbase(0x00281b0, 0x45dc03a7076691aa);
	write_fxpbase(0x00281b8, 0xbb38d7042ab8173b);
	write_fxpbase(0x00281c0, 0x4a7e27cc884d045e);
	write_fxpbase(0x0028200, 0xf072979e991c6a2e);
	write_fxpbase(0x0028208, 0x23f724a491803b22);
	write_fxpbase(0x0028210, 0xa5357763fd5726b4);
	write_fxpbase(0x0028218, 0x48b131628e7223da);
	write_fxpbase(0x0028220, 0x65218402290d7f60);
	write_fxpbase(0x0028228, 0xbbaaf204616b8e9c);
	write_fxpbase(0x0028230, 0x61bf410db2ffbe4e);
	write_fxpbase(0x0028238, 0xe18eb4ab5877d945);
	write_fxpbase(0x0028240, 0xab266ed18d162ef1);
	write_fxpbase(0x0028280, 0xdf7586df0cc78279);
	write_fxpbase(0x0028288, 0x2e1b724935dba69d);
	write_fxpbase(0x0028290, 0xd879771e7a95ce20);
	write_fxpbase(0x0028298, 0xead333db398c4d57);
	write_fxpbase(0x00282a0, 0x2552d6cc0c6e4eff);
	write_fxpbase(0x00282a8, 0x59c45ed9c256b877);
	write_fxpbase(0x00282b0, 0x9636df9376be62bb);
	write_fxpbase(0x00282b8, 0xb1b625dbe0e9d4b0);
	write_fxpbase(0x00282c0, 0xa78a5ca776aea8c3);
	write_fxpbase(0x0028400, 0x000000000000008d);
	write_fxpbase(0x0028408, 0x000000000000008d);
	write_fxpbase(0x0028410, 0x0000000000000039);
	write_fxpbase(0x0028418, 0x0000000000000029);
	write_fxpbase(0x0028420, 0x0000000000000029);
	write_fxpbase(0x0028428, 0x0000000000000055);
	write_fxpbase(0x0028430, 0x0000000000000031);
	write_fxpbase(0x0028438, 0x0000000000000042);
	write_fxpbase(0x0028440, 0x0000000000000085);
	write_fxpbase(0x0028448, 0x000000000000004d);
	write_fxpbase(0x0028450, 0x0000000000000089);
	write_fxpbase(0x0028458, 0x0000000000000025);
	write_fxpbase(0x0028460, 0x0000000000000045);
	write_fxpbase(0x0028468, 0x0000000000000085);
	write_fxpbase(0x0028470, 0x0000000000000041);
	write_fxpbase(0x0028478, 0x0000000000000040);
	write_fxpbase(0x0028480, 0x000000000000004d);
	write_fxpbase(0x0028488, 0x0000000000000045);
	write_fxpbase(0x0028490, 0x0000000000000051);
	write_fxpbase(0x0028498, 0x0000000000000085);
	write_fxpbase(0x00284a0, 0x0000000000000039);
	write_fxpbase(0x00284a8, 0x0000000000000021);
	write_fxpbase(0x00284b0, 0x0000000000000041);
	write_fxpbase(0x00284b8, 0x0000000000000040);
	write_fxpbase(0x00284c0, 0x0000000000000039);
	write_fxpbase(0x00284c8, 0x0000000000000021);
	write_fxpbase(0x00284d0, 0x0000000000000041);
	write_fxpbase(0x00284d8, 0x0000000000000081);
	write_fxpbase(0x00284e0, 0x0000000000000021);
	write_fxpbase(0x00284e8, 0x0000000000000049);
	write_fxpbase(0x00284f0, 0x000000000000004d);
	write_fxpbase(0x00284f8, 0x0000000000000042);
	write_fxpbase(0x0028000, 0x3ac570f7fe1817ee);
	write_fxpbase(0x0028400, 0x000000000000008d);
	write_fxpbase(0x0000000, 0x0000000000000000);
	write_fxpbase(0x0000008, 0x0000000000000000);
	write_fxpbase(0x0000010, 0x0000000000000000);
	write_fxpbase(0x0000018, 0x0000000000000000);
	write_fxpbase(0x0000020, 0x0000000000000000);
	write_fxpbase(0x0000028, 0x0000000000000000);
	write_fxpbase(0x0000030, 0x0000000000000000);
	write_fxpbase(0x0000038, 0x0000000000000000);
	write_fxpbase(0x0000040, 0x0000000000000000);
	write_fxpbase(0x0000048, 0x0000000000000000);
	write_fxpbase(0x0041430, 0x4253f1360f8ab907);
	write_fxpbase(0x0041428, 0x0000000000000001);
	write_fxpbase(0x0048020, 0x0000000000000001);
	write_fxpbase(0x0041000, 0x0000000000000008);
	write_fxpbase(0x0041008, 0x0000000000000008);
	write_fxpbase(0x0041010, 0x0000000000000008);
	write_fxpbase(0x0041018, 0x0000000000000008);
	write_fxpbase(0x0041020, 0x0000000000000008);
	write_fxpbase(0x0041028, 0x0000000000000008);
	write_fxpbase(0x0041030, 0x0000000000000008);
	write_fxpbase(0x0048008, 0x8000da6062836c4f);
	write_fxpbase(0x0048000, 0xa000000000000000);
	write_fxpbase(0x0048000, 0x0000000000000000);
	write_fxpbase(0x0048018, 0x8000000000000000);
	write_fxpbase(0x0048010, 0xa000000000000000);
	write_fxpbase(0x0048010, 0x0000000000000000);
	write_fxpbase(0x0586e40, 0x3fdfeff7fbfdfe30);
	write_fxpbase(0x054e400, 0x0000000000840000);
	write_fxpbase(0x054e408, 0x0000000000000000);
	write_fxpbase(0x054e410, 0x0000000000000000);
	write_fxpbase(0x054e418, 0x0000000000000000);
	write_fxpbase(0x054e420, 0x0000000000000000);
	write_fxpbase(0x054e428, 0x0000000000000000);
	write_fxpbase(0x054e430, 0x0000000000000000);
	write_fxpbase(0x054e438, 0x0000000000000000);
	write_fxpbase(0x054e440, 0x0000000000000000);
	write_fxpbase(0x054e448, 0x0000000000000000);
	write_fxpbase(0x054e450, 0x0000000000000000);
	write_fxpbase(0x054e458, 0x0000000000000000);
	write_fxpbase(0x054e460, 0x0000000000000000);
	write_fxpbase(0x054e468, 0x0000000000000000);
	write_fxpbase(0x054e470, 0x0000000000000000);
	write_fxpbase(0x054e478, 0x0000000000000000);
	write_fxpbase(0x058c040, 0x0000000000400243);
	write_fxpbase(0x058ba00, 0x0000148c0c460723);
	write_fxpbase(0x058c058, 0x0000000000400243);
	write_fxpbase(0x058bac0, 0x000014bc0c5e072f);
	write_fxpbase(0x0362810, 0x000000000000001e);
	write_fxpbase(0x0362800, 0x0000000000000015);
	write_fxpbase(0x0362808, 0x000000000000000e);
	write_fxpbase(0x0362800, 0x0000000000000015);
	write_fxpbase(0x0827500, 0x0000000f00000000);
	write_fxpbase(0x0827508, 0x0000000f00000000);
	write_fxpbase(0x0827510, 0x0000000f00000000);
	write_fxpbase(0x0827518, 0x0000000f00000000);
	write_fxpbase(0x0827520, 0x0000000f00000000);
	write_fxpbase(0x0827528, 0x0000000f00000000);
	write_fxpbase(0x0827530, 0x0000000f00000000);
	write_fxpbase(0x0827538, 0x0000000f00000000);
	/*  Registers changed from DV Default  */
	/* FXP_MD_TX_AUTO_DEL */
	write_fxpbase(0x3e8100, 0x0fffffffffffffef);
	/* FXP_MD_RX_AUTO_ADD */
	write_fxpbase(0x3C40F8, 0x0ffffffffffffde6);
	/* FXP_MD_RX_LAN_AUTO_DEL */
	write_fxpbase(0x3E80E8, 0x0fffffffffffffef);
}
